import pathlib

import numpy as np
import requests
import scipy.linalg
import timm
from loguru import logger
from numpy.linalg import LinAlgError
from numpy.random import default_rng
from sklearn import metrics
from torchvision import transforms
from tqdm import tqdm

CIFAR_NPZ_URL = "https://robustcifar.s3.us-west-2.amazonaws.com/cifar.npz"
CIFAR_10_1_DATA_NPY_URL = "https://github.com/modestyachts/CIFAR-10.1/raw/master/datasets/cifar10.1_v6_data.npy"
CIFAR_10_1_LABELS_NPY_URL = "https://github.com/modestyachts/CIFAR-10.1/raw/master/datasets/cifar10.1_v6_labels.npy"

RAND_AUG = timm.data.auto_augment.rand_augment_transform("rand-m9-mstd0.5-inc1", {})
TO_PIL = transforms.ToPILImage()
RESIZE = transforms.Resize((32,32))

def get_data_from_url(url, local_path):
    logger.info(f"Downloading: {url} to {local_path}")
    response = requests.get(url, stream=True)
    total_size_in_bytes= int(response.headers.get('content-length', 0))
    block_size = 1024 #1 Kibibyte
    progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)
    with open(local_path, 'wb') as f:
        for data in response.iter_content(block_size):
            progress_bar.update(len(data))
            f.write(data)
    progress_bar.close()
    if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
        assert False, f"Downloading from {URLS[variant]} failed"


def get_datasets():
    if not pathlib.Path("cifar.npz").exists():
        get_data_from_url(CIFAR_NPZ_URL, "cifar.npz")

    if not pathlib.Path("cifar10.1_v6_data.npy").exists():
        get_data_from_url(CIFAR_10_1_DATA_NPY_URL, "cifar10.1_v6_data.npy")

    if not pathlib.Path("cifar10.1_v6_labels.npy").exists():
        get_data_from_url(CIFAR_10_1_LABELS_NPY_URL, "cifar10.1_v6_labels.npy")

def grid_and_aug_img(x, tile_size, num_aug=100):
    imgs = []
    for i in range(0,x.shape[0],tile_size):
        for j in range(0,x.shape[0],tile_size):
            for _ in range(num_aug):
                img = RAND_AUG(RESIZE(TO_PIL(x[i:i+tile_size, j:j+tile_size])))
                imgs.append(np.array(img))
    return imgs

def solve(X,y,reg=1e-2, n_classes=10):
    X = X.reshape(X.shape[0], -1).astype('float64')
    y_enc = np.eye(n_classes)[y]
    XTX = X.T.dot(X)
    XTy = X.T.dot(y_enc)
    XTX += reg*np.eye(X.shape[1])
    return scipy.linalg.solve(XTX, XTy, sym_pos=True)

def predict(X, model):
    X = X.reshape(X.shape[0], -1).astype('float64')
    return X.dot(model)

def generate_cls_balance_sample(num_per_cls, y, classes={0,3}, seed=0):
    rng = default_rng(seed)
    idxs = []
    for cls in set(y):
        if cls in classes:
            idxs += list(rng.choice(np.where(y == cls)[0], num_per_cls, replace=False))
    return idxs

def rf_lift(X, M=4096, seed=0):
    rng = default_rng(seed)
    X = X.reshape(X.shape[0], -1).astype('float64')
    W = rng.standard_normal((X.shape[1], M))
    return np.maximum(X.dot(W), 0)

def gen_cifar_10_patch(X_train_full, y_train_full, X_val_full, y_val_full, X_test_full, y_test_full, seed, N_train=15, N_test=5):
    idxs = generate_cls_balance_sample(N_train, y_train_full, seed=seed)
    idxs_val = generate_cls_balance_sample(N_test, y_val_full, seed=seed)
    idxs_test = generate_cls_balance_sample(N_test, y_test_full, seed=seed)
    X_train_sample = X_train_full[idxs]
    X_val_sample = X_val_full[idxs_val]
    X_test_sample = X_test_full[idxs_test]

    y_train_sample = y_train_full[idxs]
    y_val_sample = y_val_full[idxs_val]
    y_test_sample = y_test_full[idxs_test]
    X_train = []
    X_test = []
    X_val = []

    y_train = []
    y_test = []
    y_val = []

    for i in range(len(X_train_sample)):
        X_samples = grid_and_aug_img(X_train_sample[i], 32, num_aug=400)
        X_train += X_samples
        y_train += [y_train_sample[i] for _ in range(len(X_samples))]

    for i in range(len(X_val_sample)):
        X_samples = grid_and_aug_img(X_val_sample[i], 32, num_aug=500)
        X_val += X_samples
        y_val += [y_val_sample[i] for _ in range(len(X_samples))]

    for i in range(len(X_test_sample)):
        X_samples = grid_and_aug_img(X_test_sample[i], 32, num_aug=500)
        X_test += X_samples
        y_test += [y_test_sample[i] for _ in range(len(X_samples))]

    X_train = np.stack(X_train) 
    X_test = np.stack(X_test) 
    X_val = np.stack(X_val)

    y_train = np.hstack(y_train) 
    y_test = np.hstack(y_test) 
    y_val = np.hstack(y_val)
    return (X_train,y_train), (X_val, y_val), (X_test, y_test)

def run_exp(dataset_seed=0, N_train=10, N_test=5):
    cifar = np.load("cifar.npz")
    X_train_full = cifar["X_train"]
    y_train_full = cifar["y_train"]
    X_val_full = cifar["X_test"]
    y_val_full = cifar["y_test"]
    X_test_full= np.load("cifar10.1_v6_data.npy")
    y_test_full = np.load("cifar10.1_v6_labels.npy")
    (X_train,y_train), (X_val, y_val), (X_test,y_test) = gen_cifar_10_patch(X_train_full, y_train_full, X_val_full, y_val_full, X_test_full, y_test_full, seed=dataset_seed, N_train=N_train, N_test=N_test)
    exps = []
    for seed in range(5):
        model = solve(X_train, y_train)
        y_val_pred = np.argmax(predict(X_val, model), axis=1)
        y_test_pred = np.argmax(predict(X_test, model), axis=1)
        exp = {}
        exp["N_train"] = N_train
        exp["N_test"] = N_test
        exp["classfier"] = "least_squares"
        exp["cifar_10_acc"] = metrics.accuracy_score(y_val_pred, y_val)
        exp["cifar_10_1_acc"] = metrics.accuracy_score(y_test_pred, y_test)
        exps.append(exp)
        logger.info(f"{seed},linear,{exp['cifar_10_acc']}, {exp['cifar_10_1_acc']}")
        for M in [128,256,512,1024]:
            for reg in [1, 1e9]:
                try:
                    model = solve(rf_lift(X_train, M=M, seed=seed), y_train, reg=reg)
                    y_val_pred = np.argmax(predict(rf_lift(X_val, M=M, seed=seed), model), axis=1)
                    y_test_pred = np.argmax(predict(rf_lift(X_test, M=M, seed=seed), model), axis=1)
                    exp = {}
                    exp["N_train"] = N_train
                    exp["N_test"] = N_test
                    exp["classfier"] = f"least_squares_rf_{M}"
                    exp["cifar_10_acc"] = metrics.accuracy_score(y_val_pred, y_val)
                    exp["cifar_10_1_acc"] = metrics.accuracy_score(y_test_pred, y_test)
                    exps.append(exp)
                    logger.info(f"{seed},{M},{reg},{exp['cifar_10_acc']}, {exp['cifar_10_1_acc']}")
                except LinAlgError:
                    logger.info(f"Singular reg {reg}")

if __name__ == "__main__":
    get_datasets()
    df0 = run_exp(N_train=100, N_test=1)
    df1 = run_exp(N_train=100, N_test=2)
    df2 = run_exp(N_train=100, N_test=5)
    df3 = run_exp(N_train=100, N_test=50)
    df4 = run_exp(N_train=100, N_test=500)
    df = pd.concat([df0,df1,df2,df2,df3,df4])
    with open("cifar_camelyon.pkl", "wb") as f:
        pickle.dump(df, f)
    print(df)




