"""Train and evaluate models in the testbed"""
import dataclasses
import re

import click
from tqdm import tqdm

import testbed
from testbed import mldb as mldb_module

@click.group()
def main():
    """Train and evaluate models."""
    pass


def parse_cmdline_hparams(param_cls, params_cmdline):
    """Parse cmdline parameters using the rule/model dataclass type hints."""
    fields = {f.name: f.type for f in dataclasses.fields(param_cls)}
    def cast_key(key):
        if key[:2] == "--":
            return key[2:]
        return key

    kwargs = {}
    for idx in range(0, len(params_cmdline), 2):
        name = cast_key(params_cmdline[idx])
        if name not in fields:
            raise ValueError(f"{name} not an rule parameter for {param_cls}")
        raw_val = params_cmdline[idx + 1]
        if raw_val.lower() == "true":
            cast_val = True
        elif raw_val.lower() == "false":
            cast_val = False
        else:
            cast_val = fields[name](raw_val)
        kwargs[name] = cast_val
    return param_cls(**kwargs)


@main.command("train", context_settings=dict(
    ignore_unknown_options=True,
))
@click.option("--rule_name", type=str, default=None)
@click.option("--model_family", type=str, default=None)
@click.option("--train_set", type=str, default=None)
@click.option("--tag_name", type=str, default=None)
@click.option('--num_model_repeats', type=int, default=1)
@click.option("--seed", type=int, default=None)
@click.option("--device", type=str, default=None)
@click.option("--store_early", is_flag=True)
@click.option('--num_checkpoints', type=int, default=10)
@click.option("--dry_run", is_flag=True, default=False)
@click.argument("rule_params_cmdline", nargs=-1, type=click.UNPROCESSED)
def train(rule_name, model_family, train_set, tag_name, seed,
          device, store_early, num_model_repeats,
          num_checkpoints, dry_run, rule_params_cmdline):
    """Train model on the specified datasets.

    Any unknown arguments passed are treated as rule_parameters, eg.
        python experiment.py train --model_family VGG13 --lr 0.001
    will train models with learning rate 1e-3.
    """
    assert model_family is not None or rule_name is not None
    assert train_set is not None

    print("Connecting to the database.")
    mldb = mldb_module.get_default_mldb()
    mldb.db.get(name=train_set, type_='dataset', assert_exists=True)


    if rule_name is not None:
        rule = mldb.learning_rule(name=rule_name)
        model_family = rule.model_family
        rule_params = rule.params
    else:
        rule_param_cls = mldb.model_registry[model_family].rule_params
        rule_params = parse_cmdline_hparams(rule_param_cls, rule_params_cmdline)
        rule = mldb.learning_rule(family=model_family, params=rule_params)

    print(f"Checking for matching {rule_name or model_family} models trained on {train_set}...")
    num_existing_models = 0
    for model in rule.get_trained_models():
        if model.train_set.name == train_set and(
                "seed" not in dataclasses.asdict(model.params)
                or seed == model.params.seed) and not model.is_checkpoint:
            num_existing_models += 1

    print(f"Found {num_existing_models} existing models trained with {rule_name or model_family}"
          f"{train_set} (rule params {rule_params})")
    if num_existing_models >= num_model_repeats:
        print("This is equal to or greater than num_model_repeats, so I am quitting")
        return


    kwargs = {}
    train_params_cls = mldb.model_registry[model_family].train_params
    train_params = [f.name for f in dataclasses.fields(train_params_cls)]
    if rule_name is not None:
        kwargs.update(dataclasses.asdict(
            parse_cmdline_hparams(train_params_cls, rule_params_cmdline)))
    if "seed" in train_params:
        kwargs["seed"] = seed
    if "device" in train_params:
        kwargs["device"] = device
    if "num_checkpoints" in train_params:
        kwargs["num_checkpoints"] = num_checkpoints

    print(f"Training model {rule_params or model_family} on {train_set} with {rule_params}, {kwargs}.")
    if not dry_run:
        model = rule.learn(dataset=train_set, store=store_early, **kwargs)
        model.store(tag=tag_name)


@main.command("evaluate", context_settings=dict(
    ignore_unknown_options=True,
))
@click.option("--model_family", type=str, default=None)
@click.option("--tag_name", type=str, default=None)
@click.option("--train_set", type=str, default=None)
@click.option("--test_set_pattern", type=str, default=None,
              help="Regex specifying datasets to evaluate on, e.g 'cifar10.*'")
@click.option("--eval_all_checkpoints", is_flag=True, default=False)
@click.option("--eval_checkpoints", type=str, default=None,
              help="Which if any checkpoints to evaluate, e.g.  '--eval_checkpoints 0,100'")
@click.option("--dry_run", is_flag=True, default=False)
@click.argument("eval_params_cmdline", nargs=-1, type=click.UNPROCESSED)
def evaluate(model_family, tag_name, train_set, test_set_pattern,
             eval_all_checkpoints, eval_checkpoints, dry_run,
             eval_params_cmdline):
    """Evaluate trained models on new datasets."""
    print("Connecting to the database.")
    mldb = mldb_module.get_default_mldb()
    mldb.db.get(name=train_set, type_='dataset', assert_exists=True)

    print(f'Fetching datasets matching {test_set_pattern}')
    testsets = [ds for ds in mldb.get_datasets()
                if re.match(test_set_pattern, ds.name)]
    print(f'Performing testing on {len(testsets)} datasets: '
          f'{", ".join([ds.name for ds in testsets])}')

    print(f"Fetching existing models trained on {train_set}...")
    matching_models = []
    if eval_checkpoints is not None:
        checkpoint_numbers = [int(x) for x in eval_checkpoints.split(",")]
    else:
        checkpoint_numbers = []
    for model in mldb.get_models(train_set=train_set, family=model_family, tag=tag_name):
        matching_models.append(model)
        if (eval_all_checkpoints or eval_checkpoints) and not model.is_checkpoint:
            matching_models.extend([c.as_model() for c in model.checkpoints
                                    if eval_all_checkpoints or
                                    (c.sequence_number in checkpoint_numbers)])
    print(f"Found {len(matching_models)} models to evaluate!")

    eval_tag_name = tag_name + f"_eval" if tag_name else None
    pbar = tqdm(total=len(matching_models) * len(testsets))
    for model in matching_models:
        existing_evals = model.get_evaluations()
        evaluated_testsets = set(ev.test_set.uuid for ev in existing_evals)
        for testset in testsets:
            if testset.uuid in evaluated_testsets:
                print(f"Model {model.name} already evaluated on {testset.name}...")
                pbar.update(1)
                continue

            print(f"Evaluating model {model.model_family} on {testset.name}")
            eval_params_cls = mldb.model_registry[model.model_family].eval_params
            eval_params = parse_cmdline_hparams(eval_params_cls, eval_params_cmdline)

            if not dry_run:
                evaluation = model.predict(dataset=testset,
                                           **dataclasses.asdict(eval_params))
                evaluation.store(tag=eval_tag_name)
            pbar.update(1)

    mldb.db.db_adapter.engine.dispose()


if __name__ == "__main__":
    main()
