"""Script to interact with (train/eval) neural models in parallel on a cluster."""
import concurrent.futures
import os
import subprocess
import sys

FREE_GPUS = [0, 1, 2, 3, 4, 5, 6, 7]

MODELS = [
    "PNASNetA",
    "RegNetX_400MF",
    "ResNeXt29_32x4d",
    "ShuffleNetV2_1",
    "densenet_cifar",
    "ShuffleNetV2_1_5",
    "ShuffleNetV2_0_5",
    "PNASNetB",
    "SENet18",
    "VGG13",
    "ShuffleNetG2",
    "MobileNet",
    "DenseNet201",
    "ShuffleNetV2_2",
]
TRAIN_DATASETS = ["cifar10-train", "cifar10.2-train", "cifar10-train_10k_0"]
TEST_DATASETS = ["cifar10-test", "cifar10.1-v6", "cifar10.2-test"]


def train_model(model, dataset, gpu_idx):
    """Launch a training job."""
    env = os.environ
    env["CUDA_VISIBLE_DEVICES"] = str(gpu_idx)
    command = [
        "python",
        "experiment.py",
        "train",
        "--model_family",
        model,
        "--train_set",
        dataset,
        "--device",
        "cuda:0", # Since we're setting CUDA VISIBLE DEVICES
    ]
    print(f"GPU: {gpu_idx}\t {' '.join(command)}")
    subprocess.call(command, env=env)


def evaluate_model(model, train_dataset, eval_dataset, gpu_idx):
    """Launch an evaluation job."""
    env = os.environ
    env["CUDA_VISIBLE_DEVICES"] = str(gpu_idx)
    command = [
        "python",
        "experiment.py",
        "evaluate",
        "--model_family",
        model,
        "--train_set",
        train_dataset,
        "--test_set_pattern",
        eval_dataset + "$",
        "--device",
        "cuda:0", # Since we're setting CUDA VISIBLE DEVICES
    ]
    print(f"GPU: {gpu_idx}\t {' '.join(command)}")
    subprocess.call(command, env=env)


def run_training():
    futures = []
    with concurrent.futures.ProcessPoolExecutor(max_workers=len(FREE_GPUS)) as executor:
        idx = 0
        for train_dataset in TRAIN_DATASETS:
            for model in MODELS:
                futures.append(
                    executor.submit(
                        train_model, model, train_dataset, FREE_GPUS[idx % len(FREE_GPUS)]
                    )
                )
                idx += 1
        for future in futures:
            _ = future.result()


def run_evaluations():
    futures = []
    with concurrent.futures.ProcessPoolExecutor(max_workers=len(FREE_GPUS)) as executor:
        idx = 0
        for train_dataset in TRAIN_DATASETS:
            for eval_dataset in TEST_DATASETS:
                for model in MODELS:
                    futures.append(
                        executor.submit(
                            evaluate_model, model, train_dataset, eval_dataset, FREE_GPUS[idx % len(FREE_GPUS)]
                        )
                    )
                    idx += 1
        for future in futures:
            _ = future.result()

if __name__ == "__main__":
    assert len(sys.argv) == 2
    if sys.argv[1] == "train":
        run_training()
    elif sys.argv[1] == "eval":
        run_evaluations()
    else:
        print("Must be one of train/eval")
        raise NotImplementedError
