"""Script to train/eval shallow models in parallel on a cluster."""
import concurrent.futures
from itertools import product
import os
import subprocess
import sys

MODELS = [
    "K_nearest_neighbors",
    "SVM",
    "RandomForest",
    "LogisticRegression",
    "RidgeClassifier",
    "QuadraticDiscrimantAnalysis",
    "AdaBoost",
]

TRAIN_DATASETS = ["cifar10-train", "cifar10.2-train", "cifar10-train_10k_0"]
TEST_DATASETS = ["cifar10-test", "cifar10.2-test"]

def get_hparams(family):
    """Return hyperparameter sweep for a given model family."""
    param_grids = {
        "SVM": {
            "kernel": ["linear", "rbf", "poly"],
            "C": [0.01, 0.1, 1.0, 10, 100],
            "gamma": ["auto", "scale"],
        },
        "RandomForest": {
            "criterion": ["gini", "entropy"],
            "max_depth": [3, 10, 20],
        },
        "LogisticRegression": {
            "C": [1e-6, 1e-2, 1.0],
            "max_iter": [2000],
        },
        "K_nearest_neighbors": {
            "n_neighbors": [3, 5, 7],
        },
        "RidgeClassifier": {
            "normalize": [True, False],
            "alpha": [1e-4, 1e-2, 1.0, 10.],
        },
        "AdaBoost": {
            "n_estimators": [20, 50, 100],
        },
    }
    if family not in param_grids:
        yield {}
    else:
        grid = param_grids[family]
        param_names = list(grid.keys())
        for param_values in product(*grid.values()):
            yield dict(zip(param_names, param_values))


def train_model(model, dataset, hparams):
    """Launch a training job."""
    command = [
        "python",
        "experiment.py",
        "train",
        "--model_family",
        model,
        "--train_set",
        dataset,
    ]
    for key, value in hparams.items():
        command.extend([
            f"--{key}",
            str(value),
        ])
    print(f"{' '.join(command)}")
    subprocess.call(command)


def evaluate_model(model, train_dataset, eval_dataset):
    """Launch an evaluation job."""
    command = [
        "python",
        "experiment.py",
        "evaluate",
        "--model_family",
        model,
        "--train_set",
        train_dataset,
        "--test_set_pattern",
        eval_dataset + "$",
    ]
    print(f"{' '.join(command)}")
    subprocess.call(command)


def run_training():
    futures = []
    with concurrent.futures.ProcessPoolExecutor() as executor:
        for train_dataset in TRAIN_DATASETS:
            for model in MODELS:
                for hparams in get_hparams(model):
                    futures.append(
                        executor.submit(train_model, model, train_dataset, hparams)
                    )
        for future in futures:
            _ = future.result()

def run_evaluations():
    futures = []
    with concurrent.futures.ProcessPoolExecutor() as executor:
        for train_dataset in TRAIN_DATASETS:
            for eval_dataset in TEST_DATASETS:
                for model in MODELS:
                    futures.append(
                        executor.submit(evaluate_model, model, train_dataset, eval_dataset)
                    )
        for future in futures:
            _ = future.result()

if __name__ == "__main__":
    assert len(sys.argv) == 2
    if sys.argv[1] == "train":
        run_training()
    elif sys.argv[1] == "eval":
        run_evaluations()
    else:
        print("Must be one of train/eval")
        raise NotImplementedError
