import collections
import sys
import traceback

sys.path.append("../")

import tqdm

from testbed.mldb import get_default_mldb


mldb = get_default_mldb(test=False)
mldb.db.allow_hard_delete = True


def checkpoint_numbers_to_keep(nums):
    if len(nums) <= 20:
        return nums
    # Always keep first 10 and the last checkpoint
    res = set(nums[:10])
    res.add(nums[-1])
    # Also keep every 10th checkpoint
    for ii in nums:
        if ii % 10 == 0:
            res.add(ii)
    return sorted(list(res))


def prune_model_checkpoints(cur_model, dry_run=True):
    cur_checkpoints = cur_model.checkpoints
    nums_to_keep = checkpoint_numbers_to_keep([x.sequence_number for x in cur_checkpoints])
    checkpoints_to_delete = []
    rels_to_delete = []
    blobs_to_delete = []
    for cur_chkpt in cur_checkpoints:
        if cur_chkpt.sequence_number not in nums_to_keep:
            checkpoints_to_delete.append(cur_chkpt)
    checkpoints_to_delete_uuids = set([x.uuid for x in checkpoints_to_delete])
    all_rels = mldb.db.get_relationship(first=cur_model.db_obj,
                                        type_='model_checkpoint',
                                        include_hidden=True)
    rels_to_delete = [x for x in all_rels if x.second in checkpoints_to_delete_uuids]
    blobs_to_delete_tmp = mldb.db.get_blobs([x.db_obj for x in checkpoints_to_delete],
                                            include_hidden=True)
    blobs_to_delete = []
    for v in blobs_to_delete_tmp.values():
        blobs_to_delete.extend(v.values())
    if not dry_run:
        mldb.db.delete_blob(blobs_to_delete, hide_only=False)
        mldb.db.delete_relationship(rels_to_delete, hide_only=False)
        mldb.db.delete([x.db_obj for x in checkpoints_to_delete], hide_only=False)
    return checkpoints_to_delete, rels_to_delete, blobs_to_delete


models = mldb.get_models(include_hidden=True)
hidden_models = [x for x in models if x.db_obj.hidden]
print(f'Total number of models: {len(models)}  ({len(hidden_models)} hidden models)')

print('Retrieving checkpoints by model ...')
num_checkpoints_by_model = {}
for x in tqdm.auto.tqdm(models):
    num_checkpoints_by_model[x.uuid] = len(x.checkpoints)

print('Checkpoint stats:')
tmpc = collections.Counter([x for x in num_checkpoints_by_model.values()])
for k, v in tmpc.most_common():
    print(f'    {v} models with {k} checkpoints')


models_to_process = [x for x in models if len(x.checkpoints) > 20]

for m in tqdm.tqdm(models_to_process):
    try:
        prune_model_checkpoints(m, dry_run=False)
    except:
        print(f'Caught exception while processing model {m.db_obj}')
        traceback.print_exc()
