"""Set up learning rules that reproduce the WILDS baseline fine-tuning
protocol for the iWildCam task"""
import tqdm
import sys

import testbed
from testbed.mldb import get_default_mldb

print("Loading MLDB")
mldb = get_default_mldb()

print("Getting all ImageNet model family names")
fam_names = [
    f for f in mldb.model_registry.families
    if f.startswith('imgnt_')]
print(f"Found {len(fam_names)} models")

print("Creating and storing learning rules")
progress_bar = tqdm.tqdm(fam_names)
for fam in progress_bar:
    net_name = fam.replace("imgnt_", "")
    progress_bar.set_postfix({'network': net_name})
    rule = mldb.learning_rule(
        family='imgnt_' + net_name,
        name='wilds_baseline_iWC_' + net_name,
        params=dict(
            num_classes=186,
            milestones=tuple(),
            epochs=18,
            lr=1e-5,
            weight_decay=0.0,
            batch_size=16,  # note: with such a small BS multi-gpu training will likely fail
            optimizer='Adam',
        )
    )
    rule.store()
    for lr in ('1e-2', '1e-3', '1e-4'):
        rule = mldb.learning_rule(
            family='imgnt_' + net_name,
            name=f'wilds_baseline_scratch_lr:{lr}_iWC_' + net_name,
            params=dict(
                pretrained=False,
                num_classes=186,
                milestones=tuple(),
                epochs=20,
                lr=float(lr),
                weight_decay=0.0,
                batch_size=16,  # note: with such a small BS multi-gpu training will likely fail
                optimizer='Adam',
            )
        )
        rule.store()
    rule = mldb.learning_rule(
        family='imgnt_' + net_name,
        name='wilds_baseline_iWC_' + net_name,
        params=dict(
            num_classes=186,
            milestones=tuple(),
            epochs=18,
            lr=1e-5,
            weight_decay=0.0,
            batch_size=16,  # note: with such a small BS multi-gpu training will likely fail
            optimizer='Adam',
        )
    )
    rule.store()
    for wd in ('1e-3',):
        for lr in ('1e-5',):
            rule = mldb.learning_rule(
                family='imgnt_' + net_name,
                name=f'wilds_baseline_lr:{lr}_wd:{wd}_iWC_' + net_name,
                params=dict(
                    num_classes=186,
                    milestones=tuple(),
                    epochs=18,
                    lr=float(lr),
                    weight_decay=float(wd),
                    batch_size=16,  # note: with such a small BS multi-gpu training will likely fail
                    optimizer='Adam',
                )
            )
            rule.store()
    for wd in ('1e-3',):
        for lr in ('1e-2',):
            rule = mldb.learning_rule(
                family='imgnt_' + net_name,
                name=f'wilds_baseline_scratch_lr:{lr}_wd:{wd}_iWC_' + net_name,
                params=dict(
                    pretrained=False,
                    num_classes=186,
                    milestones=tuple(),
                    epochs=20,
                    lr=float(lr),
                    weight_decay=float(wd),
                    batch_size=16,  # note: with such a small BS multi-gpu training will likely fail
                    optimizer='Adam',
                )
            )
            rule.store()
    rule = mldb.learning_rule(
        family='imgnt_' + net_name,
        name='wilds_baseline_lr:1e-3_iWC_' + net_name,
        params=dict(
            num_classes=186,
            milestones=tuple(),
            epochs=18,
            lr=1e-3,
            weight_decay=0.0,
            batch_size=16,  # note: with such a small BS multi-gpu training will likely fail
            optimizer='Adam',
        )
    )
    rule.store()
    rule = mldb.learning_rule(
        family='imgnt_' + net_name,
        name='wilds_baseline_lr:1e-4_iWC_' + net_name,
        params=dict(
            num_classes=186,
            milestones=tuple(),
            epochs=18,
            lr=1e-4,
            weight_decay=0.0,
            batch_size=16,  # note: with such a small BS multi-gpu training will likely fail
            optimizer='Adam',
        )
    )
    rule.store()
    rule = mldb.learning_rule(
        family='imgnt_' + net_name,
        name='wilds_baseline_scratch_iWC_' + net_name,
        params=dict(
            num_classes=186,
            pretrained=False,
            milestones=tuple(),
            epochs=18,
            lr=1e-5,
            weight_decay=0.0,
            batch_size=16,  # note: with such a small BS multi-gpu training will likely fail
            optimizer='Adam',
        )
    )
    rule.store()
    rule = mldb.learning_rule(
        family='imgnt_' + net_name,
        name='wilds_baseline_Camelyon_' + net_name,
        params=dict(
            num_classes=2,
            input_size=(96, 96),
            pretrained=False,
            batch_size=32,
            epochs=5,
            lr=0.001,
            momentum=0.9,
            optimizer='SGD',
            milestones=tuple(),
            weight_decay=0.01,
        )
    )
    rule.store()
    rule = mldb.learning_rule(
        family='imgnt_' + net_name,
        name='wilds_baseline224_Camelyon_' + net_name,
        params=dict(
            num_classes=2,
            input_size=(224, 244),
            pretrained=False,
            batch_size=32,
            epochs=5,
            lr=0.001,
            momentum=0.9,
            optimizer='SGD',
            milestones=tuple(),
            weight_decay=0.01,
        )
    )
    rule.store()
    rule = mldb.learning_rule(
        family='imgnt_' + net_name,
        name='wilds_baseline224_v2_Camelyon_' + net_name,
        params=dict(
            num_classes=2,
            input_size=(224, 224),
            pretrained=False,
            batch_size=32,
            epochs=5,
            lr=0.001,
            momentum=0.9,
            optimizer='SGD',
            milestones=tuple(),
            weight_decay=0.01,
        )
    )
    rule.store()
    rule = mldb.learning_rule(
        family='imgnt_' + net_name,
        name='wilds_baseline224_v2_pretrain_Camelyon_' + net_name,
        params=dict(
            num_classes=2,
            input_size=(224, 224),
            pretrained=True,
            batch_size=32,
            epochs=5,
            lr=0.001,
            momentum=0.9,
            optimizer='SGD',
            milestones=tuple(),
            weight_decay=0.01,
        )
    )
    rule.store()
    rule = mldb.learning_rule(
        family='imgnt_' + net_name,
        name='wilds_baseline224_v2_Camelyon_colorjitter_' + net_name,
        params=dict(
            num_classes=2,
            input_size=(224, 224),
            pretrained=False,
            batch_size=32,
            epochs=5,
            lr=0.001,
            momentum=0.9,
            optimizer='SGD',
            milestones=tuple(),
            weight_decay=0.01,
            color_jitter=True,
        )
    )
    rule.store()
