"""'In-memory' datasets stored in MLDB."""
import dataclasses
import typing as ty

import numpy as np

from ..dataset_family import DataFamily, DataProviderImplementation, DatasetImplementation


@dataclasses.dataclass
class MLDBDatasetFamilyParams:
    # Classes to ignore during evaluation, e.g. for logit subsampling.
    classes_to_exclude: ty.List[int] = None


class MLDBDatasetFamily(DataFamily):

    def provider(self, data_params):
        return MLDBDatasetProvider(data_params)
    
    @property
    def name(self):
        return "MLDBDatasetFamily"

    @property
    def params(self):
        return MLDBDatasetFamilyParams


class MLDBDatasetProvider(DataProviderImplementation):
    def __init__(self, data_params, examples=None, labels=None):
        self.data_params = data_params
        self.examples = examples
        self.labels = labels

    def _load_examples(self, mldb, db_obj):
        if self.examples is None:
            if "example_refs" in mldb.db.get_blobs(db_obj):
                example_refs = mldb.db.load_blob(db_obj, "example_refs")
                dset_uuids = set(dset_uuid for dset_uuid, _ in example_refs)
                example_map = {dset_uuid: mldb.db.load_blob(dset_uuid, "examples")
                               for dset_uuid in dset_uuids}
                examples = [example_map[ref_uuid][idx] for ref_uuid, idx in example_refs]
                self.examples = np.stack(examples, axis=0)
            else:
                self.examples = mldb.db.load_blob(db_obj, "examples")
        return self.examples

    def _load_labels(self, mldb, db_obj):
        if self.labels is None:
            if db_obj:
                if "labels" in mldb.db.get_blobs(db_obj):
                    self.labels = mldb.db.load_blob(db_obj, "labels")
                elif "example_refs" in mldb.db.get_blobs(db_obj):
                    example_refs = mldb.db.load_blob(db_obj, "example_refs")
                    dset_uuids = set(dset_uuid for dset_uuid, _ in example_refs)
                    label_map = {}
                    for dset_uuid in dset_uuids:
                        if "labels" in mldb.db.get_blobs(dset_uuid):
                            label_map[dset_uuid] = mldb.db.load_blob(dset_uuid, "labels")
                        else:
                            return None
                    labels = [label_map[ref_uuid][idx] for ref_uuid, idx in example_refs]
                    self.labels = np.array(labels)
        return self.labels

    def get_dataset(self, mldb, db_obj, show_progress):
        examples = self._load_examples(mldb, db_obj)
        labels = self._load_labels(mldb, db_obj)
        return MLDBDataset(examples, labels)

    def store_dataset(self, mldb, db_obj, show_progress):
        if self.examples is not None:
            if "examples" in mldb.db.get_blobs(db_obj):
                raise ValueError
            mldb.db.insert_blob(db_obj, "examples", data=self.examples)

        if self.labels is not None:
            if "labels" in mldb.db.get_blobs(db_obj):
                raise ValueError
            mldb.db.insert_blob(db_obj, "labels", data=self.labels)

    @property
    def family(self):
        return MLDBDatasetFamily()

    @property
    def params(self):
        return self.data_params

    def __len__(self):
        if self.examples is not None:
            return len(self.examples)
        raise ValueError

    def subsample(self, size, seed, mldb, db_obj):
        """Return a class balanced subset of the data."""

        examples = self._load_examples(mldb, db_obj)
        labels = self._load_labels(mldb, db_obj)
        full_dset_size = len(examples)

        if size is None:
            size = full_dset_size

        if size > full_dset_size:
            raise ValueError("Dataset size must be smaller than total dataset size.")

        rng = np.random.RandomState(seed)
        if labels is not None:
            # TODO: Assumes labels is a vector of scalars.
            class_idxs = np.unique(labels)
            if size % len(class_idxs) != 0:
                raise ValueError("Dataset size must be divisible by number of classes.")
            if size < len(class_idxs):
                raise ValueError("Dataset size must be larger than number of classes.")

            sampled_idxs = []
            examples_per_class = size // len(class_idxs)
            for class_idx in class_idxs:
                matching_idxs = np.argwhere(labels == class_idx).squeeze(axis=1)
                sampled_idxs.extend(
                    rng.choice(matching_idxs, examples_per_class, replace=False)
                )
        else:
            sampled_idxs = rng.choice(full_dset_size, size, replace=False)
        
        # TODO: Resurrect storing examples by reference to reduce storage size.
        #example_refs = self.get_references()
        #example_refs = [example_refs[idx] for idx in sampled_idxs]

        return MLDBDatasetProvider(self.data_params,
                                   examples=examples[sampled_idxs],
                                   labels=labels[sampled_idxs] if labels is not None else None)


class MLDBDataset(DatasetImplementation):
    def __init__(self, examples, labels):
        self._examples = examples
        self._labels = labels

    def __len__(self):
        return len(self._examples)

    def __getitem__(self, index):
        if self._labels is not None:
            return self._examples[index], self._labels[index]
        return self._examples[index]

    @property
    def examples(self):
        return self._examples

    @property
    def labels(self):
        return self._labels

def get_data_families():
    return {"MLDBDatasetFamily": MLDBDatasetFamily()}
