"""Dataloaders for WILDS datasets."""
import dataclasses
from functools import partial
import typing as ty
import os

import numpy as np
import torchvision.transforms as transforms
from wilds.datasets.camelyon17_dataset import Camelyon17Dataset
from wilds.datasets.download_utils import download_and_extract_archive
from wilds.datasets.fmow_dataset import FMoWDataset
from wilds.datasets.iwildcam_dataset import IWildCamDataset
from wilds.datasets.poverty_dataset import PovertyMapDataset
from wilds.datasets.waterbirds_dataset import WaterbirdsDataset

from ..dataset_family import DataFamily, DataProviderImplementation, DatasetImplementation, NoParameters


@dataclasses.dataclass
class WildsDataFamilyParams:
    dataset: str
    split: str
    image_size: ty.Tuple[int] = None
    image_encoding: str = "jpeg"
    # Classes to ignore during evaluation, e.g. for logit subsampling.
    classes_to_exclude: ty.List[int] = None

class WildsDataFamily(DataFamily):
    def __init__(self, data_param_cls):
        self.data_param_cls = data_param_cls

    def provider(self, data_params):
        download_url, variant = None, None
        if data_params.dataset == "IWildCamDataset":
            dataset_cls = IWildCamDataset
            if data_params.image_size == (224, 224) and data_params.image_encoding == "jpeg":
                download_url = "https://worksheets.codalab.org/rest/bundles/0xa4ee3e0fd22a4ff5a51c0e4fa7b73761/contents/blob/"
                variant = "224x224"
            elif data_params.image_size == (224, 224) and data_params.image_encoding == "png":
                download_url = "https://worksheets.codalab.org/rest/bundles/0x5709aa1ef1d14d1180a6ecf61c291540/contents/blob/"
                variant = "224x224png"
            elif data_params.image_encoding not in ["jpeg", "png"]:
                raise ValueError("Invalid image_encoding parameter for IWildCam")
            elif data_params.image_size is not None and data_params.image_size != (224, 224):
                raise ValueError("Invalid image_size parameter for IWildCam")
        elif data_params.dataset == "FMoWDataset":
            dataset_cls = partial(FMoWDataset, use_ood_val=True)
        elif data_params.dataset == "PovertyMapDataset":
            dataset_cls = PovertyMapDataset
        elif data_params.dataset == "Camelyon17Dataset":
            dataset_cls = Camelyon17Dataset
        elif data_params.dataset == "WaterbirdsDataset":
            dataset_cls = WaterbirdsDataset
        else:
            raise NotImplementedError
        size = self.lookup_size(data_params)
        return WildsDataProvider(
            dataset_cls, data_params, self, size, download_url, variant)

    def lookup_size(self, data_params):
        dataset_sizes = {
            "IWildCamDataset": {
                "train": 142202,
                "ood_val": 20784,
                "ood_test": 38943,
                "id_val": 7819,
                "id_test": 7861,
            },
            "FMoWDataset": {
                "train": 76863,
                "ood_val": 19915,
                "ood_test": 22108,
                "id_val": 11483,
                "id_test": 11327,
            },
            "PovertyMapDataset": {
                "train": 9797,
                "ood_val": 1000,
                "ood_test": 3963,
                "id_test": 1000,
            },
            "Camelyon17Dataset": {
                "train": 302436,
                "ood_val": 34904,
                "ood_test": 85054,
                "id_val": 33560,
            },
            "WaterbirdsDataset": {
                "train": 4795,
                "val": 1199,
                "test": 5794,
            },
        }
        return dataset_sizes[data_params.dataset][data_params.split]


    @property
    def name(self):
        return "WildsDataFamily"

    @property
    def params(self):
        return self.data_param_cls


class WildsDataProvider(DataProviderImplementation):
    def __init__(self, dataset_cls, data_params, family, length,
                 download_url=None, variant=None, indices=None):
        self.dataset_cls = dataset_cls
        self.data_params = data_params
        self._family = family
        self.length = length
        if download_url is not None:
            assert variant is not None
        self.download_url = download_url
        self.variant = variant
        self.indices = indices

    def _download_dataset(self, mldb):
        assert self.download_url is not None
        root_dir = os.path.join(mldb.data_cache_dir,
                                f"{self.data_params.dataset}_v{self.variant}")
        os.makedirs(root_dir, exist_ok=True)
        
        # Hack: Update dataset name and version to use the class
        # variables from self.dataset_cls once wilds is updated.
        _dataset_name = self.data_params.dataset.replace("Dataset", "").lower()
        _version = "1.0"
        data_dir = os.path.join(root_dir, f"{_dataset_name}_v{_version}")
        # If the data_dir exists and is not empty, and the download_url is set,
        # we assume the dataset is correctly set up
        if os.path.exists(data_dir) and len(os.listdir(data_dir)) > 0:
            return root_dir

        print(f"Downloading dataset to {data_dir}...")
        try:
            download_and_extract_archive(
                url=self.download_url,
                download_root=data_dir,
                filename='archive.tar.gz',
                remove_finished=True)
        except Exception as e:
            print(f"\n{os.path.join(data_dir, 'archive.tar.gz')} may be corrupted. Please try deleting it and rerunning this command.\n")
            print(f"Exception: ", e)

        return root_dir

    def _load_indices(self, mldb, db_obj):
        if self.indices is None:
            if db_obj and "indices" in mldb.db.get_blobs(db_obj):
                self.indices = mldb.db.load_blob(db_obj, "indices")
        return self.indices

    def get_dataset(self, mldb, db_obj, show_progress):
        if self.download_url is None:
            dataset = self.dataset_cls(root_dir=mldb.data_cache_dir, download=True)
        else:
            root_dir = self._download_dataset(mldb)
            dataset = self.dataset_cls(root_dir=root_dir, download=False)
        split = self.data_params.split
        if split == "ood_val":
            split = "val"
        elif split == "ood_test":
            split = "test"
        dataset_split = dataset.get_subset(split)
        
        # Check if any indices in the DB
        subset_indices = self._load_indices(mldb, db_obj)
        return WildsDataset(dataset_split, subset_indices)

    def store_dataset(self, mldb, db_obj, show_progress):
        if self.indices is not None:
            if "indices" in mldb.db.get_blobs(db_obj):
                raise ValueError
            mldb.db.insert_blob(db_obj, "indices", data=self.indices)

    @property
    def family(self):
        return self._family

    @property
    def params(self):
        return self.data_params

    def __len__(self):
        return self.length

    def subsample(self, size, seed, mldb, db_obj):
        # Note: We use IID subsampling here rather than class-balanced
        # subsampling. Wilds datasets are not classed balanced and in some cases
        # it's not possible to exactly preserve the class proportions in the
        # subsampled dataset.
        if size > self.length:
            raise ValueError("Dataset size must be smaller than the total dataset size")
        rng = np.random.RandomState(seed)
        sampled_idxs = rng.choice(self.length, size, replace=False)

        return WildsDataProvider(self.dataset_cls, self.data_params,
                                 self._family, size, self.download_url,
                                 self.variant, indices=sampled_idxs)

class WildsDataset(DatasetImplementation):
    def __init__(self, dataset, indices=None):
        self.dataset = dataset
        self._examples = None
        self._metadata = None
        self.indices = indices

    def __len__(self):
        if self.indices is None:
            return len(self.dataset)
        return len(self.indices)

    def __getitem__(self, index):
        if self.indices is None:
            return self.dataset[index]
        return self.dataset[self.indices[index]]
    
    @property
    def labels(self):
        labels = self.dataset.y_array.numpy()
        if self.indices is None:
            return labels
        return labels[self.indices]

    def get_wilds_dataset(self):
        return self.dataset

    @property
    def examples(self):
        if self._examples is None:
            if self.dataset.dataset_name == "fmow":
                self._examples = np.stack([np.copy(x) for x, _, _ in self])
            elif self.dataset.dataset_name == "camelyon17":
                self._examples = np.stack([np.array(x) for x, _, _ in self])
            else:
                raise NotImplementedError
        return self._examples

    @property
    def metadata(self):
        if self._metadata is None:
            metadata = [z.clone().detach().numpy() for _, _, z in self]
            self._metadata = np.stack(metadata)
        return self._metadata


def get_data_families():
    return {
        "WildsDataFamily": WildsDataFamily(WildsDataFamilyParams)
    }
