"""Base dataset class."""
import copy
import dataclasses
import uuid

from basicdb import Blob, Object
import numpy as np

from .data_families import default_registry
from .data_families.mldb_data import MLDBDatasetFamilyParams, MLDBDatasetProvider
from . import dataset_family


class Dataset:
    def __init__(self,
                 *,
                 name=None,
                 db_obj=None,
                 provider=None,
                 family=None,
                 params=None,
                 examples=None,
                 labels=None,
                 mldb=None):
        self.mldb = mldb
        if mldb is not None:
            self.db = mldb.db
            self.registry = mldb.data_registry
        else:
            self.db = None
            self.registry = default_registry()
        self.impl = None
        self.db_obj = db_obj
        if db_obj is not None:
            assert all(x is None for x in [name, provider, family, params])
            self._init_from_db_obj()
        elif provider is not None:
            assert all(x is None for x in [db_obj, family, params])
            self.provider = provider
            self.name = name
        elif family is not None:
            assert all(x is None for x in [db_obj, provider])
            assert params is not None
            self.name = name
            if isinstance(params, dict):
                cur_params = self.registry[family].params(**params)
            else:
                assert isinstance(params, self.registry[family].params)
                cur_params = params
            self.provider = self.registry[family].provider(cur_params)
        elif examples is not None:
            assert all(x is None for x in [db_obj, provider, family])
            if self.db is not None and name is not None:
                if self.db.get(type_='dataset', name=name) is not None:
                    raise ValueError
            self.name = name
            if params is None:
                cur_params = MLDBDatasetFamilyParams()
            elif isinstance(params, dict):
                cur_params = MLDBDatasetFamilyParams(**params)
            else:
                assert isinstance(params, MLDBDatasetFamilyParams)
                cur_params = params
            self.provider = MLDBDatasetProvider(
                data_params=cur_params, examples=examples, labels=labels)
        else:
            assert name is not None and mldb is not None
            assert all(x is None for x in [db_obj, provider, family, params])
            self.db_obj = self.db.get(type_='dataset', name=name, assert_exists=True)
            self._init_from_db_obj()

    def _init_from_db_obj(self):
        family_name = self.db_obj.subtype
        self.provider = self.registry[family_name].provider(
                self.registry[family_name].params(**self.db_obj.params))
        self.name = self.db_obj.name

    def _get_dataset_if_necessary(self, show_progress=True):
        if self.impl is None:
            self.impl = self.provider.get_dataset(mldb=self.mldb,
                                                  db_obj=self.db_obj,
                                                  show_progress=show_progress)

    def __len__(self):
        if self.db_obj is not None:
            return self.db_obj.dataset_length
        return len(self.provider)

    def __getitem__(self, index):
        self._get_dataset_if_necessary()
        return self.impl[index]

    @property
    def examples(self):
        self._get_dataset_if_necessary()
        return self.impl.examples

    @property
    def labels(self):
        self._get_dataset_if_necessary()
        return self.impl.labels

    @property
    def implementation(self):
        self._get_dataset_if_necessary()
        return self.impl

    @property
    def family(self):
        return self.provider.family
    
    @property
    def params(self):
        return self.provider.params
    
    @property
    def is_stored(self):
        return self.db_obj is not None
    
    @property
    def uuid(self):
        if self.db_obj is not None:
            return self.db_obj.uuid
        else:
            return None

    def rename(self, new_name):
        if self.is_stored:
            self.db.update(self.db_obj, new_name=new_name)
        self.name = new_name

    def store(self, assert_not_stored_yet=False, show_progress=True):
        assert self.db is not None
        if self.is_stored:
            if assert_not_stored_yet:
                raise ValueError
            return
        self.db_obj = self.db.insert(
            name=self.name,
            type_= 'dataset',
            subtype=self.family.name,
            extra={
                'params': dataclasses.asdict(self.params),
                'dataset_length': len(self),
            },
        )
        self.name = self.db_obj.name    # in case name is None, basicdb assigns the uuid
        self.provider.store_dataset(mldb=self.mldb, db_obj=self.db_obj, show_progress=show_progress)

    def __repr__(self):
        res = '<Dataset(name='
        if self.name is None:
            res += 'None'
        else:
            res += f'"{self.name}"'
        if self.is_stored:
            res += f', uuid={self.uuid}'
        res += f' at {hex(id(self))}>'
        return res

    def subsample(self, size, seed, name=None):
        new_provider = self.provider.subsample(
            size=size, seed=seed, mldb=self.mldb, db_obj=self.db_obj)
        return Dataset(provider=new_provider, name=name, mldb=self.mldb)


    # In the future
    # * merge
    # * subset
    # * stratify
    # * sample
    # * etc.  (as in Dataset before)
