from abc import ABC, abstractmethod
import dataclasses
import math
import uuid

import basicdb
import numpy as np
from numpy.random import default_rng
import sklearn.metrics
import statsmodels.stats.api as sms
import torch

from .data_families.wilds_data import WildsDataset
from .dataset import Dataset
from . import model
from .model_families import default_registry


class StorableEvaluationStatistic(ABC):
    @abstractmethod
    def compute(self, predictions, test_set):
        pass


class InapplicableStatisticError(Exception):
    pass


class SimpleStorableStatistic(StorableEvaluationStatistic):
    def __init__(self, fn):
        self.fn = fn
    
    def compute(self, predictions, test_set):
        return self.fn(predictions, test_set)


def num_correct_and_size_fn(predictions, test_set):
    if test_set.labels is None:
        raise InapplicableStatisticError
    return int(np.sum(test_set.labels == predictions)), len(test_set.labels)


def macro_f1_fn(predictions, test_set):
    if test_set.labels is None:
        raise InapplicableStatisticError
    return float(sklearn.metrics.f1_score(test_set.labels, predictions, average='macro'))


def wilds_metrics_fn(predictions, test_set):
    if not isinstance(test_set.implementation, WildsDataset):
        raise InapplicableStatisticError
    metadata = torch.from_numpy(test_set.implementation.metadata)
    labels = torch.from_numpy(test_set.labels)
    predictions = torch.from_numpy(predictions)

    wilds_dset = test_set.implementation.get_wilds_dataset()
    # By default, wilds metrics expects logits, which are internally
    # converted into predictions. Update the prediction_fn to
    # assume class predictions are being passed.
    if hasattr(wilds_dset.dataset, "_metric"):
        metrics = [wilds_dset.dataset._metric]
    elif hasattr(wilds_dset.dataset, "_metrics"):
        metrics = wilds_dset.dataset._metrics
    else:
        raise NotImplementedError
    for metric in metrics:
        metric.prediction_fn = None
    results_dict, _ = wilds_dset.eval(predictions, labels, metadata)
    return results_dict


def compute_combined_ci(per_class_counts, coverage_percentile):
    num_classes = len(per_class_counts)
    upper_cis = []
    lower_cis = []
    for cur_count in per_class_counts:
        ci = sms.proportion_confint(0.5 * cur_count,
                                    cur_count,
                                    (100.0 - coverage_percentile) / 100.0,
                                    method='beta')
        upper_cis.append(ci[1] - 0.5)
        lower_cis.append(0.5 - ci[0])
    return np.mean(upper_cis) / math.sqrt(num_classes), np.mean(lower_cis) / math.sqrt(num_classes)


def make_iwc_f1_approx_ci(coverage_percentile):
    def fn(predictions, test_set):
        if not isinstance(test_set.implementation, WildsDataset):
            raise InapplicableStatisticError
        point_estimate = wilds_metrics_fn(predictions, test_set)['F1-macro_all']
        _, per_class_counts = np.unique(test_set.labels, return_counts=True)
        ci_size = compute_combined_ci(per_class_counts, coverage_percentile)
        return float(point_estimate - ci_size[0]), float(point_estimate + ci_size[0])
    return fn


storable_statistics = {
    'num_correct_and_size': SimpleStorableStatistic(num_correct_and_size_fn),
    'macro_f1': SimpleStorableStatistic(macro_f1_fn),
    'wilds_metrics': SimpleStorableStatistic(wilds_metrics_fn),
    'iwc_f1_approx_ci_95': SimpleStorableStatistic(make_iwc_f1_approx_ci(95)),
}

default_statistics_to_store = ['num_correct_and_size', 'macro_f1', 'wilds_metrics']


class Evaluation():
    def __init__(self,
                 db_obj=None,
                 model=None,
                 test_set=None,
                 params={},
                 predictions=None,
                 extra_prediction_data=None,
                 mldb=None):
        self.mldb = mldb
        if mldb is not None:
            self.db = mldb.db
            self.registry = mldb.model_registry
        else:
            self.db = None
            self.registry = default_registry()
        if isinstance(db_obj, basicdb.Object):
            assert self.db is not None
            self.db_obj = db_obj
            self.params = self.registry[self.db_obj.subtype].eval_params(**self.db_obj.params)
            assert all([x is None for x in [model, test_set, predictions, extra_prediction_data]])
            assert params == {}
            self._test_set = None
            self._model = None
            self._predictions = None
            self._extra_prediction_data = None
            self.computed_statistics = {} if 'computed_statistics' not in self.db_obj.extra else self.db_obj.computed_statistics
        elif isinstance(db_obj, uuid.UUID):
            assert self.db is not None
            assert all([x is None for x in [model, test_set, predictions, extra_prediction_data]])
            assert params == {}
            self.db_obj = self.db.get(uuid=db_obj, type_='evaluation')
            self.params = self.registry[self.db_obj.subtype].eval_params(**self.db_obj.params)
            self._test_set = None
            self._model = None
            self._predictions = None
            self._extra_prediction_data = None
            self.computed_statistics = {} if 'computed_statistics' not in self.db_obj.extra else self.db_obj.computed_statistics
        else:
            assert db_obj is None
            assert all([x is not None for x in [model, test_set, predictions]])
            self._test_set = test_set
            self._model = model
            self.params = params
            self._predictions = predictions
            self._extra_prediction_data = extra_prediction_data if extra_prediction_data is not None else {}
            self.db_obj = None
            self.computed_statistics = {}

    @property
    def uuid(self):
        if self.db_obj is not None:
            return self.db_obj.uuid
        else:
            return None
    
    @property
    def predictions(self):
        if self._predictions is None:
            self._predictions = self.db.load_blob(self.db_obj, "prediction_data")     
        return self._predictions
    
    @property
    def extra_prediction_data(self):
        if self._extra_prediction_data is None:
            self._extra_prediction_data = self.db.load_blob(self.db_obj, "extra_prediction_data")
        return self._extra_prediction_data
    
    @property
    def model(self):
        if self._model is None:
            models = self.db.get(rel_first=self.db_obj, rel_type="model_evaluated", type_="model")
            assert len(models) == 1
            self._model = models[0]
        if isinstance(self._model, basicdb.Object):
            self._model = model.Model(db_obj=self._model, mldb=self.mldb)
        return self._model
    
    @property
    def test_set(self):
        if self._test_set is None:
            test_sets = self.db.get(rel_first=self.db_obj, rel_type="test_set", type_="dataset")
            assert len(test_sets) == 1, f'Evaluation {self.db_obj} has {len(test_sets)} associated test sets'
            self._test_set = test_sets[0]
        if isinstance(self._test_set, basicdb.Object):
            self._test_set = Dataset(db_obj=self._test_set, mldb=self.mldb)
        return self._test_set
    
    def get_statistic(self, name, recompute=False, skip_store=False):
        assert name in storable_statistics
        if recompute or name not in self.computed_statistics:
            self.computed_statistics[name] = storable_statistics[name].compute(self.predictions, self.test_set)
            if self.is_stored and not skip_store:
                assert self.db is not None
                self.db_obj.computed_statistics = self.computed_statistics
                self.db.update(self.db_obj)
        return self.computed_statistics[name]

    @property
    def is_stored(self):
        return self.db_obj is not None
    
    def store(self,
              statistics_to_store=None,      # Store all statistics in default_statistics_to_store by default
              assert_not_stored_yet=False,
              tag=None):
        assert self.db is not None and self.db_obj is None
        if self.is_stored:
            if assert_not_stored_yet:
                raise ValueError
            return

        self.model.store(assert_not_stored_yet=False)
        self.test_set.store(assert_not_stored_yet=False)

        if statistics_to_store is None:
            statistics_to_store = default_statistics_to_store
        for stat_name in statistics_to_store:
            try:
                self.get_statistic(stat_name)
            except InapplicableStatisticError:
                pass
        
        extra = {'params': dataclasses.asdict(self.params),
                 'testset_uuid': str(self.test_set.uuid),           # Why do we store this here?
                 'computed_statistics': self.computed_statistics}

        self.db_obj = self.db.insert(
            type_='evaluation',
            subtype=self.model.model_family,
            extra=extra,
            blobs={
                'prediction_data': self.predictions,
                'extra_prediction_data': self.extra_prediction_data
            }
        )
        self.db.insert_relationship(
            first=self.db_obj, second=self.test_set.db_obj, type_="test_set"
        )
        self.db.insert_relationship(
            first=self.db_obj, second=self.model.db_obj, type_="model_evaluated"
        )
        
        if tag is not None:
            self.mldb.add_tag(self, tag)

    @property
    def accuracy(self):
        num_correct, test_set_size = self.get_statistic('num_correct_and_size')
        return num_correct / test_set_size

    @property
    def acc_conf_int(self, alpha=0.05):
        num_correct, test_set_size = self.get_statistic('num_correct_and_size')
        return sms.proportion_confint(num_correct,
                                      test_set_size,
                                      alpha=alpha,
                                      method="beta")

    @property
    def error_rate(self):
        return 1.0 - self.accuracy

    @property
    def macro_f1(self):
        return self.get_statistic('macro_f1')

    @property
    def wilds_metrics(self):
        return self.get_statistic('wilds_metrics')
