import dataclasses
import uuid

import basicdb
import numpy as np

from .dataset import Dataset
from . import model
from .model_families import default_registry


class Featurization():
    def __init__(self,
                 db_obj=None,
                 model=None,
                 dataset=None,
                 params={},
                 features=None,
                 extra_featurization_data=None,
                 mldb=None):
        self.mldb = mldb
        if mldb is not None:
            self.db = mldb.db
            self.registry = mldb.model_registry
        else:
            self.db = None
            self.registry = default_registry()
        if isinstance(db_obj, basicdb.Object):
            assert self.db is not None
            self.db_obj = db_obj
            self.params = self.registry[self.db_obj.subtype].eval_params(**self.db_obj.params)
            assert all([x is None for x in [model, dataset, features, extra_featurization_data]])
            assert params == {}
            self._dataset = None
            self._model = None
            self._features = None
            self._extra_featurization_data = None
        elif isinstance(db_obj, uuid.UUID):
            assert self.db is not None
            assert all([x is None for x in [model, dataset, features, extra_featurization_data]])
            assert params == {}
            self.db_obj = self.db.get(uuid=db_obj, type_='featurization')
            self.params = self.registry[self.db_obj.subtype].eval_params(**self.db_obj.params)
            self._dataset = None
            self._model = None
            self._features = None
            self._extra_featurization_data = None
        else:
            assert db_obj is None
            assert all([x is not None for x in [model, dataset, features]])
            self._dataset = dataset
            self._model = model
            self.params = params
            self._features = features
            self._extra_featurization_data = extra_featurization_data if extra_featurization_data is not None else {}
            self.db_obj = None

    @property
    def uuid(self):
        if self.db_obj is not None:
            return self.db_obj.uuid
        else:
            return None
    
    @property
    def features(self):
        if self._features is None:
            self._features = self.db.load_blob(self.db_obj, "features")     
        return self._features
    
    @property
    def extra_featurization_data(self):
        if self._extra_featurization_data is None:
            self._extra_featurization_data = self.db.load_blob(self.db_obj, "extra_featurization_data")
        return self._extra_featurization_data
    
    @property
    def model(self):
        if self._model is None:
            models = self.db.get(rel_first=self.db_obj, rel_type="model_featurized", type_="model")
            assert len(models) == 1
            self._model = models[0]
        if isinstance(self._model, basicdb.Object):
            self._model = model.Model(db_obj=self._model, mldb=self.mldb)
        return self._model
    
    @property
    def dataset(self):
        if self._dataset is None:
            datasets = self.db.get(rel_first=self.db_obj, rel_type="featurization_dataset", type_="dataset")
            assert len(datasets) == 1, f'Featurization {self.db_obj} has {len(datasets)} associated datasets'
            self._dataset = datasets[0]
        if isinstance(self._dataset, basicdb.Object):
            self._dataset= Dataset(db_obj=self._dataset, mldb=self.mldb)
        return self._dataset
    
    @property
    def is_stored(self):
        return self.db_obj is not None
    
    def store(self,
              assert_not_stored_yet=False,
              tag=None):
        assert self.db is not None and self.db_obj is None
        if self.is_stored:
            if assert_not_stored_yet:
                raise ValueError
            return

        self.model.store(assert_not_stored_yet=False)
        self.dataset.store(assert_not_stored_yet=False)

        extra = {'params': dataclasses.asdict(self.params)}

        self.db_obj = self.db.insert(
            type_='featurization',
            subtype=self.model.model_family,
            extra=extra,
            blobs={
                'features': self.features,
                'extra_featurization_data': self.extra_featurization_data
            }
        )
        self.db.insert_relationship(
            first=self.db_obj, second=self.dataset.db_obj, type_="featurization_dataset"
        )
        self.db.insert_relationship(
            first=self.db_obj, second=self.model.db_obj, type_="model_featurized"
        )
        
        if tag is not None:
            self.mldb.add_tag(self, tag)