"""Initialize database for the project."""
import io
from itertools import chain
import pathlib
import requests
import shutil
import ssl
import tarfile
import tempfile
import urllib

import numpy as np
import torch
import torchvision
from torchvision import transforms

from testbed.mldb import get_default_mldb


def add_cifar10(mldb):
    """Add base cifar10 datasets to the database."""
    found_datasets = [dset.name for dset in mldb.get_datasets()]
    if all(dset in found_datasets for dset in ["cifar10-all", "cifar10-train", "cifar10-test"]):
        print("CIFAR10 already loaded into DB. Skipping...")
        return

    with tempfile.TemporaryDirectory() as tmpdirname:
        trainset = torchvision.datasets.CIFAR10(root=tmpdirname, train=True, download=True)
        testset = torchvision.datasets.CIFAR10(root=tmpdirname, train=False, download=True)

    train_images = np.array(trainset.data)
    test_images = np.array(testset.data)

    train_labels = np.array(trainset.targets)
    test_labels = np.array(testset.targets)

    all_images = np.concatenate([train_images, test_images])
    all_labels = np.concatenate([train_labels, test_labels])

    if "cifar10-all" not in found_datasets:
        mldb.dataset(name="cifar10-all", examples=all_images, labels=all_labels).store()
    if "cifar10-train" not in found_datasets:
        mldb.dataset(name="cifar10-train", examples=train_images, labels=train_labels).store()
    if "cifar10-test" not in found_datasets:
        mldb.dataset(name="cifar10-test", examples=test_images, labels=test_labels).store()


def add_cifar10p2(mldb):
    """Add cifar10.2 dataset to the database."""

    found_datasets = [dset.name for dset in mldb.get_datasets()]
    if all(dset in found_datasets for dset in ["cifar10.2-all", "cifar10.2-train", "cifar10.2-test"]):
        print("CIFAR10.2 already loaded into DB. Skipping...")
        return

    train_url = "https://github.com/modestyachts/cifar-10.2/raw/master/cifar102_train.npy"
    test_url = "https://github.com/modestyachts/cifar-10.2/raw/master/cifar102_test.npy"

    train_response = requests.get(train_url)
    train_response.raise_for_status()
    train = np.load(io.BytesIO(train_response.content), allow_pickle=True).item()

    test_response = requests.get(test_url)
    test_response.raise_for_status()
    test = np.load(io.BytesIO(test_response.content), allow_pickle=True).item()

    all_images = np.concatenate([train["images"], test["images"]])
    all_labels = np.concatenate([train["labels"], test["labels"]])

    if "cifar10.2-all" not in found_datasets:
        mldb.dataset(name="cifar10.2-all", examples=all_images, labels=all_labels).store()
    if "cifar10.2-train" not in found_datasets:
        mldb.dataset(name="cifar10.2-train", examples=train["images"], labels=train["labels"]).store()
    if "cifar10.2-test" not in found_datasets:
        mldb.dataset(name="cifar10.2-test", examples=test["images"], labels=test["labels"]).store()

def add_cifar10p1(mldb):
    """Add cifar10.1 v4 and v6 to the database."""
    found_datasets = [dset.name for dset in mldb.get_datasets()]
    if all(dset in found_datasets for dset in ["cifar10.1-v4", "cifar10.1-v6"]):
        print("CIFAR10.1 already loaded into DB. Skipping...")
        return

    v4_data_url = "https://github.com/modestyachts/CIFAR-10.1/raw/master/datasets/cifar10.1_v4_data.npy"
    v4_labels_url = "https://github.com/modestyachts/CIFAR-10.1/raw/master/datasets/cifar10.1_v4_labels.npy"
    v6_data_url = "https://github.com/modestyachts/CIFAR-10.1/raw/master/datasets/cifar10.1_v6_data.npy"
    v6_labels_url = "https://github.com/modestyachts/CIFAR-10.1/raw/master/datasets/cifar10.1_v6_labels.npy"

    def download(url):
        response = requests.get(url)
        response.raise_for_status()
        return np.load(io.BytesIO(response.content), allow_pickle=True)

    if "cifar10.1-v4" not in found_datasets:
        v4_images = download(v4_data_url)
        v4_labels = download(v4_labels_url)
        mldb.dataset(name="cifar10.1-v4", examples=v4_images, labels=v4_labels).store()
    if "cifar10.1-v6" not in found_datasets:
        v6_images = download(v6_data_url)
        v6_labels = download(v6_labels_url)
        mldb.dataset(name="cifar10.1-v6", examples=v6_images, labels=v6_labels).store()

def add_cifar10c(mldb):
    """Add cifar10-c (all corruptions and severity levels) to the dataset."""
    CORRUPTIONS = [
        "brightness",
        "contrast",
        "defocus_blur",
        "elastic_transform",
        "fog",
        "frost",
        "gaussian_blur",
        "gaussian_noise",
        "glass_blur",
        "impulse_noise",
        "jpeg_compression",
        "motion_blur",
        "pixelate",
        "saturate",
        "shot_noise",
        "snow",
        "spatter",
        "speckle_noise",
        "zoom_blur",
    ]
    dataset_names = [f"cifar10c_{corruption}_{level}" for corruption in CORRUPTIONS for level in list(range(1, 6)) + ["all"]]
    found_datasets = [dset.name for dset in mldb.get_datasets()]
    if all(dset in found_datasets for dset in dataset_names):
        print("CIFAR10C already loaded into DB. Skipping...")
        return

    # Download the datasets
    download_url = "https://zenodo.org/record/2535967/files/CIFAR-10-C.tar?download=1"
    with tempfile.NamedTemporaryFile("wb") as handle:
        with requests.get(download_url, stream=True) as response:
            shutil.copyfileobj(response.raw, handle)

        with tarfile.open(handle.name) as tar_object:
            label_tfile = tar_object.extractfile("CIFAR-10-C/labels.npy")
            all_labels = np.load(io.BytesIO(label_tfile.read()))
            for corruption in CORRUPTIONS:
                tf = tar_object.extractfile(f"CIFAR-10-C/{corruption}.npy")
                all_examples = np.load(io.BytesIO(tf.read()))
                for level in list(range(1, 6)) + ["all"]:
                    if level == "all":
                        examples = all_examples
                        labels = all_labels
                    else:
                        examples = all_examples[(level - 1) * 10000: level * 10000]
                        labels = all_labels[(level - 1) * 10000: level * 10000]
                    
                    dset_name = f"cifar10c_{corruption}_{level}"
                    if dset_name not in found_datasets:
                        dataset = mldb.dataset(name=dset_name, examples=examples, labels=labels)
                        dataset.store()


def add_wilds(mldb):
    """Add Wilds datasets to MLDB."""
    wilds_datasets = {
        "IWildCam": ["train", "ood_val", "ood_test", "id_val", "id_test"],
        "FMoW": ["train", "ood_val", "ood_test", "id_val", "id_test"],
        "PovertyMap": ["train", "ood_val", "ood_test", "id_test"],
        "Camelyon17": ["train", "ood_val", "ood_test", "id_val"],
        "Waterbirds": ["train", "val", "test"],
    } 
    found_datasets = [dset.name for dset in mldb.get_datasets()]
    for name, splits in wilds_datasets.items():
        for split in splits:
            dset_name = f"{name}-{split}"
            if dset_name not in found_datasets:
                mldb.dataset(family="WildsDataFamily",
                             params={"dataset": name + "Dataset", "split": split},
                             name=dset_name).store()

    # Extra variants of IWildCam
    for split in wilds_datasets["IWildCam"]:
        dset_name = f"IWildCamSize224-{split}"
        if dset_name not in found_datasets:
            mldb.dataset(family="WildsDataFamily",
                         params={
                            "dataset": "IWildCamDataset",
                            "split": split, 
                            "image_size": (224, 224),
                         },
                         name=dset_name).store()
        dset_name = f"IWildCamSize224PNG-{split}"
        if dset_name not in found_datasets:
            mldb.dataset(family="WildsDataFamily",
                         params={
                            "dataset": "IWildCamDataset",
                            "split": split, 
                            "image_size": (224, 224),
                            "image_encoding": "png",
                         },
                         name=dset_name).store()

def add_stl10(mldb):
    """Add STL10 datasets to MLDB."""
    found_datasets = [dset.name for dset in mldb.get_datasets()]
    if all(dset in found_datasets for dset in ["STL10", "cifar10-STL10classes"]):
        print("STL10 and CIFAR10-9class already loaded into DB. Skipping...")
        return

    cifar10_labels = {cls: idx for idx, cls in enumerate(
                      ["airplane", "automobile", "bird", "cat", "deer",
                       "dog", "frog", "horse", "ship", "truck"])}
    stl10_labels = {idx: cls for idx, cls in enumerate(
                    ["airplane", "bird", "automobile", "cat", "deer",
                     "dog", "horse", "monkey", "ship", "truck"])}
    stl10_classes = list(stl10_labels.values())
    cifar10_classes_to_exclude = [cls_idx for label, cls_idx in cifar10_labels.items()
                                  if label not in stl10_classes]

    with tempfile.TemporaryDirectory() as tmpdirname:
        stl10_train = torchvision.datasets.STL10(root=tmpdirname, split="train", download=True,
                                                 transform=transforms.Resize((32, 32)))
        stl10_test = torchvision.datasets.STL10(root=tmpdirname, split="test", download=True,
                                                transform=transforms.Resize((32, 32)))

    # Merge all STL10 training and testing images to create a larger dataset
    images, labels = [], []
    for img, label in chain(stl10_train, stl10_test):
        # Ignore labels not in cifar10
        if stl10_labels[label] not in cifar10_labels:
            continue
        images.append(np.array(img))
        labels.append(cifar10_labels[stl10_labels[label]])

    examples = np.stack(images)
    labels = np.array(labels)

    if "STL10" not in found_datasets:
        mldb.dataset(name="STL10", examples=examples, labels=labels,
                     params={"classes_to_exclude": cifar10_classes_to_exclude}).store()

    # Create a 9 class version of cifar10 for comparison.
    if "cifar10-test" not in found_datasets:
        add_cifar10(mldb)
    cifar10_test = mldb.dataset(name="cifar10-test")
    subset_examples, subset_labels = [], []
    for example, label in cifar10_test:
        if label in cifar10_classes_to_exclude:
            continue
        subset_examples.append(example)
        subset_labels.append(label)
    subset_examples = np.stack(subset_examples)
    subset_labels = np.array(subset_labels)

    if "cifar10-STL10classes" not in found_datasets:
        mldb.dataset(name="cifar10-STL10classes", examples=subset_examples, labels=subset_labels,
                     params={"classes_to_exclude": cifar10_classes_to_exclude}).store()


def add_cinic10(mldb):
    """Add cinic10 (imagenet only subset) datasets to the database."""
    found_datasets = [dset.name for dset in mldb.get_datasets()]
    if all(dset in found_datasets for dset in 
        ["cinic10-imgnt-only-train", "cinic10-imgnt-only-valid", "cinic10-imgnt-only-test", "cinic10-imgnt-only-all"]):
        print("cinic10-imgnt-only already loaded into DB. Skipping...")
        return

    # NOTE: prepare directory with https://github.com/BayesWatch/cinic-10/blob/master/notebooks/imagenet-extraction.ipynb
    # TODO: fix this so it doesn't rely on local state
    trainset = torchvision.datasets.ImageFolder('anon/cinic10/cinic10-imagenet/train', transform=transforms.ToTensor())
    validset = torchvision.datasets.ImageFolder('anon/cinic10/cinic10-imagenet/valid', transform=transforms.ToTensor())
    testset = torchvision.datasets.ImageFolder('anon/cinic10/cinic10-imagenet/test', transform=transforms.ToTensor())

    trainloader = torch.utils.data.DataLoader(trainset, batch_size=999999, shuffle=False, num_workers=20)
    valloader = torch.utils.data.DataLoader(validset, batch_size=999999, shuffle=False, num_workers=20)
    testloader = torch.utils.data.DataLoader(testset, batch_size=999999, shuffle=False, num_workers=20)

    train_images, train_labels = iter(trainloader).next()
    valid_images, valid_labels = iter(valloader).next()
    test_images, test_labels = iter(testloader).next()

    train_images = (train_images.permute(0, 2, 3, 1)*255).numpy().astype(np.uint8)
    valid_images = (valid_images.permute(0, 2, 3, 1)*255).numpy().astype(np.uint8)
    test_images = (test_images.permute(0, 2, 3, 1)*255).numpy().astype(np.uint8)

    train_labels = train_labels.numpy().astype(np.int64)
    valid_labels = valid_labels.numpy().astype(np.int64)
    test_labels = test_labels.numpy().astype(np.int64)

    all_images = np.concatenate([train_images, valid_images, test_images])
    all_labels = np.concatenate([train_labels, valid_labels, test_labels])

    if "cinic10-imgnt-only-train" not in found_datasets:
        mldb.dataset(name="cinic10-imgnt-only-train", examples=train_images, labels=train_labels).store()
    if "cinic10-imgnt-only-valid" not in found_datasets:
        mldb.dataset(name="cinic10-imgnt-only-valid", examples=valid_images, labels=valid_labels).store()
    if "cinic10-imgnt-only-test" not in found_datasets:
        mldb.dataset(name="cinic10-imgnt-only-test", examples=test_images, labels=test_labels).store()
    if "cinic10-imgnt-only-all" not in found_datasets:
        mldb.dataset(name="cinic10-imgnt-only-all", examples=all_images, labels=all_labels).store()


def add_rst_models(mldb):
    model_name = 'rst_wideresnet_baseline_model'
    found = False
    try:
        model = mldb.model(name=model_name)
        found = True
    except:
        # TODO: introduce a "NotFound" exception and then actually check for it
        pass
    if found:
        assert model.model_family == 'rst_wideresnet'
        return
    ssl_context = ssl.SSLContext()
    with urllib.request.urlopen(
            'anon',
            context=ssl_context) as response:
       model_data = response.read()
    lr = mldb.learning_rule(family='rst_wideresnet',
                            params={'adversarial_training': False, 'normalize_input': True})
    mldb.import_model(learning_rule=lr,
                      model_bytes=model_data,
                      name=model_name)


def add_individual_odst_model(mldb, model_name):
    full_model_name = 'odst_' + model_name + '_model'
    found = False
    try:
        model = mldb.model(name=full_model_name)
        found = True
    except:
        # TODO: introduce a "NotFound" exception and then actually check for it
        pass
    if found:
        assert model.model_family == 'odst_' + model_name
        return
    ssl_context = ssl.SSLContext()
    with urllib.request.urlopen(
            f'anon',
            context=ssl_context) as response:
       model_data = response.read()
    lr = mldb.learning_rule(family=f'odst_{model_name}',
                            params={'num_odst_iterations': 3})
    mldb.import_model(learning_rule=lr,
                      model_bytes=model_data,
                      name=full_model_name)


def add_odst_models(mldb):
    for x in ['resnet50', 'shakedroppyramid272']:
        add_individual_odst_model(mldb, x)


def main():
    """Initialize the project database."""
    default_mldb = get_default_mldb()

    add_cifar10(default_mldb)
    add_cifar10p1(default_mldb)
    add_cifar10p2(default_mldb)
    add_cifar10c(default_mldb)
    add_wilds(default_mldb)
    add_stl10(default_mldb)
    add_cinic10(default_mldb)

    # Subsample cifar10 training splits
    cifar10_train = default_mldb.dataset(name="cifar10-train")
    for idx in range(5):
        subsampled = cifar10_train.subsample(size=10000, seed=idx, name=f"cifar10-train_10k_{idx}")
        subsampled.store(assert_not_stored_yet=False)

if __name__ == "__main__":
    main()
