from abc import ABC, abstractmethod
import dataclasses

from . import dataset as dataset_module
from . import model
from .model_families import default_registry


class LearningRuleImplementation(ABC):
    # train_examples is one example per row
    @abstractmethod
    def learn(self, dataset, checkpointer, logger, params):
        pass


class SimpleFunctionLearningRule(LearningRuleImplementation):
    def __init__(self, learning_rule):
        self.learning_rule_fn = learning_rule
    
    def learn(self, dataset, checkpointer, logger, params):
        return self.learning_rule_fn(dataset,
                                     checkpointer=checkpointer,
                                     logger=logger,
                                     params=params)

class Checkpointer():
    def __init__(self, model):
        self.model = model
    
    def add(self, sequence_number, model_bytes, extra_info=None):
        if extra_info is None:
            extra_info = {}
        assert isinstance(sequence_number, (int, float))
        assert isinstance(model_bytes, bytes)
        assert isinstance(extra_info, dict)
        self.model.add_checkpoint(sequence_number, model_bytes, extra_info)
        

class Logger():
    def __init__(self, model):
        self.model = model
    
    def add(self, key, entry_data, extra_info=None):
        if extra_info is None:
            extra_info = {}
        assert isinstance(key, str)
        assert isinstance(extra_info, dict)
        self.model.train_log.add_entry(key, entry_data, extra_info)


class LearningRule():
    def __init__(self,
                 *,
                 family=None,
                 params={},
                 db_obj=None,
                 implementation=None,
                 mldb=None,
                 name=None,
                 **kwargs):
        self.mldb = mldb
        if mldb is not None:
            self.db = mldb.db
            self.registry = mldb.model_registry
        else:
            self.db = None
            self.registry = default_registry()

        assert len(kwargs) == 0 or len(params) == 0
        if len(kwargs) > 0:
            params = kwargs

        self.name = name
        # First init option: specify a model family name that is registered
        if db_obj is not None:
            assert all([x is None for x in [family, implementation, name]])
            assert params == {} and len(kwargs) == 0
            self.db_obj = db_obj
            self._init_from_self_db_obj()
        elif family is not None:
            assert implementation is None
            self.model_family = family
            cur_family = self.registry[self.model_family]
            # Convert params to the dataclass of the model family if the user passed in a dict
            if isinstance(params, dict):
                self.params = self.registry[self.model_family].rule_params(**params)
            else:
                assert isinstance(params, self.registry[self.model_family].rule_params)
                self.params = params
            self.impl = cur_family.learning_rule(self.params)
            # Find a matching learning rule in the db if there is one
            if self.db is not None:
                candidates = self.db.get(type_='learning_rule', subtype=self.model_family)
                matching_candidate = None
                for cand in candidates:
                    cur_params_dataclass = cur_family.rule_params(**cand.params)
                    if cur_params_dataclass == self.params:
                        matching_candidate = cand
                        if self.name is not None and self.name != matching_candidate.name:
                            raise ValueError(f'Found a matching learning rule with name "{matching_candidate.name}" instead of "{self.name}".')
                        break
                self.db_obj = matching_candidate
                if self.db_obj is None and name is not None:
                    if self.db.get(type_='learning_rule', name=name) is not None:
                        raise ValueError(f'A different learning rule with name "{name}" already exists.')
            else:
                self.db_obj = None
        else:
            if implementation is None:
                assert name is not None and self.db is not None and params == {} and len(kwargs) == 0
                self.db_obj = self.db.get(type_='learning_rule', name=name, assert_exists=True)
                self._init_from_self_db_obj()
            else:
                assert name is None and params == {} and len(kwargs) == 0
                self.model_family = None
                self.db_obj = None
                self.impl = implementation
                self.params = None

    def _init_from_self_db_obj(self):
        self.model_family = self.db_obj.subtype
        self.name = self.db_obj.name
        cur_family = self.registry[self.model_family]
        self.params = cur_family.rule_params(**self.db_obj.params)
        self.impl = cur_family.learning_rule(self.params)

    def store(self, assert_not_stored_yet=False, tag=None):
        assert self.db is not None
        if self.is_stored:
            if assert_not_stored_yet:
                raise ValueError
            return
        assert self.model_family is not None
        assert self.params is not None

        # Check that another learning rule with the same parameters
        # hasn't already been created and stored.
        matching_rule = None
        cur_family = self.registry[self.model_family]
        for rule in self.db.get(type_='learning_rule', subtype=self.model_family):
            if self.params == cur_family.rule_params(**rule.params):
                matching_rule = rule
                break

        if matching_rule is not None:
            self.db_obj = rule
        else:
            self.db_obj = self.db.insert(
                name=self.name,
                type_='learning_rule',
                subtype=self.model_family,
                extra = {
                    'params': dataclasses.asdict(self.params),
                })
        if tag is not None:
            self.mldb.add_tag(self, tag)

        return self
    
    @property
    def is_stored(self):
        return self.db_obj is not None
    
    @property
    def uuid(self):
        if self.db_obj is not None:
            return self.db_obj.uuid
        else:
            return None

    def learn(self, *,
              dataset=None,
              examples=None,
              labels=None,
              name=None,
              store=False,
              params={},
              **kwargs):
        assert len(kwargs) == 0 or len(params) == 0
        if len(kwargs) > 0:
            params = kwargs
        if isinstance(params, dict):
            params = self.registry[self.model_family].train_params(**params)
        else:
            assert isinstance(params, self.registry[self.model_family].train_params)

        if isinstance(dataset, str):
            assert self.db is not None
            assert examples is None and labels is None
            return self.learn(dataset=dataset_module.Dataset(name=dataset, mldb=self.mldb),
                              name=name,
                              store=store,
                              params=params)
        if not isinstance(dataset, dataset_module.Dataset):
            assert examples is not None and labels is not None
            dataset = dataset_module.Dataset(name=None,
                                             examples=examples,
                                             labels=labels,
                                             mldb=self.mldb)
        cur_model = model.Model(implementation=None,
                                train_completed=False,
                                learning_rule=self,
                                train_set=dataset,
                                name=name,
                                params=params,
                                mldb=self.mldb)
        if store:
            cur_model.store()
        checkpointer = Checkpointer(cur_model)
        logger = Logger(cur_model)
        model_impl = self.impl.learn(dataset,
                                     checkpointer=checkpointer,
                                     logger=logger,
                                     params=params)
        cur_model.complete_training(model_impl)
        return cur_model

    def get_trained_models(self, only_completed=True):
        if not self.is_stored:
            return []
        models = self.db.get(rel_first=self.db_obj, rel_type="trained", type_="model")
        models = [model.Model(db_obj=model_obj, mldb=self.mldb) for model_obj in models]
        if only_completed:
            models = [m for m in models if m.train_completed == True]
        return models
