import basicdb
import os
import pathlib

from . import dataset
from . import data_families
from . import evaluation
from . import featurization
from . import learning_rule
from . import model
from . import model_families


class MLDB:
    def __init__(self, db, data_cache_dir, data_registry=None, model_registry=None):
        self.db = db
        self.model_registry = model_registry if model_registry is not None else model_families.default_registry()
        self.data_registry = data_registry if data_registry is not None else data_families.default_registry()
        os.makedirs(data_cache_dir, exist_ok=True)
        self.data_cache_dir = data_cache_dir

    def dataset(self, *args, **kwargs):
        return dataset.Dataset(*args, **kwargs, mldb=self)
    
    def evaluation(self, *args, **kwargs):
        return evaluation.Evaluation(*args, **kwargs, mldb=self)
    
    def featurization(self, *args, **kwargs):
        return featurization.Featurization(*args, **kwargs, mldb=self)
    
    def learning_rule(self, *args, **kwargs):
        return learning_rule.LearningRule(*args, **kwargs, mldb=self)
    
    def model(self, *args, **kwargs):
        return model.Model(*args, **kwargs, mldb=self)

    def get_datasets(self):
        return [self.dataset(db_obj=x) for x in self.db.get(type_="dataset")]
    
    def get_evaluations(self, tag=None, test_set=None):
        if tag is not None:
            assert test_set is None
            tag = self._get_tag(tag)
            if tag is None: return []
            return [self.evaluation(db_obj=x) for x in self.db.get(rel_first=tag, type_="evaluation")]
        elif test_set is not None:
            dset_obj = self.dataset(name=test_set)
            return [self.evaluation(db_obj=x) for x in self.db.get(rel_second=dset_obj.db_obj,
                                                                   rel_type="test_set",
                                                                   type_="evaluation")]
        return [self.evaluation(db_obj=x) for x in self.db.get(type_="evaluation")]
    
    def get_featurizations(self, tag=None):
        if tag is not None:
            tag = self._get_tag(tag)
            if tag is None: return []
        return [self.featurization(db_obj=x) for x in self.db.get(rel_first=tag, type_="featurization")]

    def get_learning_rules(self, tag=None):
        if tag is not None:
            tag = self._get_tag(tag)
            if tag is None: return []
        return [self.learning_rule(db_obj=x) for x in self.db.get(rel_first=tag, type_='learning_rule')]
    
    def get_models(self, train_set=None, family=None, tag=None, only_completed=True, include_hidden=False):
        if tag is not None:
            tag = self._get_tag(tag)
            if tag is None: return []
        
        uuids = None
        if train_set:
            assert train_set in [x.name for x in self.get_datasets()]
            train_set = self.dataset(name=train_set)
            train_models = self.db.get(rel_first=train_set.db_obj, subtype=family, rel_type="train_set", type_="model")
            uuids = [x.uuid for x in train_models]
        tag_models = self.db.get(rel_first=tag, uuids=uuids, subtype=family, type_="model", include_hidden=include_hidden)
        models = [self.model(db_obj=x) for x in tag_models]
        if only_completed:
            models = [m for m in models if m.train_completed == True]
        return models

    def add_tag(self, obj, tag_name):
        if not obj.is_stored:
            raise ValueError("Object not stored in DB!")
        tag = self.db.get(type_="tag", name=tag_name)
        if tag is None:
            tag = self.db.insert(type_='tag', name=tag_name)

        existing_rel = self.db.get_relationship(first=tag, second=obj.db_obj, type_="tag")
        if existing_rel is None:
            self.db.insert_relationship(first=tag, second=obj.db_obj, type_="tag")

    def _get_tag(self, tag_name):
        # this returns None if tag_name is not found
        return self.db.get(type_='tag', name=tag_name)

    def get_tags(self):
        return [x.name for x in self.db.get(type_="tag")]
    
    def import_model(self,
                     learning_rule,
                     model_bytes,
                     train_set=None,
                     name=None,
                     params=None):
        dummy_impl = model.OnlySerializableModelImplementation(model_bytes)
        tmp_model = model.Model(name=name,
                                implementation=dummy_impl,
                                learning_rule=learning_rule,
                                train_set=train_set,
                                params=params,
                                imported=True,
                                train_completed=True,
                                mldb=self)
        tmp_model.store()
        


SQL_STRING_LIVE = "anon"
S3_BUCKET_LIVE = "anon"

CACHE_ROOT = pathlib.Path(__file__) / '../../shared_mldb_root'

def get_default_mldb():
    return MLDB(
        db=basicdb.BasicDB(
            sql_string=SQL_STRING_LIVE,
            stash_s3_bucket=S3_BUCKET_LIVE,
            stash_kwargs={
                "endpoint_url": "anon",
                "access_key": "cifar10p2",
                "secret_key": "anon",
                "region_name": "us-east-1",
                "signature_version": "s3v4",
                "ssl_certificate": (pathlib.Path(__file__) / "../../configs/chain.cer").resolve(),
                "cache_on_local_disk": True,
                "cache_root_path": (CACHE_ROOT / 's3_cache').resolve(),
            },
            namespace="",
            max_extra_fields_size=8192,
        ),
        data_cache_dir=(CACHE_ROOT / 'data_cache').resolve(),
    )


def get_local_mldb(path):
    path = pathlib.Path(path)
    db_path = path / 'db.sqlite'
    return MLDB(
        db=basicdb.BasicDB(
            sql_string=f'sqlite:///{db_path}',
            db_kwargs={'create_db': True},
            stash_rootdir=(path / 'stash').resolve(),
            namespace="",
            max_extra_fields_size=8192,
        ),
        data_cache_dir=(path / 'data_cache').resolve(),
    )


# TODO: make mldbi readonly in this case (implement this at the basicdb level)
def get_semilocal_mldb(db_file):
    db_file = pathlib.Path(db_file).resolve()
    return MLDB(
        db=basicdb.BasicDB(
            sql_string=f'sqlite:///{db_file}',
            db_kwargs={'create_db': False},
            stash_s3_bucket=S3_BUCKET_LIVE,
            stash_kwargs={
                "endpoint_url": "anon",
                "access_key": "cifar10p2",
                "secret_key": "anon",
                "region_name": "us-east-1",
                "signature_version": "s3v4",
                "ssl_certificate": (pathlib.Path(__file__) / "../../configs/chain.cer").resolve(),
                "cache_on_local_disk": True,
                "cache_root_path": (CACHE_ROOT / 's3_cache').resolve(),
            },
            namespace="",
            max_extra_fields_size=8192,
        ),
        data_cache_dir=(CACHE_ROOT / 'data_cache').resolve(),
    )
