from abc import ABC, abstractmethod
import dataclasses
import gc

from . import dataset as dataset_module
from . import evaluation
from . import featurization
from . import learning_rule
from .model_families import default_registry


class ModelImplementation(ABC):
    # examples is one example per row
    @abstractmethod
    def predict(self, dataset, params):
        pass

    def featurize(self, dataset, params):
        raise NotImplementedError

    @abstractmethod
    def serialize(self):
        pass


class OnlySerializableModelImplementation(ModelImplementation):
    def __init__(self, model_bytes):
        self.model_bytes = model_bytes

    def predict(self, dataset, params):
        raise NotImplementedError

    def serialize(self):
        return self.model_bytes


class TrainLog():
    def __init__(self,
                 *,
                 parent_model=None,
                 db_obj=None,
                 mldb=None):
        self.mldb = mldb
        if mldb is not None:
            self.db = mldb.db
        else:
            self.db = None
        self.db_obj = None
        self._parent_model = parent_model
        self._added_log_entries = {}
        self._added_log_data = {}
        self._log_entries = None 
        if self.db is not None and self.db_obj is None and parent_model.is_stored:
            res = self.db.get(rel_first=parent_model.db_obj, rel_type='train_log', type_='train_log')
            assert len(res) <= 1
            if len(res) == 1:
                self.db_obj = res[0]
        
    @property
    def is_stored(self):
        return self.db_obj is not None
    
    @property
    def parent_model(self):
        if self._parent_model is None:
            assert self.db is not None
            res = self.db.get(rel_second=self.db_obj, rel_type='train_log', type_='model')
            assert len(res) == 1
            self._parent_model = Model(db_obj=res[0], mldb=self.mldb)
        return self._parent_model
    
    def add_entry(self, key, entry_data, extra_info):
        if self.is_stored:
            if key in self.entries:
                self.db.update_blob(self.db_obj, key, data=entry_data, extra={'extra_info': extra_info})
            else:
                self.db.insert_blob(self.db_obj, key, data=entry_data, extra={'extra_info': extra_info})
            self._log_entries[key] = extra_info
        else:
            self._added_log_entries[key] = extra_info
            self._added_log_data[key] = entry_data

    @property
    def entries(self):
        if not self.is_stored:
            assert self._log_entries is None
            return self._added_log_entries
        else:
            if self._log_entries is None:
                db_res = self.db.get_blobs(self.db_obj)
                self._log_entries = {k: v.extra['extra_info'] for k, v in db_res.items()}
            for k in self._added_log_entries.keys():
                assert k not in self._log_entries
            return {**self._log_entries, **self._added_log_entries}
    
    def load_data(self, keys):
        if isinstance(keys, str):
            single_key = True
            keys = [keys]
        else:
            single_key = False
        res_local = {}
        db_keys = []
        for k in keys:
            assert k in self.entries
            if k in self._added_log_data:
                res_local[k] = self._added_log_data[k]
            else:
                db_keys.append(k)
        if self.is_stored:
            res_db = self.db.load_blobs(self.db_obj, db_keys)
        else:
            res_db = {}
        tmp_res = {**res_local, **res_db}
        if single_key:
            return tmp_res[keys[0]]
        else:
            return tmp_res
    
    def store(self, assert_not_stored_yet=False):
        assert self.db is not None
        if self.is_stored:
            if assert_not_stored_yet:
                raise ValueError
            return
        self.parent_model.store()
        self.db_obj = self.db.insert(
            type_='train_log',
            subtype=self.parent_model.model_family
        )
        self.db.insert_relationship(first=self.parent_model.db_obj,
                                    second=self.db_obj,
                                    type_='train_log')
        assert set(self._added_log_data.keys()) == set(self._added_log_entries.keys())
        for k in self._added_log_data.keys():
            self.db.insert_blob(self.db_obj,
                                k,
                                data=self._added_log_data[k],
                                extra={'extra_info': self._added_log_entries[k]})
        if self._log_entries is not None:
            for k, v in self._added_log_entries.items():
                assert k not in self._log_entries
                self._log_entries[k] = v
        self._added_log_entries = {}
        self._added_log_data = {}

        
class Checkpoint():
    def __init__(self,
                 *,
                 parent_model=None,
                 sequence_number=None,
                 model_bytes=None,
                 extra_info={},
                 db_obj=None,
                 mldb=None):
        self.mldb = mldb
        if mldb is not None:
            self.db = mldb.db
            self.registry = mldb.model_registry
        else:
            self.db = None
            self.registry = default_registry()
        self._model_bytes = None
        if db_obj is not None:
            assert mldb is not None
            assert all([x is None for x in [parent_model, sequence_number, model_bytes]])
            assert extra_info == {}
            self.db_obj = db_obj
            self._parent_model = None
        else:
            assert all([x is not None for x in [model_bytes, extra_info, sequence_number, parent_model]])
            self._sequence_number = sequence_number
            self._model_bytes = model_bytes
            self._extra_info = extra_info
            self._parent_model = parent_model
            self.db_obj = None

    @property
    def sequence_number(self):
        if self.db_obj is not None:
            return self.db_obj.sequence_number
        else:
            return self._sequence_number
    
    @property
    def extra_info(self):
        if self.db_obj is not None:
            return self.db_obj.extra_info
        else:
            return self._extra_info
    
    @property
    def model_bytes(self):
        if self._model_bytes is None:
            assert self.db_obj is not None
            self._model_bytes = self.db.load_blob(self.db_obj, 'serialized_model')
        return self._model_bytes
    
    def release_model_bytes(self):
        assert self.db_obj is not None
        gc.collect()
        del self._model_bytes
        self._model_bytes = None
    
    @property
    def parent_model(self):
        if self._parent_model is None:
            assert self.db is not None
            res = self.db.get(type_='model',
                              rel_second=self.db_obj,
                              rel_type='model_checkpoint')
            assert len(res) == 1
            self._parent_model = Model(db_obj=res[0], mldb=self.mldb)
        return self._parent_model
    
    def as_model(self):
        if self.db is not None and self.db_obj is not None:
            res = self.db.get(type_='model',
                              rel_first=self.db_obj,
                              rel_type='checkpoint_model')
            assert len(res) <= 1
            if len(res) == 1:
                return Model(db_obj=res[0], parent_checkpoint=self, mldb=self.mldb)
        return Model(learning_rule=self.parent_model.learning_rule,
                     train_set=self.parent_model.train_set,
                     params=self.parent_model.params,
                     parent_checkpoint=self,
                     mldb=self.mldb)
    
    @property
    def is_stored(self):
        return self.db_obj is not None
    
    @property
    def uuid(self):
        if self.db_obj is not None:
            return self.db_obj.uuid
        else:
            return None

    def store(self, assert_not_stored_yet=False):
        assert self.db is not None
        if self.is_stored:
            if assert_not_stored_yet:
                raise ValueError
            return
        self.parent_model.store(skip_checkpoint=self)
        self.db_obj = self.db.insert(
            type_= 'checkpoint',
            subtype=self.parent_model.model_family,
            extra={
                'extra_info': self.extra_info,
                'sequence_number': self.sequence_number,
            },
            blobs={
                'serialized_model': self.model_bytes
            }
        )
        self.db.insert_relationship(first=self.parent_model.db_obj,
                                    second=self.db_obj,
                                    type_='model_checkpoint')


# TODO: overload __eq__ for this and other classes
class Model():
    def __init__(self,
                 *,
                 name=None,
                 implementation=None,
                 learning_rule=None,
                 train_set=None,
                 db_obj=None,
                 params=None,
                 train_completed=None,
                 parent_checkpoint=None,
                 imported=None,
                 mldb=None):
        self.mldb = mldb
        if mldb is not None:
            self.db = mldb.db
            self.registry = mldb.model_registry
        else:
            self.db = None
            self.registry = default_registry()
        self._added_checkpoints = []
        self._checkpoints = None
        self._parent_checkpoint = None
        self._train_log = None
        # Build from db_obj
        if db_obj is not None:
            assert self.db is not None
            assert all([x is None for x in [implementation, name, learning_rule, train_set, train_completed, params, imported]])
            self.db_obj = db_obj
            self._parent_checkpoint = parent_checkpoint
            self._init_from_db_obj()
        # Load via name
        elif implementation is None and train_completed is None and parent_checkpoint is None:
                assert all([x is not None for x in [name, mldb]])
                assert all([x is None for x in [learning_rule, train_set, train_completed, parent_checkpoint, params, imported]])
                self.db_obj = self.db.get(type_='model', name=name, assert_exists=True)
                self._init_from_db_obj()
        else:
            assert learning_rule is not None
            if implementation is None:
                assert imported is None
                assert parent_checkpoint is not None or (train_completed is not None and not train_completed)
            else:
                assert train_completed is None or train_completed
            if train_set is None:
                assert imported
            self.name = name
            self._impl = implementation
            self._train_set = train_set
            self._has_train_set = train_set is not None
            self._learning_rule = learning_rule
            self.db_obj = None
            if params is None:
                params = {}
            if isinstance(params, dict):
                params = self.registry[self.model_family].train_params(**params)
            else:
                assert isinstance(params, self.registry[self.model_family].train_params)
            self.params = params
            self.is_checkpoint = parent_checkpoint is not None
            self.train_completed = self.is_checkpoint or implementation is not None
            self.imported = imported if imported is not None else False
            self._parent_checkpoint = parent_checkpoint
    
    def _init_from_db_obj(self):
        self.name = self.db_obj.name
        self.params = self.registry[self.model_family].train_params(**self.db_obj.params)
        self.train_completed = self.db_obj.train_completed
        self.imported = self.db_obj.imported if 'imported' in self.db_obj.extra else False
        self._impl = None
        self._learning_rule = None
        self._has_train_set = None
        self._train_set = None
        self.is_checkpoint = self.db_obj.is_checkpoint
        if self._parent_checkpoint is not None:
            assert self.is_checkpoint
    
    @property
    def impl(self):
        if self.train_completed:
            if self._impl is None:
                if self.is_checkpoint:
                    # TODO: bypass the model_bytes field in the parent to avoid extra references that keep the raw
                    # bytes in memory?
                    self._impl = self.registry[self.model_family].deserialize_model(self.parent_checkpoint.model_bytes)
                else:
                    assert self.is_stored
                    self._impl = self.registry[self.model_family].deserialize_model(
                            self.db.load_blob(self.db_obj, 'serialized_model'))
            return self._impl
        else:
            return None
    
    @property
    def train_log(self):
        if self._train_log is None:
            self._train_log = TrainLog(parent_model=self, mldb=self.mldb)
        return self._train_log
    
    def complete_training(self, implementation):
        assert not self.train_completed
        self.train_completed = True
        self._impl = implementation
        if self.db_obj is not None:
            self.db_obj.train_completed = True
            self.db.update(self.db_obj)
            self.db.insert_blob(self.db_obj, name='serialized_model', data=self.impl.serialize())
    
    def add_checkpoint(self, sequence_number, model_bytes, extra_info):
        assert not self.train_completed and not self.is_checkpoint
        new_checkpoint = Checkpoint(parent_model=self,
                                    sequence_number=sequence_number,
                                    model_bytes=model_bytes,
                                    extra_info=extra_info,
                                    mldb=self.mldb)
        if self.db_obj is not None:
            new_checkpoint.store()
            new_checkpoint.release_model_bytes()
            if self._checkpoints is not None:
                self._checkpoints.append(new_checkpoint)
        else:
            self._added_checkpoints.append(new_checkpoint)
        return new_checkpoint
    
    @property
    def checkpoints(self):
        if self.is_checkpoint:
            return []
        if self.db_obj is None:
            res = [x for x in self._added_checkpoints]
        else:
            if self._checkpoints is None:
                db_res = self.db.get(type_='checkpoint',
                                     rel_first=self.db_obj,
                                     rel_type='model_checkpoint')
                self._checkpoints = [Checkpoint(db_obj=x, mldb=self.mldb) for x in db_res]
            res = [x for x in self._checkpoints]
            res_uuids = set([x.uuid for x in res])
            # Someone could have stored one of the checkpoints in _added_checkpoints independently of this
            # model object, in which case the checkpoint would appear twice. So we deduplicate here.
            for x in self._added_checkpoints:
                if x.is_stored:
                    if x.uuid not in res_uuids:
                        res.append(x)
                else:
                    res.append(x)
        return sorted(res, key=lambda x: x.sequence_number)

    def predict(self, *,
                dataset=None,
                examples=None,
                labels=None,
                return_eval=True,
                params={},
                **kwargs):
        assert len(kwargs) == 0 or len(params) == 0
        if len(kwargs) > 0:
            params = kwargs
        if isinstance(params, dict):
            params = self.registry[self.model_family].eval_params(**params)
        else:
            assert isinstance(params, self.registry[self.model_family].eval_params)
        assert self.train_completed
        if isinstance(dataset, str):
            assert self.db is not None
            assert examples is None and labels is None
            return self.predict(dataset=dataset_module.Dataset(name=dataset, mldb=self.mldb),
                                return_eval=return_eval,
                                params=params)
        if not isinstance(dataset, dataset_module.Dataset):
            assert examples is not None
            dataset = dataset_module.Dataset(name=None,
                                             examples=examples,
                                             labels=labels,
                                             mldb=self.mldb)
        predictions, extra_prediction_data = self.impl.predict(dataset=dataset,
                                                               params=params)
        if return_eval:
            return evaluation.Evaluation(model=self,
                                         test_set=dataset,
                                         params=params,
                                         predictions=predictions,
                                         extra_prediction_data=extra_prediction_data,
                                         mldb=self.mldb)
        else:
            return predictions
    
    def featurize(self,
                  dataset=None,
                  examples=None,
                  params={},
                  **kwargs):
        assert len(kwargs) == 0 or len(params) == 0
        if len(kwargs) > 0:
            params = kwargs
        if isinstance(params, dict):
            params = self.registry[self.model_family].eval_params(**params)
        else:
            assert isinstance(params, self.registry[self.model_family].eval_params)
        assert self.train_completed
        if isinstance(dataset, str):
            assert self.db is not None
            assert examples is None
            return self.featurize(dataset=dataset_module.Dataset(name=dataset, mldb=self.mldb),
                                  params=params)
        if not isinstance(dataset, dataset_module.Dataset):
            assert examples is not None
            dataset = dataset_module.Dataset(name=None,
                                             examples=examples,
                                             mldb=self.mldb)
        features, extra_featurization_data = self.impl.featurize(dataset=dataset,
                                                                 params=params)
        return featurization.Featurization(model=self,
                                           dataset=dataset,
                                           params=params,
                                           features=features,
                                           extra_featurization_data=extra_featurization_data,
                                           mldb=self.mldb)
    
    @property
    def learning_rule(self):
        if self._learning_rule is None:
            res = self.db.get(rel_second=self.db_obj, rel_type='trained', type_='learning_rule')
            assert len(res) == 1
            self._learning_rule = learning_rule.LearningRule(db_obj=res[0], mldb=self.mldb)
        return self._learning_rule
    
    @property
    def has_train_set(self):
        if self._has_train_set is None:
            res = self.db.get(rel_second=self.db_obj, rel_type='train_set', type_='dataset')
            if len(res) == 0:
                self._has_train_set = False
                assert self.imported
            else:
                assert len(res) == 1
                self._has_train_set = True
                self._train_set = dataset_module.Dataset(db_obj=res[0], mldb=self.mldb)
        return self._has_train_set
    
    @property
    def train_set(self):
        if not self.has_train_set:
            return None
        if self._train_set is None:
            res = self.db.get(rel_second=self.db_obj, rel_type='train_set', type_='dataset')
            assert len(res) == 1
            self._train_set = dataset_module.Dataset(db_obj=res[0], mldb=self.mldb)
        return self._train_set
    
    @property
    def parent_checkpoint(self):
        if not self.is_checkpoint:
            return None
        if self._parent_checkpoint is None:
            res = self.db.get(rel_second=self.db_obj, rel_type='checkpoint_model', type_='checkpoint')
            assert len(res) == 1
            self._parent_checkpoint = Checkpoint(db_obj=res[0], mldb=self.mldb)
        return self._parent_checkpoint

    @property
    def is_stored(self):
        return self.db_obj is not None
    
    @property
    def uuid(self):
        if self.db_obj is not None:
            return self.db_obj.uuid
        else:
            return None
    
    def rename(self, new_name):
        if self.is_stored:
            self.db.update(self.db_obj, new_name=new_name)
        self.name = new_name
    
    @property
    def model_family(self):
        if self.db_obj is not None:
            return self.db_obj.subtype
        else:
            return self.learning_rule.model_family

    def store(self, assert_not_stored_yet=False, skip_checkpoint=None, tag=None):
        assert self.db is not None
        if self.is_stored:
            if assert_not_stored_yet:
                raise ValueError
            return
        # If the model is a checkpoint model, make sure no other model for this
        # checkpoint is already in the db
        if self.is_checkpoint and self.parent_checkpoint.is_stored:
            res = self.db.get(type_='model',
                              rel_first=self.parent_checkpoint.db_obj,
                              rel_type='checkpoint_model')
            assert len(res) == 0
        cur_blobs = {}
        if not self.is_checkpoint and self.impl is not None:
            cur_blobs = {
                'serialized_model': self.impl.serialize()
            }
        self.db_obj = self.db.insert(
            name=self.name,
            type_= 'model',
            subtype=self.model_family,
            extra={
                'params': dataclasses.asdict(self.params),
                'is_checkpoint': self.is_checkpoint,
                'train_completed': self.train_completed,
                'imported': self.imported
            },
            blobs=cur_blobs
        )
        self.name = self.db_obj.name    # in case name is None, basicdb assigns the uuid
        self.learning_rule.store()
        self.db.insert_relationship(first=self.learning_rule.db_obj,
                                    second=self.db_obj,
                                    type_='trained')
        if self.has_train_set:
            self.train_set.store()
            self.db.insert_relationship(first=self.train_set.db_obj,
                                        second=self.db_obj,
                                        type_='train_set')
        if self.is_checkpoint:
            self.parent_checkpoint.store()
            self.db.insert_relationship(first=self.parent_checkpoint.db_obj,
                                        second=self.db_obj,
                                        type_='checkpoint_model')
        else:
            for checkpoint in self._added_checkpoints:
                if checkpoint != skip_checkpoint:
                    checkpoint.store()
            if self._checkpoints is not None:
                self._checkpoints.extend(self._added_checkpoints)
            self._added_checkpoints = []
        self.train_log.store()

        if tag is not None:
            self.mldb.add_tag(self, tag)

    def get_evaluations(self):
        '''Return the evaluations associated with the model.'''
        if not self.is_stored:
            return []
        evals = self.db.get(rel_second=self.db_obj, rel_type="model_evaluated", type_="evaluation")
        return [evaluation.Evaluation(db_obj=ev, mldb=self.mldb) for ev in evals]
