import dataclasses
import pickle
from tqdm import tqdm

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
import clip

from .. import model
from ..model_family import ModelFamily
from ..learning_rule import LearningRuleImplementation
from .imagenet_models import ImageDataset


@dataclasses.dataclass
class ClipFinetuneFamilyRuleParams:
    num_classes: int

@dataclasses.dataclass
class ClipFinetuneFamilyTrainParams:
    device: str = 'auto' # 'cuda:0' if cuda available else 'cpu'
    num_workers: int = 2
    eval_batch_size: int = 128

@dataclasses.dataclass
class ClipFinetuneFamilyEvalParams:
    device: str = 'auto' # 'cuda:0' if cuda available else 'cpu'
    num_workers: int = 2
    eval_batch_size: int = 128


class ClipFinetuneLearningRule(LearningRuleImplementation):
    def __init__(self, family, params):
        self.family = family
        self.hparams = params
    
    def learn(self, dataset, checkpointer, logger, params):
        torch.backends.cudnn.benchmark = True
        device = params.device
        if device == 'auto':
            device = 'cuda:0' if torch.cuda.is_available() else 'cpu'

        model, preprocess = clip.load(self.family)

        trainset = ImageDataset(dataset, transform=preprocess)
        trainloader = DataLoader(trainset, batch_size=params.eval_batch_size, shuffle=False,
                                num_workers=params.num_workers, pin_memory=True)

        model = model.to(device).eval()
        with torch.no_grad():
            image_features, labels = [], []
            for inputs, targets in tqdm(trainloader):
                inputs = inputs.to(device)
                image_features_batch = model.encode_image(inputs).float()
                image_features.append(image_features_batch)
                labels.append(targets.to(device))
            image_features = torch.cat(image_features)
            labels = torch.cat(labels)
        
        labels_hot = torch.eye(self.hparams.num_classes).to(device)[labels]
        features_biased = torch.cat((image_features, torch.ones(image_features.size(0), 1).to(device)), dim=1)
        weights, _ = torch.lstsq(labels_hot, features_biased)
        weights = weights[:features_biased.size(1)]

        linear = nn.Linear(image_features.size(1), self.hparams.num_classes, bias=True).to(device)
        linear.weight[:] = weights[:-1].t()
        linear.bias[:] = weights[-1]
        
        return ClipFinetune(model, linear, preprocess, self.hparams)


class ClipFinetune(model.ModelImplementation):
    def __init__(self, model, linear, preprocess, hparams):
        self.model = model
        self.linear = linear
        self.preprocess = preprocess
        self.hparams =  hparams
    
    def predict(self, dataset, params):
        torch.backends.cudnn.benchmark = True
        device = params.device
        if device == 'auto':
            device = 'cuda:0' if torch.cuda.is_available() else 'cpu'

        testset = ImageDataset(dataset, transform=self.preprocess)
        testloader = DataLoader(testset, batch_size=params.eval_batch_size, shuffle=False,
                                num_workers=params.num_workers, pin_memory=True)

        self.model = self.model.to(device).eval()
        with torch.no_grad():
            image_features = []
            for inputs, _ in tqdm(testloader):
                inputs = inputs.to(device)
                image_features_batch = self.model.encode_image(inputs).float()
                image_features.append(image_features_batch)
            image_features = torch.cat(image_features)

            self.linear = self.linear.to(device).eval()
            logits = self.linear(image_features)
            preds = logits.max(dim=1).indices

        return preds.cpu().numpy(), logits.cpu().numpy()
    
    def serialize(self):
        state_dict = {k: v.cpu() for k, v in self.linear.state_dict().items()}
        model_dict = {'state_dict': state_dict, 'hparams': self.hparams}
        return pickle.dumps(model_dict)


class ClipFinetuneFamily(ModelFamily):
    def __init__(self, family):
        self.family = family

    def learning_rule(self, params):
        return ClipFinetuneLearningRule(self.family, params)
    
    def deserialize_model(self, model_bytes):
        model_dict = pickle.loads(model_bytes)
        model, preprocess = clip.load(self.family)
        linear = nn.Linear(model_dict['state_dict']['weight'].size(1), model_dict['hparams'].num_classes)
        linear.load_state_dict(model_dict['state_dict'])
        return ClipFinetune(model, linear, preprocess, model_dict['hparams'])

    @property 
    def rule_params(self):
        return ClipFinetuneFamilyRuleParams
    
    @property
    def train_params(self):
        return ClipFinetuneFamilyTrainParams
    
    @property
    def eval_params(self):
        return ClipFinetuneFamilyEvalParams


def get_model_families():
    return {'clip_vit_b32_ft': ClipFinetuneFamily('ViT-B/32'), 'clip_resnet50_ft': ClipFinetuneFamily('RN50')}
