import dataclasses
from tqdm import tqdm

import numpy as np
import torch
from torch.utils.data import DataLoader
import clip

from .. import model
from ..model_family import ModelFamily
from ..learning_rule import LearningRuleImplementation
from .imagenet_models import ImageDataset


@dataclasses.dataclass
class ClipZSFamilyRuleParams:
    pass

@dataclasses.dataclass
class ClipZSFamilyTrainParams:
    pass

@dataclasses.dataclass
class ClipZSFamilyEvalParams:
    device: str = 'auto' # 'cuda:0' if cuda available else 'cpu'
    cpu_workers: int = 4
    eval_batch_size: int = 256  


class CLIPZSLearningRule(LearningRuleImplementation):
    def __init__(self, params):
        pass

    def learn(self, dataset, checkpointer, logger, params):
        raise NotImplementedError


classnames = [
['airplane', 'fighter', 'bomber', 'aircraft', 'plane'],
['car', 'automobile', 'suv', 'sedan', 'sports car'],
['bird'],
['cat', 'house cat', 'domestic cat', 'kitten'],
['deer'],
['dog', 'puppy'],
['frog', 'toad'],
['horse', 'race horse', 'pony'],
['ship', 'boat', 'motor boat', 'sail boat', 'yacht'],
['truck', 'freight truck', '18 wheeler', 'semi-trailer truck'],
]

templates = [
'a photo of a {}.',
'a photo of the {}.',
'a blurry photo of a {}.',
'a blurry photo of the {}.',
'a black and white photo of a {}.',
'a black and white photo of the {}.',
'a low contrast photo of a {}.',
'a low contrast photo of the {}.',
'a high contrast photo of a {}.',
'a high contrast photo of the {}.',
'a bad photo of a {}.',
'a bad photo of the {}.',
'a good photo of a {}.',
'a good photo of the {}.',
'a photo of a small {}.',
'a photo of the small {}.',
'a photo of a big {}.',
'a photo of the big {}.',
]


class ClipZSModel(model.ModelImplementation):
    def __init__(self, model, preprocess):
        self.model = model
        self.preprocess = preprocess
    
    def predict(self, dataset, params):
        torch.backends.cudnn.benchmark = True
        device = params.device
        if device == 'auto':
            device = 'cuda:0' if torch.cuda.is_available() else 'cpu'

        testset = ImageDataset(dataset, transform=self.preprocess)
        testloader = DataLoader(testset, batch_size=params.eval_batch_size, shuffle=False,
                                num_workers=params.cpu_workers, pin_memory=True)
        testloader = tqdm(testloader)

        self.model = self.model.to(device).eval()

        with torch.no_grad():
            image_features = []
            for inputs, _ in testloader:
                inputs = inputs.to(device)
                image_features_batch = self.model.encode_image(inputs).float()
                image_features.append(image_features_batch)
            image_features = torch.cat(image_features)

            text_features = []
            for labels in classnames:
                text_descriptions = [template.format(label) for label in labels for template in templates]
                text_input = clip.tokenize(text_descriptions).to(device)
                text_features_class = self.model.encode_text(text_input).float().mean(dim=0)
                text_features.append(text_features_class)
            text_features = torch.stack(text_features)

            image_features = image_features / image_features.norm(dim=-1, keepdim=True)
            text_features = text_features / text_features.norm(dim=-1, keepdim=True)

            logits = self.model.logit_scale.exp() * image_features @ text_features.t()
            preds = logits.max(dim=1).indices

        return preds.cpu().numpy(), logits.cpu().numpy()
    
    def serialize(self):
        return {}


class ClipZSFamily(ModelFamily):
    def __init__(self, family):
        self.family = family

    def learning_rule(self, params):
        return CLIPZSLearningRule(params)
    
    def deserialize_model(self, model_bytes):
        model, preprocess = clip.load(self.family)
        return ClipZSModel(model, preprocess)

    @property 
    def rule_params(self):
        return ClipZSFamilyRuleParams
    
    @property
    def train_params(self):
        return ClipZSFamilyTrainParams
    
    @property
    def eval_params(self):
        return ClipZSFamilyEvalParams


def get_model_families():
    return {'clip_vit_b32_zs': ClipZSFamily('ViT-B/32'), 'clip_resnet50_zs': ClipZSFamily('RN50')}
