import dataclasses
import pickle
import sys

import numpy as np

from .. import learning_rule
from .. import model
from .. import model_family


@dataclasses.dataclass
class RuleParameters:
    param_a: int = 0
    param_b: float = 1.0
    param_c: str = 'hello'
    output_value: int = 0


@dataclasses.dataclass
class TrainParameters:
    additional_output: int = 0
    num_checkpoints: int = 0
    num_log_entries: int = 0
    extra_payload_size: int = 0


@dataclasses.dataclass
class EvalParameters:
    output_value: int = None


class DebugModel(model.ModelImplementation):
    def __init__(self, output_value, extra_payload):
        self.output_value = output_value
        self.extra_payload = extra_payload
    
    def predict(self, dataset, params):
        output_value = self.output_value
        if params.output_value is not None:
            output_value = params.output_value
        res = np.full(len(dataset), output_value)
        return res, None
    
    def featurize(self, dataset, params):
        output_value = self.output_value
        if params.output_value is not None:
            output_value = params.output_value
        res = np.full(len(dataset), output_value + 1)
        return res, None
    
    def serialize(self):
        return pickle.dumps([self.output_value, self.extra_payload])


def deserializer(model_bytes):
    output_value, extra_payload = pickle.loads(model_bytes)
    return DebugModel(output_value, extra_payload)


def rule_generator(hparams):
    def cur_learning_rule(dataset,
                          checkpointer,
                          logger,
                          params):
        for ii in range(params.num_checkpoints):
            extra_payload = np.random.default_rng().bytes(params.extra_payload_size)
            cur_bytes = DebugModel(hparams.output_value + (ii + 1), extra_payload).serialize()
            checkpointer.add(ii, cur_bytes, {'step': ii})
        for ii in range(params.num_log_entries):
            logger.add(str(ii), [ii] * 2, {'ii': ii})
        extra_payload = np.random.default_rng().bytes(params.extra_payload_size)
        return DebugModel(hparams.output_value + params.additional_output, extra_payload)
    return learning_rule.SimpleFunctionLearningRule(cur_learning_rule)


class DebugLearningRule(learning_rule.LearningRuleImplementation):
    def __init__(self, params):
        self.params = params
    
    def learn(self, dataset, checkpointer, logger, params):
        extra_payload = np.random.default_rng().bytes(params.extra_payload_size)
        return DebugModel(self.params.output_value + params.additional_output, extra_payload)


class DebugModelFamily(model_family.ModelFamily):
    def __init__(self):
        pass

    def learning_rule(self, hparams):
        return DebugLearningRule(hparams)
    
    def deserialize_model(self, model_bytes):
        output_value, extra_payload = pickle.loads(model_bytes)
        return DebugModel(output_value, extra_payload)

    @property 
    def rule_params(self):
        return RuleParameters
    
    @property 
    def train_params(self):
        return TrainParameters
    
    @property 
    def eval_params(self):
        return EvalParameters


# This is going to get called automatically and added to the model family registry
def get_model_families():
    return {
        'debug_1': model_family.SimpleModelFamily(rule_generator,
                                                  deserializer,
                                                  RuleParameters,
                                                  TrainParameters,
                                                  EvalParameters),
        'debug_2': model_family.SimpleModelFamily(rule_generator,
                                                  deserializer,
                                                  RuleParameters,
                                                  TrainParameters,
                                                  EvalParameters),
        'debug_3': DebugModelFamily()
    }
