import dataclasses
import typing as ty
import pickle
from tqdm import tqdm
import os

import numpy as np
from PIL import Image
import torch
from torch.utils.data import Dataset, DataLoader
from torchvision import transforms
from torch.optim import SGD
from torch.optim.lr_scheduler import MultiStepLR

from ..model import ModelImplementation
from ..model_family import ModelFamily
from ..learning_rule import LearningRuleImplementation

from .fast_neural_model_impls.core import *
from .fast_neural_model_impls.torch_backend import *
from .fast_neural_model_impls.dawn_utils import net, dataframe


@dataclasses.dataclass
class FastNeuralHyperparams:
    total_examples: int = 50000 * 10  # specifying this way instead of dataset size allows making number of steps scale with batch_size but be independent of datset size
    final_training_fraction: float = 0.2
    ema_momentum: float = 0.99
    ema_freq: int = 5
    batch_size: int = 512
    gbn_splits: int = 16
    momentum: float = 0.9
    weight_decay_coef: float = 5e-4
    bias_coef: float = 64
    lr_schedule: str = 'triangle_20'
    lr_max: float = 1.0
    lr_min: float = 0.1
    whitening_examples: int = int(1e4)
    # TODO: more hparams for architecture?


@dataclasses.dataclass
class FastNeuralTrainingParams:
    seed: int = None
    deterministic: bool = True  # this seems to only make things ~10% slower on 1 single GPU
    device: str = None
    half_precision: bool = True
    log_dir: str = None
    verbose: str = False
    num_checkpoints: int = 0


@dataclasses.dataclass
class FastNeuralEvalParams:
    batch_size: int = None
    device: str = None
    tta: bool = True  # test time augmentation


class FastNeuralLearningRule(LearningRuleImplementation):
    def __init__(self, hparams: FastNeuralHyperparams):
        self.hparams = hparams

    def learn(self, dataset,
              checkpointer, logger, params: FastNeuralTrainingParams):
        
        if dataset.examples is None or dataset.labels is None:
            raise ValueError("Fast neural models only implemented for in-memory datasets.")

        examples, labels = dataset.examples, dataset.labels

        # Resolve hyper params
        hp, tp = self.hparams, params
        batch_size = hp.batch_size
        n_data = len(labels)
        epochs = int(np.ceil(hp.total_examples / n_data))
        final_epochs = int(np.round(epochs * hp.final_training_fraction))
        if hp.lr_schedule.startswith('triangle'):
            rise_epochs = float(hp.lr_schedule.split('_')[1]) / 100 * epochs
            lr_schedule = PiecewiseLinear(
                [0, rise_epochs, epochs - final_epochs],
                [0, hp.lr_max, hp.lr_min])
        else:
            raise ValueError(f'Unknown learning rate schedule '
                             f'{hp.lr_schedule}')

        assert batch_size % hp.gbn_splits == 0  # for ghost BN config
        weight_decay = hp.weight_decay_coef * batch_size
        bias_coef = hp.bias_coef
        momentum = hp.momentum
        train_transforms = [Crop(32, 32), FlipLR()]

        # Resolve device
        device = determine_device(tp.device)
        if tp.seed is not None:
            torch.manual_seed(tp.seed)
            np.random.seed(tp.seed)
        torch.backends.cudnn.deterministic = tp.deterministic

        # looks like torch doesn't support half precision conv on cpu
        assert device != cpu or not tp.half_precision
        def cast(x):
            return x.half() if tp.half_precision else x.float()

        def printo(*args):
            if tp.verbose:
                print(*args)

        printo('Warming up torch')
        random_data = cast(torch.tensor(
            np.random.randn(1000, 3, 32, 32), device=device))
        eigens(patches(random_data))

        loss = label_smoothing_loss(0.2)
        random_batch = lambda bs: {
            'input': cast(torch.Tensor(np.random.rand(bs, 3, 32, 32))).to(device),
            'target': torch.LongTensor(
                np.random.randint(0, 10, bs)).to(device)
        }
        printo('Warming up cudnn on random inputs')
        model = FastNeuralModel(hp, tp).torch_model.to(device)
        warmup_cudnn(model, loss, random_batch(batch_size))

        printo('Starting timer')
        timer = Timer(synch=torch.cuda.synchronize if device != cpu else None)

        printo('Preprocessing training data')
        examples = torch.tensor(examples).float()
        labels = torch.tensor(labels)
        data_mean = examples.mean(dim=(0, 1, 2))
        data_std = examples.std(dim=(0, 1, 2))
        examples = cast(examples).permute(0, 3, 1, 2)
        train_set = {'data': examples.to(device), 'targets': labels.to(device)}
        train_set = preprocess(train_set, [partial(pad, border=4)])
        printo(f'Finished in {timer():.2} seconds')

        # adding normalization as an explicit layer in the model
        # streamlines model serialization and loading
        noramlize = Normalization(
            mean=cast(data_mean), std=cast(data_std)).to(device)
        Λ, V = eigens(patches(
            noramlize(train_set['data'][:hp.whitening_examples,
                      :, 4:-4, 4:-4])))  # center crop to remove padding
        output = FastNeuralModel(hp, tp,
                                 normalization=(data_mean, data_std, Λ, V))
        av_model = output.torch_model.to(device)
        model = copy.deepcopy(av_model)

        train_batches = GPUBatches(batch_size=batch_size,
                                   transforms=train_transforms,
                                   dataset=train_set,
                                   shuffle=True, drop_last=True,
                                   max_options=200)
        is_bias = group_by_key(
            ('bias' in k, v) for k, v in trainable_params(model).items())
        opts = [
            SGD(is_bias[False], {'lr': (
                lambda step: lr_schedule(
                    step / len(train_batches)) / batch_size),
                'weight_decay': Const(weight_decay),
                'momentum': Const(momentum)}),
            SGD(is_bias[True], {'lr': (
                lambda step: lr_schedule(step / len(train_batches)) * (
                        bias_coef / batch_size)),
                'weight_decay': Const(weight_decay / bias_coef),
                'momentum': Const(momentum)})
        ]
        logs, state = Table(report=tp.verbose), {MODEL: model,
                                                 VALID_MODEL: av_model,
                                                 LOSS: loss,
                                                 OPTS: opts}

        # Main train loop
        checkpoint_interval = (int(np.ceil(epochs / tp.num_checkpoints))
                               if tp.num_checkpoints else epochs)
        for epoch in range(epochs):
            ema_step = update_ema(momentum=hp.ema_momentum,
                                  update_freq=hp.ema_freq)
            logs.append(union({'epoch': epoch + 1},
                              train_epoch(state, timer, train_batches,
                                          train_steps=(*default_train_steps,
                                                       ema_step))))
            if (epochs - (epoch + 1)) % checkpoint_interval == 0:
                checkpointer.add(
                    epoch + 1, output.serialize(), {'epoch': epoch + 1})
            logger.add('log dump', dataframe(logs.log).to_csv(index=False),
                       dict(epoch=epoch + 1))

        if tp.log_dir is not None:
            dataframe(logs.log).to_csv(
                os.path.join(os.path.expanduser(tp.log_dir), 'log.csv'),
                index=False)

        return output


class FastNeuralModel(ModelImplementation):
    def __init__(self, hp: FastNeuralHyperparams,
                 tp: FastNeuralTrainingParams,
                 normalization=None, state_dict=None):
        # Setups architecture elements (why is this outside core code?)
        batch_norm = partial(GhostBatchNorm,
                             num_splits=hp.gbn_splits, weight_freeze=True)
        relu = partial(nn.CELU, alpha=0.3)

        def conv_bn(c_in, c_out, pool=None):
            block = {
                'conv': nn.Conv2d(c_in, c_out, kernel_size=3, stride=1,
                                  padding=1, bias=False),
                'bn': batch_norm(c_out),
                'relu': relu(),
            }
            if pool: block = {'conv': block['conv'], 'pool': pool,
                              'bn': block['bn'], 'relu': block['relu']}
            return block

        def whitening_block(c_in, c_out, Λ=None, V=None, eps=1e-2):
            return {
                'whiten': whitening_filter(Λ, V, eps),
                'conv': nn.Conv2d(27, c_out, kernel_size=(1, 1), bias=False),
                'norm': batch_norm(c_out),
                'act': relu()
            }

        if normalization is None:
            mean = torch.zeros(3)
            std = torch.ones(3)
            Λ = None
            V = None
        else:
            mean, std, Λ, V = map(to(cpu), normalization)
        prep = lambda a,b: union(
            {'normalize': Normalization(mean=mean, std=std)},
            whitening_block(a, b, Λ=Λ, V=V))

        self.torch_model = Network(net(
            weight=1 / 16, conv_bn=conv_bn, prep=prep))
        if tp.half_precision:
            self.torch_model.half()

        if state_dict is not None:
            self.torch_model.load_state_dict(state_dict)

        self.hp = hp
        self.tp = tp

    def predict(self, dataset, params: FastNeuralEvalParams):

        if dataset.examples is None:
            raise ValueError("Fast neural models only implemented for in-memory datasets.")
        examples = dataset.examples
        batch_size = params.batch_size or self.hp.batch_size
        device = determine_device(params.device)
        assert device != cpu or not self.tp.half_precision
        examples = torch.FloatTensor(examples).to(device).permute(0, 3, 1, 2)
        if self.tp.half_precision:
            examples = examples.half()
        # add fake labels
        dataset = dict(data=examples,
                       targets=torch.zeros(len(examples)).long().to(device))
        batches = GPUBatches(batch_size=batch_size, dataset=dataset,
                             shuffle=False, drop_last=False)
        model = self.torch_model
        model.to(device)
        model.train(False)
        tta_transforms = [(lambda x: x), flip_lr]  # TODO: options for other transforms

        logits = []
        for batch in batches:
            with torch.no_grad():
                if params.tta:
                    logits_ = torch.mean(torch.stack([model(
                        {'input': transform(batch['input'].clone())})[
                                                          'logits'].detach()
                                                      for transform in
                                                      tta_transforms],
                                                     dim=0), dim=0)
                else:
                    logits_ = model(batch)['logits']
                logits.append(logits_.cpu().numpy())
        logits = np.concatenate(logits)
        preds = logits.argmax(-1)

        model.to(cpu)

        return preds, logits

    def serialize(self):
        device = next(self.torch_model.parameters()).device
        model_dict = {'state_dict': self.torch_model.cpu().state_dict(),
                      'hp': self.hp, 'tp': self.tp}
        self.torch_model.to(device)
        return pickle.dumps(model_dict)


class FastNeuralModelFamily(ModelFamily):
    def learning_rule(self, hparams):
        return FastNeuralLearningRule(hparams)

    def deserialize_model(self, model_bytes):
        return FastNeuralModel(**pickle.loads(model_bytes))

    @property
    def rule_params(self):
        return FastNeuralHyperparams

    @property
    def train_params(self):
        return FastNeuralTrainingParams

    @property
    def eval_params(self):
        return FastNeuralEvalParams


def get_model_families():
    return {'myrtlenet': FastNeuralModelFamily()}

