import os
import random
import dataclasses
import typing as ty
import pickle
from tqdm import tqdm

import numpy as np
from PIL import Image
import torch
import torch.nn as nn
from torch.utils.data import DataLoader, Dataset
from torchvision import transforms
from torch.optim import SGD, Adam
from torch.optim.lr_scheduler import MultiStepLR, StepLR
import torch.multiprocessing as mp
import torch.distributed as dist
import cnn_finetune
from cnn_finetune import make_model

from .. import model
from ..dataset_family import DatasetImplementation
from ..model_family import ModelFamily
from ..learning_rule import LearningRuleImplementation
from ..data_families.wilds_data import WildsDataset
from testbed.data_families import color_jitter


@dataclasses.dataclass
class ImagenetModelFamilyRuleParams:
    num_classes: int
    input_size: ty.Tuple[int] = (224, 224)
    pretrained: bool = True
    finetune_only_last_layer: bool = False
    batch_size: int = 256  # total across all gpus
    epochs: int = 90
    optimizer: str = 'SGD'
    lr: float = 0.1
    momentum: float = 0.9
    weight_decay: float = 1e-4
    lr_scheduler: str = 'MultiStepLR'
    milestones: ty.Tuple[int] = (30, 60, 80)
    gamma: float = 0.1
    dropout_p: float = None # None disables dropout layer
    color_jitter: bool = False # add color jitter transform (for camelyon)

@dataclasses.dataclass
class ImagenetModelFamilyTrainParams:
    cpu_workers: int = 4 # number per gpu
    seed: int = None
    early_stop: bool = False
    num_checkpoints: int = 10
    additional_checkpoint_epochs: ty.Tuple[int] = ()
    dist_url: str = f'tcp://127.0.0.1:{np.random.randint(20000, 30000)}'

@dataclasses.dataclass
class ImagenetModelFamilyEvalParams:
    device: str = 'auto' # 'cuda:0' if cuda available else 'cpu'
    cpu_workers: int = 4
    eval_batch_size: int = None # defaults to ImagenetModelFamilyRuleParams().batch_size if not provided


class ImagenetLearningRule(LearningRuleImplementation):
    def __init__(self, family, params):
        self.family = family
        self.hparams = params
    
    def learn(self, dataset, checkpointer, logger, params):
        if params.early_stop:
            raise NotImplementedError('Early stopping for imagenet learning rules not yet implemented.')

        os.environ['MKL_THREADING_LAYER'] = 'GNU' # don't know why I need this
        num_gpus = torch.cuda.device_count()
        print(f'Found {num_gpus} cuda devices, training across all.')
        # Use torch.multiprocessing.spawn to launch distributed processes on gpus 1...n
        process_context = mp.spawn(main_worker, nprocs=num_gpus-1, join=False, 
                                    args=(num_gpus, dataset.implementation, self.family, self.hparams, params))
        # Launch rank 0 gpu process that will also do logging/checkpointing and return the final model
        model = main_worker(-1, num_gpus, dataset.implementation, self.family, self.hparams, params, checkpointer, logger)
        # wait for all processes to finish
        process_context.join()
        dist.destroy_process_group()
        return model

def main_worker(gpu, num_gpus, dataset, family, hparams, params, checkpointer=None, logger=None):
    # workaround for pytorch launching on gpus 1...n with ids 0...n-1
    gpu += 1
    
     # only one process will do logging, checkpointing, etc.
    is_master = gpu == 0

    dist.init_process_group(backend='nccl', init_method=params.dist_url, 
                            world_size=num_gpus, rank=gpu)
    torch.cuda.set_device(gpu)

    torch.backends.cudnn.benchmark = True
    if params.seed is not None:
        if is_master:
            print('You have chosen to seed training. This will incur a significant training ' + \
                'speed and (possible) performance hit. Please make sure this is intentional.')
        seed_process(params.seed)
        torch.backends.cudnn.benchmark = False 
        torch.backends.cudnn.deterministic = True

    # make the model
    if is_master:
        print(f'creating model {family}')
    model = make_model(family, num_classes=hparams.num_classes, input_size=hparams.input_size,
                        pretrained=hparams.pretrained, dropout_p=hparams.dropout_p)
    if hparams.finetune_only_last_layer:
        last_layer_param_ptrs = [param.data_ptr() for param in model._classifier.parameters()]
        for param in model.parameters():
            if param.data_ptr() not in last_layer_param_ptrs:
                param.requires_grad = False

    # add normalization layer
    info = model.original_model_info
    max_input_range = max(info.input_range)
    mean, std = [x / max_input_range for x in info.mean], [x / max_input_range for x in info.std]
    model = nn.Sequential(
        NormalizationLayer(mean, std, info.input_space),
        model,
    )
    model = model.cuda()
    model = torch.nn.parallel.DistributedDataParallel(model, device_ids=[gpu])

    # make optimizers
    if hparams.optimizer == 'SGD':
        optimizer = SGD(model.parameters(), lr=hparams.lr, 
                        momentum=hparams.momentum, weight_decay=hparams.weight_decay)
    elif hparams.optimizer == 'Adam':
        optimizer = Adam(model.parameters(), lr=hparams.lr,
                         weight_decay=hparams.weight_decay,
                         betas=(hparams.momentum, 0.999))

    if hparams.lr_scheduler == 'MultiStepLR':
        scheduler = MultiStepLR(optimizer, milestones=hparams.milestones, gamma=hparams.gamma)
    if hparams.lr_scheduler == 'StepLR':
        scheduler = StepLR(optimizer, step_size=1, gamma=hparams.gamma)
    criterion = nn.CrossEntropyLoss().cuda()

    # create the train loders
    if is_master:
        print('initializing dataloaders')
    transform = [transforms.Resize(hparams.input_size), transforms.ToTensor()]
    if hparams.color_jitter:
        transform.append(color_jitter.ColorAug(0.05))
    trainset = ImageDataset(dataset, transforms.Compose(transform))
    per_process_batch_size = hparams.batch_size // num_gpus + int(gpu < hparams.batch_size % num_gpus)
    train_sampler = torch.utils.data.distributed.DistributedSampler(trainset)
    trainloader = DataLoader(trainset, batch_size=per_process_batch_size, num_workers=params.cpu_workers,
                            sampler=train_sampler, pin_memory=True, worker_init_fn=_worker_init_fn)

    model.train()
    checkpoint_epochs = [*np.linspace(0, hparams.epochs-1, num=params.num_checkpoints, dtype=np.int)] + \
                        list(params.additional_checkpoint_epochs)
    train_losses = []

    # main training loop
    for epoch in range(hparams.epochs):
        losses_epoch = []
        if is_master:
            trainloader = tqdm(trainloader, desc=f'Training, Epoch {epoch}/{hparams.epochs}')

        train_sampler.set_epoch(epoch)
        for inputs, targets in trainloader:
            inputs, targets = inputs.cuda(non_blocking=True), targets.cuda(non_blocking=True)
            outputs = model(inputs)
            loss = criterion(outputs, targets)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()

            losses_epoch.append(loss.item())
            if is_master:
                trainloader.set_postfix({'train_loss': loss.item()})
            
        scheduler.step()
        train_losses.append(losses_epoch)
        if is_master:
            logger.add('train_losses', train_losses, {})
            if epoch in checkpoint_epochs:
                print(f"Adding a checkpoint of epoch {epoch}")
                checkpointer.add(epoch, ImagenetModel(model.module, hparams).serialize(), {'epoch': epoch})

    if is_master:
        return ImagenetModel(model.module.cpu(), hparams)


class NormalizationLayer(nn.Module):
    def __init__(self, mean, std, input_space='RGB'):
        super().__init__()
        self.mean = nn.Parameter(torch.tensor(mean).float().view(3, 1, 1), requires_grad=False)
        self.std = nn.Parameter(torch.tensor(std).float().view(3, 1, 1), requires_grad=False)
        self.input_space = input_space

    def forward(self, x):
        if self.input_space == 'BGR':
            x = x.flip(1)
        return (x - self.mean) / self.std


class ImagenetModel(model.ModelImplementation):
    def __init__(self, model, hparams):
        self.model = model
        self.hparams = hparams
    
    def predict(self, dataset, params):
        torch.backends.cudnn.benchmark = True
        device = params.device
        if device == 'auto':
            device = 'cuda:0' if torch.cuda.is_available() else 'cpu'

        testset = ImageDataset(dataset, transform=transforms.Compose([
            transforms.Resize(self.hparams.input_size),
            transforms.ToTensor(),
        ]))
        batch_size = self.hparams.batch_size if params.eval_batch_size is None else params.eval_batch_size
        testloader = DataLoader(testset, batch_size=batch_size, shuffle=False,
                                num_workers=params.cpu_workers, pin_memory=True)
        testloader = tqdm(testloader)

        self.model = self.model.to(device).eval()
        logits = []
        for inputs, _ in testloader:
            inputs = inputs.to(device)
            with torch.no_grad():
                outputs = self.model(inputs)
            logits.append(outputs)

        logits = torch.cat(logits)
        preds = logits.max(dim=1).indices
        return preds.cpu().numpy(), logits.cpu().numpy()
    
    def serialize(self):
        state_dict = {k: v.cpu() for k, v in self.model.state_dict().items()}
        model_dict = {'state_dict': state_dict, 'hparams': self.hparams, 
                    'input_space': self.model[0].input_space}
        return pickle.dumps(model_dict)


class ImagenetModelFamily(ModelFamily):
    def __init__(self, family):
        self.family = family

    def learning_rule(self, params):
        return ImagenetLearningRule(self.family, params)
    
    def deserialize_model(self, model_bytes):
        model_dict = pickle.loads(model_bytes)
        model = nn.Sequential(
            NormalizationLayer([0]*3, [0]*3, model_dict['input_space']),
            make_model(self.family, num_classes=model_dict['hparams'].num_classes, 
                        input_size=model_dict['hparams'].input_size, pretrained=False),
        )
        model.load_state_dict(model_dict['state_dict'])
        return ImagenetModel(model, model_dict['hparams'])

    @property 
    def rule_params(self):
        return ImagenetModelFamilyRuleParams
    
    @property
    def train_params(self):
        return ImagenetModelFamilyTrainParams
    
    @property
    def eval_params(self):
        return ImagenetModelFamilyEvalParams


def get_model_families():
    invalid_models = set(['inception_v3', 'polynet', 'inceptionresnetv2', 'inception_v4'])
    valid_models = set(cnn_finetune.base.MODEL_REGISTRY.keys()) - invalid_models
    return {'imgnt_' + name: ImagenetModelFamily(name) for name in valid_models}


class ImageDataset(Dataset):
    def __init__(self, dataset, transform=None):
        self.dataset = dataset
        self.transform = transform

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        example = self.dataset[idx]
        if isinstance(example, tuple) and len(example) == 3:
            if isinstance(self.dataset, DatasetImplementation):
                assert isinstance(self.dataset, WildsDataset)
            else:
                assert isinstance(self.dataset.implementation, WildsDataset)
            image, label, _ = example
        elif isinstance(example, tuple) and len(example) == 2:
            image, label = example
        else:
            image, label = example, None

        # convert image to PIL if not already that way
        if isinstance(image, Image.Image):
            pass
        elif isinstance(image, np.ndarray):
            if image.dtype == np.uint8:
                image = Image.fromarray(image)
            elif image.min() >= 0 and image.max() <= 1 and isinstance(image, np.floating):
                image = Image.fromarray((image * 255).astype(np.uint8))
            else:
                raise NotImplementedError(f'data type {type(image)} not recognized')
        else:
            raise NotImplementedError(f'data type {type(image)} not recognized')

        if self.transform:
            image = self.transform(image)
        if label is not None:
            return image, label
        return image


def seed_process(seed):
    random.seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed) # if you are using multi-GPU.

def _worker_init_fn(worker_id):
    seed = torch.initial_seed()
    random.seed(seed)
    np.random.seed(seed // 2**32-1)
