import os
import random
import dataclasses
import typing as ty
import pickle
from tqdm import tqdm

import numpy as np
from PIL import Image
import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
from torchvision import transforms
from torch.optim import SGD
from torch.optim.lr_scheduler import MultiStepLR

from .. import model
from ..model_family import ModelFamily
from ..model_families import neural_model_impls
from ..learning_rule import LearningRuleImplementation
from .imagenet_models import ImageDataset, seed_process


DEFAULT_TRAIN_EXAMPLES = 50000

@dataclasses.dataclass
class NeuralModelFamilyRuleParams:
    batch_size: int = 128
    total_examples: int = DEFAULT_TRAIN_EXAMPLES * 350
    optimizer: str = 'SGD'
    lr: float = 0.1
    momentum: float = 0.9
    weight_decay: float = 5e-4
    lr_scheduler: str = 'MultiStepLR'
    milestones: ty.Tuple[int] = (DEFAULT_TRAIN_EXAMPLES * 150, DEFAULT_TRAIN_EXAMPLES * 250)
    gamma: float = 0.1
    
@dataclasses.dataclass
class NeuralModelFamilyTrainParams:
    device: str = 'auto' # 'cuda:0' if cuda available else 'cpu'
    num_workers: int = 2
    seed: int = None
    early_stop: bool = False
    num_checkpoints: int = 10
    additional_checkpoint_epochs: ty.Tuple[int] = ()

@dataclasses.dataclass
class NeuralModelFamilyEvalParams:
    device: str = 'auto' # 'cuda:0' if cuda available else 'cpu'
    num_workers: int = 2
    eval_batch_size: int = None # defaults to ImagenetModelFamilyRuleParams().batch_size if not provided


class NeuralLearningRule(LearningRuleImplementation):
    def __init__(self, family, params):
        self.family = family
        self.hparams = params
        self.criterion = torch.nn.CrossEntropyLoss()
    
    def learn(self, dataset, checkpointer, logger, params):
        if params.early_stop:
            raise NotImplementedError("Early stopping for neural learning rules not yet implemented.")

        torch.backends.cudnn.benchmark = True
        if params.seed is not None:
            print('You have chosen to seed training. This will incur a significant training ' + \
                'speed and (possible) performance hit. Please make sure this is intentional.')
            seed_process(params.seed)
            torch.backends.cudnn.benchmark = False 
            torch.backends.cudnn.deterministic = True

        device = params.device
        if device == 'auto':
            device = 'cuda:0' if torch.cuda.is_available() else 'cpu'

        epochs = self.hparams.total_examples // len(dataset)
        milestones = [milestone // len(dataset) for milestone in self.hparams.milestones]

        # TODO: Compute via dataset iteration?
        # Some dataset may not support .examples
        examples_temp = dataset.examples.astype(np.float) / 255
        data_mean = torch.from_numpy(examples_temp.mean(axis=(0, 1, 2))).float()
        data_std = torch.from_numpy(examples_temp.std(axis=(0, 1, 2))).float()

        model = getattr(neural_model_impls, self.family)()
        model = nn.Sequential(
            NormalizationLayer(data_mean, data_std),
            model,
        )

        transform_train = transforms.Compose([
            transforms.RandomCrop(32, padding=4),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
        ])
        trainset = ImageDataset(dataset, transform_train)
        worker_init_fn = None if params.seed is None else \
                        lambda worker_id: np.random.seed(params.seed+worker_id+1)
        trainloader = DataLoader(trainset,
                                 batch_size=self.hparams.batch_size,
                                 shuffle=True, 
                                 num_workers=params.num_workers,
                                 worker_init_fn=worker_init_fn)

        model = model.to(device)
        if self.hparams.optimizer == 'SGD':
            optimizer = SGD(model.parameters(), lr=self.hparams.lr, 
                            momentum=self.hparams.momentum, weight_decay=self.hparams.weight_decay)
        if self.hparams.lr_scheduler == 'MultiStepLR':
            scheduler = MultiStepLR(optimizer, milestones=milestones, gamma=self.hparams.gamma)

        train_losses = []
        checkpoint_epochs = [*np.linspace(0, epochs-1, num=params.num_checkpoints, dtype=np.int)] + \
                            list(params.additional_checkpoint_epochs)

        for epoch in range(epochs):
            model.train()
            losses_epoch = []
            loop = tqdm(trainloader, desc=f'Training, Epoch {epoch}/{epochs}')

            for inputs, targets in loop:
                inputs, targets = inputs.to(device), targets.to(device)
                optimizer.zero_grad()
                outputs = model(inputs)
                loss = self.criterion(outputs, targets)
                loss.backward()
                optimizer.step()
                loop.set_postfix({'train_loss': loss.item()})
                losses_epoch.append(loss.item())
                
            scheduler.step()

            train_losses.append(losses_epoch)
            logger.add('train_losses', train_losses, {})
            if epoch in checkpoint_epochs:
                checkpointer.add(epoch, NeuralModel(model, self.hparams).serialize(), {'epoch': epoch})

        return NeuralModel(model, self.hparams)


class NormalizationLayer(nn.Module):
    def __init__(self, mean, std):
        super().__init__()
        self.mean = nn.Parameter(mean.unsqueeze(1).unsqueeze(2), requires_grad=False)
        self.std = nn.Parameter(std.unsqueeze(1).unsqueeze(2), requires_grad=False)

    def forward(self, x):
        return (x - self.mean) / self.std


class NeuralModel(model.ModelImplementation):
    def __init__(self, model, hparams):
        self.model = model
        self.hparams = hparams
    
    def predict(self, dataset, params):
        torch.backends.cudnn.benchmark = True
        device = params.device
        if device == 'auto':
            device = 'cuda:0' if torch.cuda.is_available() else 'cpu'

        testset = ImageDataset(dataset, transform=transforms.ToTensor())
        batch_size = self.hparams.batch_size if params.eval_batch_size is None else params.eval_batch_size
        testloader = DataLoader(testset, batch_size=batch_size, shuffle=False,
                                num_workers=params.num_workers)

        self.model = self.model.to(device)
        self.model = self.model.eval()
        logits = []
        for inputs, _ in testloader:
            inputs = inputs.to(device)
            with torch.no_grad():
                outputs = self.model(inputs)
            logits.append(outputs)

        logits = torch.cat(logits)

        if dataset.params.classes_to_exclude is not None:
            keep_columns = torch.ones(logits.shape[1])
            keep_columns[dataset.params.classes_to_exclude] = 0
            
            logits = logits[:, keep_columns.bool()]
            preds = logits.max(dim=1).indices

            # Remap the predictions from the subsampled label space
            # to the original label space
            original_cls_idxs = torch.arange(len(keep_columns))
            original_cls_idxs = original_cls_idxs[keep_columns.bool()]
            preds = original_cls_idxs[preds]
        else:
            preds = logits.max(dim=1).indices
        return preds.cpu().numpy(), logits.cpu().numpy()
    
    def serialize(self):
        state_dict = {k: v.cpu() for k, v in self.model.state_dict().items()}
        model_dict = {'state_dict': state_dict, 'hparams': self.hparams}
        return pickle.dumps(model_dict)


class NeuralModelFamily(ModelFamily):
    def __init__(self, family):
        self.family = family

    def learning_rule(self, params):
        return NeuralLearningRule(self.family, params)
    
    def deserialize_model(self, model_bytes):
        model_dict = pickle.loads(model_bytes)
        model = nn.Sequential(
            NormalizationLayer(torch.randn(3), torch.randn(3)),
            getattr(neural_model_impls, self.family)(),
        )
        model.load_state_dict(model_dict['state_dict'])
        return NeuralModel(model, model_dict['hparams'])

    @property 
    def rule_params(self):
        return NeuralModelFamilyRuleParams
    
    @property
    def train_params(self):
        return NeuralModelFamilyTrainParams
    
    @property
    def eval_params(self):
        return NeuralModelFamilyEvalParams


def get_model_families():
    return {name: NeuralModelFamily(name) for name in neural_model_impls.names}
