import io
import os
import random
import dataclasses
import typing as ty
import pickle

import numpy as np
from PIL import Image
import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
from torchvision import transforms

from .. import model
from ..model_family import ModelFamily
from .odst_impls import pyramid, resnet
from ..learning_rule import LearningRuleImplementation
from .imagenet_models import ImageDataset, seed_process


@dataclasses.dataclass
class ODSTModelFamilyRuleParams:
    num_odst_iterations: int


@dataclasses.dataclass
class ODSTModelFamilyTrainParams:
    pass


@dataclasses.dataclass
class ODSTModelFamilyEvalParams:
    device: str = 'auto' # 'cuda:0' if cuda available else 'cpu'
    num_workers: int = 2
    eval_batch_size: int = 128


class ODSTLearningRule(LearningRuleImplementation):
    def __init__(self, family, params):
        self.family = family
        self.hparams = params

    def learn(self, dataset, checkpointer, logger, params):
        raise NotImplementedError


class NormalizationLayer(nn.Module):
    def __init__(self, mean, std):
        super().__init__()
        self.mean = nn.Parameter(mean.unsqueeze(1).unsqueeze(2), requires_grad=False)
        self.std = nn.Parameter(std.unsqueeze(1).unsqueeze(2), requires_grad=False)

    def forward(self, x):
        return (x - self.mean) / self.std


class ODSTModel(model.ModelImplementation):
    def __init__(self, model):
        self.model = model
    
    def predict(self, dataset, params):
        torch.backends.cudnn.benchmark = True
        device = params.device
        if device == 'auto':
            device = 'cuda:0' if torch.cuda.is_available() else 'cpu'
        testset = ImageDataset(dataset, transform=transforms.ToTensor())
        testloader = DataLoader(testset,
                                batch_size=params.eval_batch_size,
                                shuffle=False,
                                num_workers=params.num_workers)
        
        self.model = self.model.to(device)
        self.model = self.model.eval()
        logits = []
        for inputs, _ in testloader:
            inputs = inputs.to(device)
            with torch.no_grad():
                outputs = self.model(inputs)
            logits.append(outputs)

        logits = torch.cat(logits)

        if dataset.params.classes_to_exclude is not None:
            keep_columns = torch.ones(logits.shape[1])
            keep_columns[dataset.params.classes_to_exclude] = 0
            
            logits = logits[:, keep_columns.bool()]
            preds = logits.max(dim=1).indices

            # Remap the predictions from the subsampled label space
            # to the original label space
            original_cls_idxs = torch.arange(len(keep_columns))
            original_cls_idxs = original_cls_idxs[keep_columns.bool()]
            preds = original_cls_idxs[preds]
        else:
            preds = logits.max(dim=1).indices
        return preds.cpu().numpy(), logits.cpu().numpy()
    
    def serialize(self):
        raise NotImplementedError


class ODSTModelFamily(ModelFamily):
    def __init__(self, family):
        self.family = family

    def learning_rule(self, params):
        return ODSTLearningRule(self.family, params)
    
    def deserialize_model(self, model_bytes):
        if self.family == 'resnet50':
            model_class = resnet.ResNet50
        elif self.family == 'shakedroppyramid272':
            model_class = pyramid.aa_PyramidNet
        else:
            raise ValueError(self.family)
        state_dict = torch.load(io.BytesIO(model_bytes), map_location=torch.device('cpu'))
        inner_model = model_class()
        inner_model.load_state_dict(state_dict)
        model = nn.Sequential(
            NormalizationLayer(mean=torch.tensor([0.4913997551666284, 0.48215855929893703, 0.4465309133731618]),
                               std=torch.tensor([0.24703225141799082, 0.24348516474564, 0.26158783926049628])),
            inner_model)
        return ODSTModel(model)

    @property 
    def rule_params(self):
        return ODSTModelFamilyRuleParams
    
    @property
    def train_params(self):
        return ODSTModelFamilyTrainParams
    
    @property
    def eval_params(self):
        return ODSTModelFamilyEvalParams


def get_model_families():
    return {'odst_resnet50': ODSTModelFamily('resnet50'),
            'odst_shakedroppyramid272': ODSTModelFamily('shakedroppyramid272')}