import dataclasses
import typing as ty
import pickle
from tqdm import tqdm

import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from torchvision import transforms
import cnn_finetune
from cnn_finetune import make_model

from .. import model
from ..model_family import ModelFamily
from ..learning_rule import LearningRuleImplementation
from .imagenet_models import ImageDataset, NormalizationLayer


@dataclasses.dataclass
class PretrainedModelFamilyRuleParams:
    input_size: ty.Tuple[int] = (224, 224)

@dataclasses.dataclass
class PretrainedModelFamilyTrainParams:
    pass

@dataclasses.dataclass
class PretrainedModelFamilyEvalParams:
    device: str = 'auto' # 'cuda:0' if cuda available else 'cpu'
    cpu_workers: int = 4
    eval_batch_size: int = 256  


class PretrainedLearningRule(LearningRuleImplementation):
    def __init__(self, family, params):
        self.family = family
        self.hparams = params
    
    def learn(self, dataset, checkpointer, logger, params):
        raise NotImplementedError


cifar10_idxs = [
    [404, 908, 895], 
    [407, 436, 468, 511, 609, 627, 661, 751, 817], 
    [10, 8, 134, 90, 92, 18, 145, 143, 23, 96, 93, 127, 94, 97, 95, 141, 140, 89, 22, 7, 12, 83, 138, 139, 98, 146, 129, 24, 16, 136, 133, 14, 128, 15, 13, 85, 11, 20, 131, 132, 130, 137, 82, 87, 142, 91, 100, 81, 9, 84, 135, 17, 99, 88, 80, 19, 21, 144, 86], 
    [285, 283, 284, 281, 282], 
    [353, 351, 352], 
    [162, 249, 202, 205, 199, 238, 258, 218, 267, 265, 233, 232, 245, 241, 213, 153, 247, 242, 180, 214, 263, 266, 251, 156, 181, 220, 224, 195, 171, 234, 179, 164, 191, 216, 186, 174, 217, 160, 261, 177, 226, 189, 155, 257, 252, 239, 157, 182, 176, 187, 264, 229, 246, 211, 244, 175, 259, 250, 203, 243, 192, 201, 169, 165, 260, 172, 197, 254, 228, 158, 255, 248, 206, 154, 212, 151, 168, 237, 159, 194, 161, 230, 178, 225, 184, 222, 210, 268, 193, 240, 200, 221, 152, 236, 256, 253, 173, 231, 183, 198, 167, 166, 219, 235, 208, 170, 196, 163, 215, 204, 262, 223, 209, 188, 207, 190, 185, 227], 
    [30, 32, 31], 
    [339], 
    [403, 510, 554, 576, 625, 628, 724, 780, 814, 871, 914, 472, 484], 
    [555, 569, 675, 864, 867]
]


class PretrainedModel(model.ModelImplementation):
    def __init__(self, model, hparams):
        self.model = model
        self.hparams = hparams
    
    def predict(self, dataset, params):
        torch.backends.cudnn.benchmark = True
        device = params.device
        if device == 'auto':
            device = 'cuda:0' if torch.cuda.is_available() else 'cpu'

        testset = ImageDataset(dataset, transform=transforms.Compose([
            transforms.Resize(self.hparams.input_size),
            transforms.ToTensor(),
        ]))
        testloader = DataLoader(testset, batch_size=params.eval_batch_size, shuffle=False,
                                num_workers=params.cpu_workers, pin_memory=True)
        testloader = tqdm(testloader)

        self.model = self.model.to(device).eval()
        logits = []
        for inputs, _ in testloader:
            inputs = inputs.to(device)
            with torch.no_grad():
                outputs = self.model(inputs)
            logits.append(outputs)

        logits = torch.cat(logits)

        cifar10_logits = torch.empty(logits.size(0), 10).to(device)
        for i in range(10):
            cifar10_logits[:, i] = logits[:, cifar10_idxs[i]].max(dim=1).values

        preds = cifar10_logits.max(dim=1).indices
        return preds.cpu().numpy(), logits.cpu().numpy()
    
    def serialize(self):
        state_dict = {k: v.cpu() for k, v in self.model.state_dict().items()}
        model_dict = {'state_dict': state_dict, 'hparams': self.hparams, 
                    'input_space': self.model[0].input_space}
        return pickle.dumps(model_dict)


class PretrainedModelFamily(ModelFamily):
    def __init__(self, family):
        self.family = family

    def learning_rule(self, params):
        return PretrainedLearningRule(self.family, params)
    
    def deserialize_model(self, model_bytes):
        model_dict = pickle.loads(model_bytes)
        model = nn.Sequential(
            NormalizationLayer([0]*3, [0]*3, model_dict['input_space']),
            make_model(self.family, num_classes=1000, use_original_classifier=True,
                        input_size=model_dict['hparams'].input_size, pretrained=False),
        )
        model.load_state_dict(model_dict['state_dict'])
        return PretrainedModel(model, model_dict['hparams'])

    @property 
    def rule_params(self):
        return PretrainedModelFamilyRuleParams
    
    @property
    def train_params(self):
        return PretrainedModelFamilyTrainParams
    
    @property
    def eval_params(self):
        return PretrainedModelFamilyEvalParams


def get_model_families():
    invalid_models = set(['inception_v4', 'nasnetamobile', 'nasnetalarge', 'pnasnet5large'])
    valid_models = set(cnn_finetune.base.MODEL_REGISTRY.keys()) - invalid_models
    return {'pretrained_' + name: PretrainedModelFamily(name) for name in valid_models}
