import argparse
import time
import multiprocessing as mp
import numpy as np
import scipy.linalg
import torch

from . import coatesng
from . import utils


def build_featurizer(patch_size, pool_size,
                     pool_stride, bias, patch_distribution,
                     num_filters, num_channels, seed,
                     filter_scale, X_train=None, filter_batch_size=2048,
                     max_threads=16, tot_patches=int(1e5)):
    dtype = 'float32'
    np.random.seed(seed)
    if (patch_distribution == 'empirical'):
        assert X_train is not None, "X_train must be provided when patch distribution == empirical"
        all_patches, idxs = utils.grab_patches(
            X_train, patch_size=patch_size, max_threads=max_threads, seed=seed, tot_patches=tot_patches)
        all_patches = utils.normalize_patches(
            all_patches, zca_bias=filter_scale)
        idxs = np.random.choice(
            all_patches.shape[0], num_filters, replace=False)
        filters = all_patches[idxs].astype(dtype)
        print("filters shape", filters.shape)
    elif (patch_distribution == 'gaussian'):
        filters = np.random.randn(
            num_filters, num_channels, patch_size, patch_size).astype(dtype) * filter_scale
        print("filters shape", filters.shape)
    elif (patch_distribution == 'laplace'):
        filters = np.random.laplace(loc=0.0, scale=filter_scale, size=(
            num_filters*num_channels*patch_size*patch_size)).reshape(num_filters, num_channels, patch_size, patch_size)
        filters = filters.astype('float32')
        print("filters shape", filters.shape)
    else:
        raise Exception(
            f"Unsupported patch distribution : {patch_distribution}")
    net = coatesng.BasicCoatesNgNet(filters, pool_size=pool_size, pool_stride=pool_stride,
                                    bias=bias, patch_size=patch_size, filter_batch_size=filter_batch_size)
    return net


def featurize(dataset, patch_size, patch_distribution, num_filters, pool_size, pool_stride, bias, filter_scale, seed, data_batchsize=1024, filter_batch_size=1024, zca=False):
    data = utils.load_dataset(dataset)
    gpu = torch.cuda.is_available()
    num_channels = 3
    dtype = 'float32'
    X_train = data['X_train']
    X_test = data['X_test']
    if zca:
        X_train, X_test = utils.preprocess(X_train, X_test)
    y_train = data['y_train']
    y_test = data['y_test']
    if ("cifar" in dataset):
        dataset = "cifar-10"
    elif ("mnist" in dataset):
        dataset = "mnist"
    else:
        raise Exception("unsupported dataset")
    featurizer = build_featurizer(patch_size, pool_size, pool_stride, bias, patch_distribution,
                                  num_filters, num_channels, seed, filter_scale, X_train, filter_batch_size)
    start = time.time()
    if not zca:
        assert X_train.dtype == np.uint8
        print("Scaling input by 255..")
        X_train = X_train.astype(dtype)
        X_test = X_test.astype(dtype)
        X_train /= 255.0
        X_test /= 255.0
    X_train_lift = coatesng.coatesng_featurize(
        featurizer, X_train.astype(dtype), data_batchsize=data_batchsize, gpu=gpu)
    X_test_lift = coatesng.coatesng_featurize(
        featurizer, X_test.astype(dtype), data_batchsize=data_batchsize, gpu=gpu)
    end = time.time()
    featurizer = featurizer.cpu()
    print(
        f"featurization complete, featurized {X_train.shape[0]} training points  and {X_test.shape[0]} testing points, {X_train_lift.shape[1]} output features, took {end - start} seconds")
    return X_train_lift, X_test_lift, y_train, y_test, featurizer.cpu()

def train_ls_model(X_train, y_train, reg):
    X_train = X_train.astype('float64')
    y = np.eye(np.max(y_train) + 1)[y_train]
    XTX = X_train.T.dot(X_train)
    XTy = X_train.T.dot(y)
    idxs = np.diag_indices(X_train.shape[1])
    XTX[idxs] += reg
    model = scipy.linalg.solve(XTX, XTy)
    XTX[idxs] -= reg
    return model

def train_ls_model_dual(X_train, y_train, reg):
    X_train = X_train.astype('float64')
    y = np.eye(np.max(y_train) + 1)[y_train]
    XXT = X_train.dot(X_train.T)
    idxs = np.diag_indices(X_train.shape[0])
    XXT[idxs] += reg
    model = scipy.linalg.solve(XXT, y)
    XXT[idxs] -= reg
    model = X_train.T.dot(model)
    return model

def eval_ls_model(model, X, y):
    y_pred = X.dot(model)
    return np.sum(np.argmax(y_pred, axis=1) == y)/y.shape[0]
