import dataclasses
import pickle
from tqdm import tqdm
import os

import numpy as np
import torch

from ..model import ModelImplementation
from ..model_family import ModelFamily
from ..learning_rule import LearningRuleImplementation

from .random_features_impls import utils
from .random_features_impls import featurize_train_model
from .random_features_impls import coatesng


@dataclasses.dataclass
class RandomFeaturesHyperparams:
    num_filters: int = None
    patch_size: int = 6
    pool_size: int = 15
    pool_stride: int = 6
    bias: float = 1.0
    filter_scale: float = 1e-3
    patch_distribution: str = 'empirical'
    regularizer: float = 1e-4
    zca: bool = False
    tot_patches: int = int(1e5)


@dataclasses.dataclass
class RandomFeaturesTrainingParams:
    seed: int = None
    batch_size: int = 256
    deterministic: bool = True
    device: str = None
    filter_batch_size: int = 1024
    max_threads: int = 16
    verbose: str = False


@dataclasses.dataclass
class RandomFeaturesEvalParams:
    batch_size: int = None
    filter_batch_size: int = None
    device: str = None


def resolve_device(device):
    return device or ('cuda:0' if torch.cuda.is_available() else 'cpu')


class RandomFeaturesLearningRule(LearningRuleImplementation):
    def __init__(self, hparams: RandomFeaturesHyperparams):
        self.hparams = hparams

    def learn(self, dataset,
              checkpointer, logger, params: RandomFeaturesTrainingParams):
        
        if dataset.examples is None or dataset.labels is None:
            raise ValueError("Random Features only supports in-memory datasets.")
        examples, labels = dataset.examples, dataset.labels

        # Resolve hyper params
        hp, tp = self.hparams, params
        if hp.num_filters is None:
            raise ValueError('Must set a value for num_filters!')
        torch.backends.cudnn.deterministic = tp.deterministic

        def printo(*args):
            if tp.verbose:
                print(*args)

        # arrange in BCHW format
        examples = examples.transpose((0, 3, 1, 2)).astype('float64') / 255.0

        printo('Doing preprocessing')
        if hp.zca:
            x_prep, zca_mat = utils.preprocess(examples)
        else:
            x_prep, zca_mat = examples, None

        printo('Building featurizer')
        featurizer = featurize_train_model.build_featurizer(
            hp.patch_size, hp.pool_size, hp.pool_stride,
            hp.bias, hp.patch_distribution, hp.num_filters, x_prep.shape[1],
            tp.seed, hp.filter_scale,
            X_train=x_prep, filter_batch_size=tp.filter_batch_size,
            max_threads=tp.max_threads, tot_patches=hp.tot_patches)

        device = resolve_device(tp.device)

        printo('Featurizing')
        x_lift = coatesng.coatesng_featurize(
            featurizer, x_prep.astype('float32'),
            data_batchsize=tp.batch_size, device=device)
        printo(f'Feature dimension is {np.prod(x_lift.shape[1:])}')

        printo('Computing least-squares fit of linear weights')
        if x_lift.shape[0] > x_lift.shape[1]:
            lin_weights = featurize_train_model.train_ls_model(
                x_lift, labels, hp.regularizer)
        else:
            lin_weights = featurize_train_model.train_ls_model_dual(
                x_lift, labels, hp.regularizer)

        printo('Done')
        return RandomFeaturesModel(
            zca_mat, featurizer.filters, lin_weights, hp, tp)


class RandomFeaturesModel(ModelImplementation):
    def __init__(self, zca_mat, filters, lin_weights,
                 hp: RandomFeaturesHyperparams,
                 tp: RandomFeaturesTrainingParams):
        # Setups architecture elements (why is this outside core code?)
        self.zca_mat, self.filters, self.lin_weights = zca_mat, filters, lin_weights
        self.hp, self.tp = hp, tp

    def predict(self, dataset, params: RandomFeaturesEvalParams):
        if dataset.examples is None:
            raise ValueError("Random features only supports in-memory dataset.")
        examples = dataset.examples
        hp, tp = self.hp, self.tp
        batch_size = params.batch_size or tp.batch_size
        device = resolve_device(params.device)

        featurizer = coatesng.BasicCoatesNgNet(
            self.filters, pool_size=hp.pool_size, pool_stride=hp.pool_stride,
            bias=hp.bias, patch_size=hp.patch_size,
            filter_batch_size=params.filter_batch_size or tp.filter_batch_size)

        examples = examples.transpose((0, 3, 1, 2)).astype('float64') / 255.0

        if self.hp.zca:
            x_prep = utils.preprocess(examples, global_ZCA=self.zca_mat)
        else:
            x_prep = examples

        x_lift = coatesng.coatesng_featurize(
            featurizer, x_prep.astype('float32'),
            data_batchsize=batch_size, device=device)

        scores = x_lift @ self.lin_weights

        if dataset.params.classes_to_exclude is not None:
            keep_columns = np.ones(scores.shape[1], dtype=bool)
            keep_columns[dataset.params.classes_to_exclude] = False
            
            scores = scores[:, keep_columns]
            preds = scores.argmax(-1)

            # Remap the predictions from the subsampled label space
            # to the original label space
            original_cls_idxs = np.arange(len(keep_columns))
            original_cls_idxs = original_cls_idxs[keep_columns]
            preds = original_cls_idxs[preds]
        else:
            preds = scores.argmax(-1)

        return preds, scores

    def serialize(self):
        model_dict = {'zca_mat': self.zca_mat,
                      'filters': self.filters,
                      'lin_weights': self.lin_weights,
                      'hp': self.hp, 'tp': self.tp}
        return pickle.dumps(model_dict)


class RandomFeaturesFamily(ModelFamily):
    def __init__(self, num_filters=None):
        self.num_filters = num_filters

    def learning_rule(self, hparams):
        if self.num_filters is not None:
            if hparams.num_filters is not None:
                assert hparams.num_filters == self.num_filters
            else:
                hparams.num_filters = self.num_filters
        return RandomFeaturesLearningRule(hparams)

    def deserialize_model(self, model_bytes):
        return RandomFeaturesModel(**pickle.loads(model_bytes))

    @property
    def rule_params(self):
        return RandomFeaturesHyperparams

    @property
    def train_params(self):
        return RandomFeaturesTrainingParams

    @property
    def eval_params(self):
        return RandomFeaturesEvalParams


def get_model_families():
    output = {'RandFeaturesCoatesNg': RandomFeaturesFamily()}
    output.update({f'RandFeaturesCoatesNg_2to{k}': RandomFeaturesFamily(
        num_filters=2 ** k) for k in range(16)})
    return output

