import io
import dataclasses

import torch
import torch.nn as nn
from torch.utils.data import DataLoader
from torchvision import transforms

from .. import model
from ..model_family import ModelFamily
from .rst_impls import wideresnet 
from ..learning_rule import LearningRuleImplementation
from .imagenet_models import ImageDataset, seed_process


@dataclasses.dataclass
class RSTModelFamilyRuleParams:
    adversarial_training: bool
    normalize_input: bool


@dataclasses.dataclass
class RSTModelFamilyTrainParams:
    pass


@dataclasses.dataclass
class RSTModelFamilyEvalParams:
    device: str = 'auto' # 'cuda:0' if cuda available else 'cpu'
    num_workers: int = 2
    eval_batch_size: int = 128


class RSTLearningRule(LearningRuleImplementation):
    def __init__(self, family, params):
        self.family = family
        self.hparams = params

    def learn(self, dataset, checkpointer, logger, params):
        raise NotImplementedError


class RSTModel(model.ModelImplementation):
    def __init__(self, model):
        self.model = model
    
    def predict(self, dataset, params):
        torch.backends.cudnn.benchmark = True
        device = params.device
        if device == 'auto':
            device = 'cuda:0' if torch.cuda.is_available() else 'cpu'
        testset = ImageDataset(dataset, transform=transforms.ToTensor())
        testloader = DataLoader(testset,
                                batch_size=params.eval_batch_size,
                                shuffle=False,
                                num_workers=params.num_workers)
        
        self.model = self.model.to(device)
        self.model = self.model.eval()
        logits = []
        for inputs, _ in testloader:
            inputs = inputs.to(device)
            with torch.no_grad():
                outputs = self.model(inputs)
            logits.append(outputs)

        logits = torch.cat(logits)

        if dataset.params.classes_to_exclude is not None:
            keep_columns = torch.ones(logits.shape[1])
            keep_columns[dataset.params.classes_to_exclude] = 0
            
            logits = logits[:, keep_columns.bool()]
            preds = logits.max(dim=1).indices

            # Remap the predictions from the subsampled label space
            # to the original label space
            original_cls_idxs = torch.arange(len(keep_columns))
            original_cls_idxs = original_cls_idxs[keep_columns.bool()]
            preds = original_cls_idxs[preds]
        else:
            preds = logits.max(dim=1).indices
        return preds.cpu().numpy(), logits.cpu().numpy()
    
    def serialize(self):
        raise NotImplementedError


class NormalizeInput(nn.Module):
    def __init__(self,
                 mean=(0.4914, 0.4822, 0.4465),
                 std=(0.2023, 0.1994, 0.2010)):
        super().__init__()

        self.register_buffer('mean', torch.Tensor(mean).reshape(1, -1, 1, 1))
        self.register_buffer('std', torch.Tensor(std).reshape(1, -1, 1, 1))

    def forward(self, x):
        return (x - self.mean) / self.std


class RSTModelFamily(ModelFamily):
    def __init__(self, family):
        self.family = family

    def learning_rule(self, params):
        return RSTLearningRule(self.family, params)
    
    def deserialize_model(self, model_bytes):
        if self.family == 'wideresnet':
            model_data = torch.load(io.BytesIO(model_bytes), map_location=torch.device('cpu'))
            assert model_data['num_classes'] == 10
            inner_model = wideresnet.WideResNet()
            # TODO: change the code so that deserialize_model gets the parameters and can check
            # normalize_input matching the flag in the learning rule here
            assert model_data['normalize_input']
            if model_data['normalize_input']:
                model = nn.Sequential(NormalizeInput(), inner_model)
            else:
                model = inner_model
            def strip_data_parallel(s):
                if s.startswith('module'):
                    return s[len('module.'):]
                else:
                    return s
            state_dict = {strip_data_parallel(k): v for k, v in model_data['state_dict'].items()}
            model.load_state_dict(state_dict)
            return RSTModel(model)
        else:
            raise ValueError(self.family)

    @property 
    def rule_params(self):
        return RSTModelFamilyRuleParams
    
    @property
    def train_params(self):
        return RSTModelFamilyTrainParams
    
    @property
    def eval_params(self):
        return RSTModelFamilyEvalParams


def get_model_families():
    return {'rst_wideresnet': RSTModelFamily('wideresnet')}