"""A collection of models from SKLearn."""
import dataclasses
import pickle

import numpy as np
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from sklearn.ensemble import AdaBoostClassifier, RandomForestClassifier
from sklearn.linear_model import LogisticRegression, RidgeClassifier, SGDClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import FunctionTransformer, StandardScaler
from sklearn import svm


from .. import model
from ..learning_rule import LearningRuleImplementation
from ..model_family import ModelFamily, NoParameters


class SKLearnModel(model.ModelImplementation):
    def __init__(self, pipeline_object):
        self.pipeline = pipeline_object

    def predict(self, dataset, params):
        if dataset.examples is None:
            raise ValueError("SKLearnModels can only be evaluated on in-memory datasets.")
        predictions = self.pipeline.predict(dataset.examples)
        return predictions, None

    def serialize(self):
        return pickle.dumps(self.pipeline)


def featurize(examples):
    """Featurize examples for SKLearn models."""
    examples = np.array(examples)
    if examples.dtype == "uint8":
        examples = examples.astype(np.float64) / 255.0
    return examples.reshape(len(examples), -1)


class SkLearnLearningRule(LearningRuleImplementation):
    def __init__(self, model_class, hparams):
        self.model_class = model_class
        self.hparams = hparams

    def learn(self, dataset, checkpointer, logger, params):
        if dataset.examples is None or dataset.labels is None:
            raise ValueError("SKLearnModels can only be trained on in-memory datasets.")
        model_kwargs = dataclasses.asdict(self.hparams)
        del model_kwargs["scale_features"]
        classifier = self.model_class(**model_kwargs)
        if self.hparams.scale_features:
            pipeline = make_pipeline(
                FunctionTransformer(featurize), StandardScaler(), classifier)
        else:
            pipeline = make_pipeline(FunctionTransformer(featurize), classifier)
        pipeline.fit(dataset.examples, dataset.labels)
        return SKLearnModel(pipeline)


class SKLearnModelFamily(ModelFamily):
    def __init__(self, model_class, hparam_class):
        self.model_class = model_class
        self.hparam_class = hparam_class

    def learning_rule(self, rule_params):
        return SkLearnLearningRule(self.model_class, rule_params)

    def deserialize_model(self, model_bytes):
        pipeline = pickle.loads(model_bytes)
        return SKLearnModel(pipeline)

    @property
    def rule_params(self):
        return self.hparam_class

    @property
    def train_params(self):
        return NoParameters

    @property
    def eval_params(self):
        return NoParameters


@dataclasses.dataclass
class KNNHyperparameters:
    """Parameters for K-Nearest Neighbors classifiers.

    For a more detailed description, see here:
        https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KNeighborsClassifier.html
    """

    n_neighbors: int = 3
    weights: str = "uniform"
    p: int = 2
    scale_features: bool = True


@dataclasses.dataclass
class SVMHyperparameters:
    """Parameters for Support Vector Classifiers

    For a more detailed description, see here:
        https://scikit-learn.org/stable/modules/generated/sklearn.svm.SVC.html
    """

    kernel: str = "linear"
    C: float = 1.0
    gamma: str = "scale"
    scale_features: bool = True


@dataclasses.dataclass
class RandomForestHyperparameters:
    """Parameters for Random Forest Classifiers

    For a more detailed description, see here:
        https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.RandomForestClassifier.html
    """

    n_estimators: int = 100
    criterion: str = "gini"
    max_depth: int = 10
    max_features: str = "auto"
    scale_features: bool = True


@dataclasses.dataclass
class LogisticRegressionHyperparameters:
    """Parameters for the logistic regression model.

    For a more detailed description, see here:
        https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LogisticRegression.html
    """

    penalty: str = "l2"
    C: float = 1.0
    max_iter: int = 1000
    scale_features: bool = True


@dataclasses.dataclass
class RidgeClassifierHyperparameters:
    """Parameters for a fast linear model.

    For a more detailed description, see here:
        https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.RidgeClassifier.html
    """

    alpha: float = 1.0
    normalize: bool = False
    solver: str = "auto"
    scale_features: bool = True
    tol: float = 1e-3


@dataclasses.dataclass
class QDAHyperparameters:
    """Parameters for quadratic discrimant analysis.

    For a more detailed description, see here:
        https://scikit-learn.org/stable/modules/generated/sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis.html
    """

    reg_param: float = 0.0
    scale_features: bool = True


@dataclasses.dataclass
class AdaBoostHyperparameters:
    """Parameters for adaboost models.

    For a more detailed description, see here:
        https://scikit-learn.org/stable/modules/generated/sklearn.ensemble.AdaBoostClassifier.html.
    """

    n_estimators: int = 50
    learning_rate: float = 1.0
    algorithm: str = "SAMME.R"
    scale_features: bool = True

@dataclasses.dataclass
class SGDClassifierHyperparameters:
    """Parameters for SGDClassifier.

    For a more detailed description, see here:
        https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.SGDClassifier.html#sklearn.linear_model.SGDClassifier
    """

    loss: str = "hinge"
    penalty: str = "l2"
    # The higher the value, the stronger the regularization
    alpha: float = 1e-4
    scale_features: bool = True


def get_model_families():
    return {
        "K_nearest_neighbors": SKLearnModelFamily(
            KNeighborsClassifier,
            KNNHyperparameters,
        ),
        "SVM": SKLearnModelFamily(
            svm.SVC,
            SVMHyperparameters,
        ),
        "RandomForest": SKLearnModelFamily(
            RandomForestClassifier,
            RandomForestHyperparameters,
        ),
        "LogisticRegression": SKLearnModelFamily(
            LogisticRegression,
            LogisticRegressionHyperparameters,
        ),
        "RidgeClassifier": SKLearnModelFamily(
            RidgeClassifier,
            RidgeClassifierHyperparameters,
        ),
        "QuadraticDiscrimantAnalysis": SKLearnModelFamily(
            QuadraticDiscriminantAnalysis,
            QDAHyperparameters,
        ),
        "AdaBoost": SKLearnModelFamily(
            AdaBoostClassifier,
            AdaBoostHyperparameters,
        ),
        "SGDClassifier": SKLearnModelFamily(
            SGDClassifier,
            SGDClassifierHyperparameters,
        ),
    }
