"""Basic tests for checkpoint functionality."""
from functools import partial
import os
import sys

import basicdb
import numpy as np
import psutil
import pytest

from testbed.mldb import MLDB
from testbed.dataset import Dataset
from testbed import learning_rule


def test_no_checkpoints(mldb, rng):
    mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2])).store()

    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    rule1.store()
    assert rule1.impl is not None

    model1 = rule1.learn(dataset='dset1')
    assert model1.learning_rule == rule1
    assert model1.model_family == 'debug_1'
    assert model1.train_completed
    assert len(mldb.get_models()) == 0
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))


def test_no_checkpoints_continuous_storing(mldb, rng):
    mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2])).store()

    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    rule1.store()
    assert rule1.impl is not None

    model1 = rule1.learn(dataset='dset1', store=True)
    assert model1.train_completed
    assert len(mldb.get_models()) == 1
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))


def test_two_checkpoints_continuous_storing(mldb, rng):
    mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2])).store()
    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    
    model1 = rule1.learn(dataset='dset1', store=True, num_checkpoints=2)
    assert model1.train_completed
    assert len(mldb.get_models()) == 1
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))

    assert len(model1.checkpoints) == 2
    
    def check_checkpoint(checkpoint, ii, checkpoint_model_stored=False):
        assert checkpoint.parent_model.uuid == model1.uuid
        assert checkpoint.is_stored
        assert checkpoint.sequence_number == ii
        assert checkpoint._model_bytes is None
        assert checkpoint.extra_info['step'] == ii
        c_model = checkpoint.as_model()
        assert checkpoint._model_bytes is None
        assert c_model.train_set.uuid == model1.train_set.uuid
        assert c_model.learning_rule.uuid == model1.learning_rule.uuid
        assert c_model.model_family == model1.model_family
        assert c_model.is_checkpoint
        assert c_model.parent_checkpoint == checkpoint
        assert len(c_model.checkpoints) == 0
        assert c_model.is_stored == checkpoint_model_stored
        assert np.array_equal(c_model.predict(examples=np.zeros(5), return_eval=False),
                              (rule1.params.output_value + ii + 1) * np.ones(5))
        assert checkpoint._model_bytes is not None
        checkpoint.release_model_bytes()
        assert checkpoint._model_bytes is None

    checkpoint1 = model1.checkpoints[0]
    check_checkpoint(checkpoint1, 0)
    checkpoint2 = model1.checkpoints[1]
    check_checkpoint(checkpoint2, 1)

    assert len(mldb.get_models()) == 1
    c_model_1 = checkpoint1.as_model()
    c_model_1.store()
    assert len(mldb.get_models()) == 2
    c_model_1_again = checkpoint1.as_model()
    c_model_1_again.store()
    assert len(mldb.get_models()) == 2
    c_model_2 = checkpoint2.as_model()
    c_model_2.store()
    assert len(mldb.get_models()) == 3
    assert set([x.uuid for x in mldb.get_models()]) == set([model1.uuid, c_model_1.uuid, c_model_2.uuid])
    model1.rename('model1')

    model1_again = mldb.model(name='model1')
    assert model1.uuid == model1_again.uuid
    [c1_again, c2_again] = model1_again.checkpoints
    check_checkpoint(c1_again, 0, checkpoint_model_stored=True)
    check_checkpoint(c2_again, 1, checkpoint_model_stored=True)


def test_two_checkpoints_store_afterwards(mldb, rng):
    mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2])).store()
    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    
    model1 = rule1.learn(dataset='dset1', store=False, num_checkpoints=2)
    assert model1.train_completed
    assert len(mldb.get_models()) == 0
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))

    assert len(model1.checkpoints) == 2
    
    def check_checkpoint(checkpoint, ii, stored=False, has_model_bytes=True):
        if stored:
            assert checkpoint.parent_model.uuid == model1.uuid
        else:
            assert checkpoint.parent_model == model1
        assert checkpoint.is_stored == stored
        assert (checkpoint._model_bytes is not None) == has_model_bytes
        assert checkpoint.sequence_number == ii
        assert checkpoint.extra_info['step'] == ii
        c_model = checkpoint.as_model()
        if stored:
            assert c_model.train_set.uuid == model1.train_set.uuid
        else:
            assert c_model.train_set == model1.train_set
        if stored:
            assert c_model.learning_rule.uuid == model1.learning_rule.uuid
        else:
            assert c_model.learning_rule == model1.learning_rule
        assert c_model.model_family == model1.model_family
        assert c_model.is_checkpoint
        assert c_model.parent_checkpoint == checkpoint
        assert len(c_model.checkpoints) == 0
        assert not c_model.is_stored
        assert np.array_equal(c_model.predict(examples=np.zeros(5), return_eval=False),
                            (rule1.params.output_value + ii + 1) * np.ones(5))

    checkpoint1 = model1.checkpoints[0]
    check_checkpoint(checkpoint1, 0)
    checkpoint2 = model1.checkpoints[1]
    check_checkpoint(checkpoint2, 1)

    model1.store()
    assert len(mldb.get_models()) == 1
    assert model1.is_stored
    assert checkpoint1.is_stored
    assert checkpoint2.is_stored
    assert len(model1.checkpoints) == 2
    check_checkpoint(checkpoint1, 0, stored=True)
    check_checkpoint(checkpoint2, 1, stored=True)
    
    model1.rename('model1')
    model1_again = mldb.model(name='model1')
    assert model1.uuid == model1_again.uuid
    [c1_again, c2_again] = model1_again.checkpoints
    check_checkpoint(c1_again, 0, stored=True, has_model_bytes=False)
    check_checkpoint(c2_again, 1, stored=True, has_model_bytes=False)


def test_checkpoint_not_stored_yet(mldb, rng):
    mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2])).store()
    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    model1 = rule1.learn(dataset='dset1', store=False, num_checkpoints=1)
 
    assert not model1.is_stored
    checkpoint1 = model1.checkpoints[0]
    c_model = checkpoint1.as_model()
    c_model.store()


@pytest.mark.slow
def test_memory_consumption(mldb, rng):
    mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2])).store()
    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    payload_size = 1e9
    model1 = rule1.learn(dataset='dset1', store=True, num_checkpoints=10, extra_payload_size=payload_size)
    assert model1.train_completed
    assert psutil.Process(os.getpid()).memory_info().rss <= 2.1 * payload_size