"""Basic tests for dataset functionality."""
import pytest
import sys

import basicdb
import numpy as np

from testbed import dataset
from testbed.mldb import MLDB

def dset_names(mldb):
    return set(dset.name for dset in mldb.get_datasets())

def test_mldb(tmp_path):
    """Basic test of mldb functionality."""
    db1 = basicdb.BasicDB(sql_string="sqlite:///:memory:",
                          stash_rootdir=tmp_path,
                          namespace="mockdb_1")
    db2 = basicdb.BasicDB(sql_string="sqlite:///:memory:",
                          stash_rootdir=tmp_path,
                          namespace="mockdb_2")

    mldb1 = MLDB(db1, data_cache_dir=tmp_path)
    mldb2 = MLDB(db2, data_cache_dir=tmp_path)
    
    mldb1.dataset(name="dset1", examples=[0]).store()

    assert dset_names(mldb1) == set(["dset1"])
    assert dset_names(mldb2) == set()


def test_dataset(mldb, rng):
    """Simple tests of dataset functionality."""
    examples = rng.randn(531, 32, 32, 4)
    labels = rng.randint(2, size=(531, ))

    cifar10 = mldb.dataset(name="cifar10", examples=examples, labels=labels)
    
    assert len(cifar10) == 531

    assert dset_names(mldb) == set()
    cifar10.store()
    assert dset_names(mldb) == set(["cifar10"])
    with pytest.raises(ValueError):
        cifar10.store(assert_not_stored_yet=True)
    
    # Can't create a dataset with examples/labels if the dataset is in the DB
    with pytest.raises(ValueError):
        mldb.dataset(name="cifar10", examples=examples, labels=labels)
    
    from_cache_cifar10 = mldb.dataset(name="cifar10")
    assert np.allclose(cifar10.examples, from_cache_cifar10.examples)
    assert np.allclose(cifar10.labels, from_cache_cifar10.labels)
    assert len(from_cache_cifar10) == len(cifar10)

    # Can't load a dataset that doesn't exist
    with pytest.raises(AssertionError):
        mldb.dataset(name="cifar222")

    examples2 = rng.randn(127, 32, 32, 4)
    labels2 = rng.randint(2, size=(127, ))
    cifar10p2 = mldb.dataset(name="cifar10p2", examples=examples2, labels=labels2)
    cifar10p2.store()
    assert dset_names(mldb) == set(["cifar10", "cifar10p2"])
    from_cache_cifar10p2 = mldb.dataset(name="cifar10p2")
    assert np.allclose(cifar10p2.examples, from_cache_cifar10p2.examples)
    assert np.allclose(cifar10p2.labels, from_cache_cifar10p2.labels)

    ## Passing examples by reference
    #example_refs = [(cifar10.name, 0), (cifar10p2.name, 3)]
    #mixed_dset = mldb.dataset(name="mixed_dset", example_refs=example_refs)
    #assert np.allclose(mixed_dset.examples[0], cifar10.examples[0])
    #assert np.allclose(mixed_dset.examples[1], cifar10p2.examples[3])
    #    
    #mixed_dset.store()
    #
    #from_cache_mixed = mldb.dataset(name="mixed_dset")
    #assert np.allclose(from_cache_mixed.examples[0], cifar10.examples[0])
    #assert np.allclose(from_cache_mixed.examples[1], cifar10p2.examples[3])

    #with pytest.raises(ValueError):
    #    mldb.dataset(name="mixed_dset", example_refs=example_refs)

    #with pytest.raises(ValueError):
    #    mldb.dataset(name="new_dset", example_refs=example_refs, examples=examples, labels=labels)

    # Create a dataset without a name
    no_name = mldb.dataset(examples=examples, labels=labels)
    no_name.store()
    from_cache = mldb.dataset(db_obj=no_name.db_obj)
    assert np.allclose(from_cache.examples, no_name.examples)

    # Create a dataset without labels
    no_labels = mldb.dataset(examples=examples)
    no_labels.store()
    from_cache = mldb.dataset(db_obj=no_labels.db_obj)
    assert np.allclose(from_cache.examples, no_labels.examples)
    assert from_cache.labels is None


def test_subsample_dset(mldb, rng):
    """Test subsampling functionality."""
    examples = rng.randn(200, 32, 32, 4)
    labels = np.array([0] * 100 + [1] * 100)

    full_dset = mldb.dataset(name="full_dset", examples=examples, labels=labels)
    full_dset.store()

    subsampled = full_dset.subsample(size=100, seed=1234)
    assert subsampled.examples.shape == (100, 32, 32, 4)
    # Check class balance
    assert subsampled.labels.sum() == 50
    
    # Storing by reference not currently implemented
    # Since full_dset is stored, the sample should be stored by reference
    #assert subsampled.example_refs is not None
    #assert set([p.name for p in subsampled.parents]) == set(["full_dset"])
    #assert subsampled.extra["seed"] == 1234
    #assert subsampled.extra["size"] == 100
    #assert subsampled.extra["derived_by"] == "subsampling"
    
    # Ensure determinism
    resampled = full_dset.subsample(size=100, seed=1234)
    assert np.allclose(subsampled.examples, resampled.examples)
    assert np.allclose(subsampled.labels, resampled.labels)
    resampled.store()
    from_cache = mldb.dataset(name=resampled.name)
    assert np.allclose(from_cache.examples, resampled.examples)
    assert np.allclose(from_cache.labels, resampled.labels)
        
    # Can't sampled a non-class balanced dataset (or one that's too big)
    with pytest.raises(ValueError):
        full_dset.subsample(size=101, seed=1234)
    with pytest.raises(ValueError):
        full_dset.subsample(size=300, seed=1234)
    with pytest.raises(ValueError):
        full_dset.subsample(size=1, seed=1234)

    subsubsampled = subsampled.subsample(size=50, seed=1234)
    assert subsubsampled.examples.shape == (50, 32, 32, 4)
    assert subsubsampled.labels.sum() == 25
    #assert subsubsampled.example_refs is not None
    #assert set([p.name for p in subsubsampled.parents]) == set([subsampled.name])
    
    no_labels = mldb.dataset(examples=examples)
    subsampled = no_labels.subsample(size=50, seed=1234)
    assert len(subsampled.examples) == 50


    #ref_dset, idx = subsubsampled.example_refs[0]
    #assert np.allclose(mldb.dataset(name=ref_dset.name).examples[idx], subsubsampled.examples[0])
    #assert np.allclose(mldb.dataset(name=ref_dset.name).labels[idx], subsubsampled.labels[0])
    
    # Storing the final node in the chain stores all ancestors.
    #subsubsampled.store()
    #assert dset_names(mldb) == set(["full_dset", subsampled.name, subsubsampled.name])
    #assert subsampled.is_stored

    #from_cache = mldb.dataset(name=subsampled.name)
    #assert set([p.name for p in from_cache.parents]) == set(["full_dset"])
    #assert np.allclose(from_cache.examples, subsampled.examples)
    #assert np.allclose(from_cache.labels, subsampled.labels)

    #sub_from_cache = mldb.dataset(name=subsubsampled.name)
    #assert set([p.name for p in sub_from_cache.parents]) == set([subsampled.name])
    #assert np.allclose(sub_from_cache.examples, subsubsampled.examples)
    #assert np.allclose(sub_from_cache.labels, subsubsampled.labels)


#def test_merge(mldb, rng):
#    """Test dataset merging functionality."""
#    examples1 = rng.randn(100, 32, 32, 4)
#    labels1 = rng.randint(2, size=(100, ))
#    examples2 = rng.randn(100, 32, 32, 4)
#    labels2 = rng.randint(2, size=(100, ))
#
#    dset1 = mldb.dataset(name="dset1", examples=examples1, labels=labels1)
#    dset2 = mldb.dataset(name="dset2", examples=examples2, labels=labels2)
#    dset1.store()
#    dset2.store()
#
#    merged = dset1.merge(dset2)
#    assert merged.examples.shape == (200, 32, 32, 4)
#    assert merged.labels.shape == (200, )
#    assert np.allclose(merged.examples[99], examples1[99])
#    assert np.allclose(merged.examples[100], examples2[0])
#    assert set(p.name for p in merged.parents) == set(["dset1", "dset2"])
#    
#    merged.store() 
#    from_cache = mldb.dataset(name=merged.name)
#    assert from_cache.examples.shape == (200, 32, 32, 4)
#    assert from_cache.labels.shape == (200, )
#    assert np.allclose(from_cache.examples[99], examples1[99])
#    assert np.allclose(from_cache.examples[100], examples2[0])
#    assert set(p.name for p in from_cache.parents) == set(["dset1", "dset2"])
#    
#    # Ensure merge order doesn't matter
#    with pytest.raises(AssertionError):
#        switched = dset2.merge(dset1)
#        switched.store(assert_not_stored_yet=True)
#
#def test_combined_ops(mldb, rng):
#    """Test integration of several operations."""
#    examples1 = rng.randn(100, 32, 32, 4)
#    labels1 = rng.randint(2, size=(100, ))
#    examples2 = rng.randn(100, 32, 32, 4)
#    labels2 = rng.randint(2, size=(100, ))
#
#    dset1 = mldb.dataset(name="dset1", examples=examples1, labels=labels1)
#    dset2 = mldb.dataset(name="dset2", examples=examples2, labels=labels2)
#
#    subsampled_1 = dset1.subsample(size=50, seed=1234)
#    subsampled_2 = dset2.subsample(size=50, seed=1234)
#    
#    merged = subsampled_1.merge(subsampled_2)
#
#    assert dset_names(mldb) == set()
#
#    merged.store()
#
#    assert dset_names(mldb) == set(["dset1", "dset2", subsampled_1.name, subsampled_2.name, merged.name])
#    
#    assert set([p.name for p in merged.parents]) == set([subsampled_1.name, subsampled_2.name])
#    parents = mldb.db.get(rel_second=subsampled_1.db_obj, rel_type="child_dset")
#    assert set([p.name for p in parents]) == set(["dset1"])
#
#    sampled_again = merged.subsample(size=50, seed=1234)
#    test_idx = 27
#    example_dset, idx = sampled_again.example_refs[test_idx]
#
#    assert np.allclose(mldb.dataset(name=example_dset.name).examples[idx], sampled_again.examples[test_idx])
#
#
#def test_perturb(mldb, rng):
#    """Test functionality for perturbing the dataset."""
#
#    def add_gaussian_noise(examples, seed, sigma=1.0):
#        rng = np.random.RandomState(seed)
#        noise = sigma * rng.randn(*examples.shape)
#        return examples + noise
#
#    def replace_with_constant(examples, constant=1.0):
#        return constant * np.ones_like(examples)
#
#    gaussian_noise = dataset.Perturbation(
#        name="gaussian_noise",
#        example_perturb_fcn=add_gaussian_noise,
#    )
#
#    replace_constant = dataset.Perturbation(
#        name="constant",
#        example_perturb_fcn=replace_with_constant,
#    )
#        
#    examples = rng.randn(100, 32, 32, 4)
#    labels = rng.randint(2, size=(100, ))
#
#    base_dset = mldb.dataset(name="base_dset", examples=examples, labels=labels)
#    base_dset.store()
#
#    perturbed = base_dset.perturb(gaussian_noise, seed=1234)
#    assert perturbed.examples.shape == (100, 32, 32, 4)
#    assert perturbed.labels.shape == (100, )
#    assert np.allclose(perturbed.labels, labels)
#    assert set(p.name for p in perturbed.parents) == set(["base_dset"])
#    
#    # Ensure determinism
#    perturbed_2 = base_dset.perturb(gaussian_noise, seed=1234)
#    assert np.allclose(perturbed.examples, perturbed_2.examples)
#    assert np.allclose(perturbed.labels, perturbed_2.labels)
#
#
#    perturbed = base_dset.perturb(replace_constant)
#    assert np.allclose(perturbed.examples, 1.0)
#
#    perturbed = base_dset.perturb(replace_constant, constant=2.0)
#    perturbed.store()
#    from_cache = mldb.dataset(name=perturbed.name)
#    assert from_cache.extra["params"]["constant"] == 2.0
#    assert np.allclose(from_cache.examples, 2.0)
#
#
#def test_subset(mldb, rng):
#    examples = rng.randn(100, 32, 32, 4)
#    labels = rng.randint(2, size=(100, ))
#
#    base_dset = mldb.dataset(name="base_dset", examples=examples, labels=labels)
#
#    subset = base_dset.subset(name="subset", indices=[0, 10, 20])
#    subset.store()
#
#    assert dset_names(mldb) == set(["base_dset", "subset"])
#    assert np.allclose(subset.examples[1], examples[10])
#    assert np.allclose(subset.labels[2], labels[20])
#    
#    # Can't create a subset with the same name
#    with pytest.raises(ValueError):
#        base_dset.subset(name="subset", indices=[0, 1, 2])
#
#def test_intersect(mldb, rng):
#    examples = rng.randn(100, 32, 32, 4)
#    labels = rng.randint(2, size=(100, ))
#    base = mldb.dataset(name="base", examples=examples, labels=labels)
#    base.store()
#    
#    refs1 = [("base", 0), ("base", 1)]
#    refs2 = [("base", 2), ("base", 3)]
#    overlap_refs = [("base", 1), ("base", 2)]
#
#    dset1 = mldb.dataset(name="dset1", example_refs=refs1)
#    dset2 = mldb.dataset(name="dset2", example_refs=refs2)
#    overlap = mldb.dataset(name="dset_overlap", example_refs=overlap_refs)
#
#    assert dset1.intersect(dset2) == set()
#    assert dset1.intersect(overlap) == set([("base", 1)])
#    
#    # Can't check intersection is not both stored by ref
#    with pytest.raises(NotImplementedError):
#        base.intersect(dset1)
#
#
#def train_test_split(mldb, rng):
#    examples = rng.randn(100, 32, 32, 4)
#    labels = rng.randint(2, size=(100, ))
#    base = mldb.dataset(name="base", examples=examples, labels=labels)
#
#    train_test_split = mldb.train_test_split(base, train_size=25, seed=1234)
#    assert len(train_test_split.train.examples) == 25
#    assert len(train_test_split.test.examples) == 75
#    train_refs = set(train_test_split.train.example_refs)
#    test_refs = set(train_test_split.test.example_refs) 
#    assert train_test_split.train.intersect(train_test_split.test) == set()
#
#
#    train_test_split2 = mldb.train_test_split(base, train_size=25, seed=1234)
#    assert np.allclose(train_test_split.train.examples, train_test_split2.train.examples)
#    assert np.allclose(train_test_split.train.labels, train_test_split2.train.labels)
#
#    train_test_split.train.store()
#    train_test_split.test.store()
#
#    assert dset_names(mldb) == set(["base_dset", train_test_split.train.name, train_test_split.test.name])

def test_name_changes(mldb, rng):
    examples = [0, 1, 2, 3]
    labels = [0, 1, 2, 3]
    original = mldb.dataset(name="original", examples=examples, labels=labels)
    original.store()
    original.rename("new_name")
    assert original.name == "new_name"
    assert dset_names(mldb) == set(["new_name"])
    from_cache = mldb.dataset(name="new_name")
    assert np.allclose(from_cache.examples, examples)
    assert np.allclose(from_cache.labels, labels)


#def test_references(mldb, rng):
#    examples = np.array([0, 1, 2, 3])
#    labels = np.array([0, 1, 2, 3])
#    base = mldb.dataset(name="base", examples=examples, labels=labels)
#    base.store()
#    
#    example_refs = [(base.db_obj.uuid, 0), (base, 2), ("base", 1)]
#    referenced = mldb.dataset(name="referenced", example_refs=example_refs)
#    referenced.store()
#    
#    # Change the internal name of the referenced dataset
#    base.rename("new_name")
#    
#    from_cache = mldb.dataset(name="referenced")
#    assert np.allclose(from_cache.examples, [0, 2, 1])
#    assert np.allclose(from_cache.labels, [0, 2, 1])
#    
#    no_name_unstored = mldb.dataset(examples=examples, labels=labels)
#    no_name_stored = mldb.dataset(examples=examples, labels=labels)
#    no_name_stored.store()
#
#    example_refs = [(no_name_unstored, 1), (no_name_stored, 0)]
#    composed = mldb.dataset(example_refs=example_refs)
#    composed.store()
#    from_cache = mldb.dataset(db_obj=composed.db_obj)
#    assert np.allclose(from_cache.examples, [1, 0])
#    assert np.allclose(from_cache.labels, [1, 0])
#    
#    no_labels = mldb.dataset(examples=examples)
#    example_refs = [(no_name_stored, 2), (no_labels, 0)]
#    no_labels_ref = mldb.dataset(example_refs=example_refs)
#    no_labels_ref.store()
#    from_cache = mldb.dataset(db_obj=no_labels_ref.db_obj)
#    assert from_cache.labels is None
#
#
#def test_indices_by_class(mldb, rng):
#    examples = np.array([0, 1, 2, 3, 4])
#    labels = np.array([0, 1, 1, 0, 0])
#    dset = mldb.dataset(name="dset", examples=examples, labels=labels)
#    assert dset.get_indices_by_class() == {0: [0, 3, 4],
#                                           1: [1, 2]}
#
#
#def test_stratify(mldb, rng):
#    examples = np.array([8, 7, 1, 0, 4, 6, 3, 2])
#    labels = np.array([0, 1, 1, 0, 0, 0, 1, 1])
#    dset = mldb.dataset(name='dset', examples=examples, labels=labels)
#    dset1, dset2 = dset.stratify(lambda x: x, 2, 'stratify_')
#    assert set(dset1.examples) == set([0, 1, 2, 4])
#    assert set(dset2.examples) == set([3, 6, 7, 8])
#    
#    dset3, dset4 = dset.stratify(lambda x: x, 2, 'stratify2_', ascending=False)
#    assert set(dset4.examples) == set([0, 1, 2, 4])
#    assert set(dset3.examples) == set([3, 6, 7, 8])


#def test_no_database(mldb):
#    examples = np.array([0, 1, 2, 3])
#    labels = np.array([0, 1, 0, 1])
#    base = dataset.Dataset(examples=examples, labels=labels)
#    subsampled = base.subsample(size=2, seed=1234)
#    
#    example_refs = [(base, 0), (subsampled, 1)]
#    referenced = dataset.Dataset(example_refs=example_refs)
#
#    assert np.allclose(referenced.examples[0], base.examples[0])
#    assert np.allclose(referenced.examples[1], subsampled.examples[1])
#    assert np.allclose(referenced.labels[0], base.labels[0])
#    assert np.allclose(referenced.labels[1], subsampled.labels[1])
#
#    with pytest.raises(ValueError):
#        referenced.store()
#
#    with pytest.raises(ValueError):
#        dataset.Dataset(name="no_db_name")
#    
#    dset = mldb.dataset(name="new", examples=examples, labels=labels)
#    dset.store()
#    with pytest.raises(ValueError):
#        dataset.Dataset(db_obj=dset.db_obj)
