"""Basic tests for dataset functionality."""
import pytest
import sys

import numpy as np

from testbed import dataset
from testbed import model
from testbed.evaluation import Evaluation, InapplicableStatisticError


def test_evaluation(mldb, rng):
    """Simple tests of evaluation functionality."""
    examples = rng.randn(20, 3)
    labels = np.array([0] * 10 + [1] * 10)

    dataset = mldb.dataset(examples=examples, labels=labels)
    
    rule = mldb.learning_rule(family='debug_1', name='lr1', params={"output_value": 1})
    rule.store()
    model = rule.learn(dataset=dataset)
    evaluation = model.predict(dataset=dataset)

    assert np.allclose(evaluation.predictions, 1)
    assert np.allclose(evaluation.accuracy, 0.5)
    
    evaluation.store()

    num_correct, test_set_size = evaluation.get_statistic('num_correct_and_size')
    assert num_correct / test_set_size == evaluation.accuracy
    assert evaluation.db_obj.computed_statistics['num_correct_and_size'] == (num_correct, test_set_size)

    from_cache = Evaluation(mldb=mldb, db_obj=evaluation.db_obj.uuid)
    assert np.allclose(from_cache.predictions, 1)
    assert np.allclose(from_cache.accuracy, 0.5)
    num_correct, test_set_size = from_cache.get_statistic('num_correct_and_size')
    assert from_cache.db_obj.computed_statistics['num_correct_and_size'] == (num_correct, test_set_size)
    assert from_cache.test_set.uuid == dataset.db_obj.uuid
    assert from_cache.model.uuid == model.db_obj.uuid
    assert len(from_cache.acc_conf_int) == 2
    assert from_cache.acc_conf_int[0] > 0.1 and from_cache.acc_conf_int[1] < 0.9

    no_labels = mldb.dataset(examples=examples)
    rule = mldb.learning_rule(family='debug_1', name='lr1', params={"output_value": 1})
    model = rule.learn(dataset=no_labels)
    evaluation = model.predict(dataset=no_labels)
    assert np.allclose(evaluation.predictions, 1)
    with pytest.raises(InapplicableStatisticError):
        _ = evaluation.accuracy
    evaluation.store()
    from_cache = Evaluation(mldb=mldb, db_obj=evaluation.db_obj.uuid)
    assert np.allclose(from_cache.predictions, 1)
    with pytest.raises(InapplicableStatisticError):
        _ = from_cache.accuracy
    with pytest.raises(InapplicableStatisticError):
        _ = from_cache.acc_conf_int


def test_cached_statistic_computation(mldb, rng):
    examples = rng.randn(20, 3)
    labels = np.array([0] * 10 + [1] * 10)
    dataset = mldb.dataset(examples=examples, labels=labels)
    
    rule = mldb.learning_rule(family='debug_1', name='lr1', params={"output_value": 1})
    rule.store()
    model = rule.learn(dataset=dataset)
    evaluation = model.predict(dataset=dataset)

    assert np.allclose(evaluation.predictions, 1)
    assert np.allclose(evaluation.accuracy, 0.5)
    evaluation.store()
    
    from_cache = Evaluation(mldb=mldb, db_obj=evaluation.db_obj.uuid)
    assert np.allclose(from_cache.accuracy, 0.5)
    assert from_cache._test_set is None    # Make sure it didn't need to load the test set
    assert from_cache._model is None       # Make sure it didn't need to load the model


def test_lazy_statistic_computation(mldb, rng):
    examples = rng.randn(20, 3)
    labels = np.array([0] * 10 + [1] * 10)
    dataset = mldb.dataset(examples=examples, labels=labels)
    
    rule = mldb.learning_rule(family='debug_1', name='lr1', params={"output_value": 1})
    rule.store()
    model = rule.learn(dataset=dataset)
    evaluation = model.predict(dataset=dataset)
    evaluation.store(statistics_to_store=[])
    
    from_cache = Evaluation(mldb=mldb, db_obj=evaluation.db_obj.uuid)
    assert np.allclose(from_cache.accuracy, 0.5)
    assert from_cache._test_set is not None    # Make sure it loaded the test set
    assert from_cache._model is None       # Make sure it didn't need to load the model


def test_eager_computation(mldb, rng):
    examples = rng.randn(20, 3)
    labels = np.array([0] * 10 + [1] * 10)
    dataset = mldb.dataset(examples=examples, labels=labels)
    
    rule = mldb.learning_rule(family='debug_1', name='lr1', params={"output_value": 1})
    rule.store()
    model = rule.learn(dataset=dataset)
    evaluation = model.predict(dataset=dataset)
    evaluation.store()
    
    from_cache = Evaluation(mldb=mldb, db_obj=evaluation.db_obj.uuid)
    assert np.allclose(from_cache.accuracy, 0.5)
    assert from_cache._test_set is None    # Make sure it didn't need to load the test set
    assert from_cache._model is None       # Make sure it didn't need to load the model


def test_params_stored(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1')
    dset1 = mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset1, name='model1')
    mldb.dataset(name='dset2', examples=np.array([1, 1, 1]), labels=np.array([2, 2, 2])).store()
    eval1 = model1.predict(dataset='dset2', output_value=5)
    assert np.array_equal(eval1.predictions, 5 * np.ones(3))
    assert eval1.params.output_value == 5
    assert len(mldb.get_evaluations()) == 0
    eval1.store()
    assert len(mldb.get_evaluations()) == 1
    eval1_again = mldb.get_evaluations()[0]
    assert eval1_again.params.output_value == 5


def test_properties(mldb):
    rule = mldb.learning_rule(family='debug_1')
    dset = mldb.dataset(name='dset', examples=np.array([1, 1]),
                         labels=np.array([2, 2]))
    model = rule.learn(dataset=dset, name='model')
    eval = model.predict(dataset=dset, output_value=5)
    # let's see that the model and test set properties are what we expected
    assert eval.model == model
    assert eval.test_set == dset

    # and let's make sure that storing and retrieving doesn't change a thing
    eval.store()
    retrieved_eval = mldb.get_evaluations()[0]
    assert retrieved_eval.model.uuid == model.uuid
    assert retrieved_eval.test_set.uuid == dset.uuid
    assert type(retrieved_eval.model) == type(model)
    assert type(retrieved_eval.test_set) == type(dset)
