"""Testing  neural learning rule functionality."""
import pytest
import sys

import basicdb
import numpy as np
import torch
import pathlib

from testbed.dataset import Dataset
from testbed import learning_rule
from testbed import init_db
from testbed.mldb import MLDB


@pytest.mark.gpu
def test_cifar(mldb, rng):
    seed = rng.randint(0, 99999)
    init_db.add_cifar10(mldb)
    
    rule = mldb.learning_rule(family='imgnt_resnet18', pretrained=True, num_classes=10, 
                            input_size=(32, 32), epochs=5, lr=0.01, milestones=(10,))
    rule.store()

    # Check learning rule lookups work correctly
    rule2 = mldb.learning_rule(family='imgnt_resnet18', pretrained=True, num_classes=10, 
                            input_size=(32, 32), epochs=5, lr=0.01, milestones=(10,))
    rule2.store()

    assert rule2.uuid == rule.uuid

    # test that network actually learns
    model = rule.learn(dataset='cifar10-train', name='model', cpu_workers=2)
    evals1 = model.predict(dataset='cifar10-test')
    assert evals1.accuracy >= 0.7

    # test model serializing & deserializing
    model.store()
    model_again = mldb.model(name='model')
    evals3 = model_again.predict(dataset='cifar10-test')
    assert np.array_equal(evals1.predictions, evals3.predictions)

    # test checkpointing works
    cifar10_train = mldb.dataset(name='cifar10-train').subsample(size=10, seed=seed)
    
    rule = mldb.learning_rule(family='imgnt_resnet18', pretrained=False, num_classes=10, input_size=(32, 32), epochs=30, lr=0.01)
    model = rule.learn(dataset=cifar10_train, cpu_workers=2, additional_checkpoint_epochs=[*range(30)])

    assert len(model.checkpoints) == 30


@pytest.mark.gpu
def test_wilds(tmp_path, rng):
    # create mldb with cache dir corresponding to actual file cache
    db = basicdb.BasicDB(sql_string="sqlite:///:memory:",
                        stash_rootdir=tmp_path,
                        namespace="mockdb")
    mldb = MLDB(db, data_cache_dir=(pathlib.Path(__file__) / "../../shared_mldb_root/data_cache").resolve())
    init_db.add_wilds(mldb)

    
    rule = mldb.learning_rule(family='imgnt_resnet18', pretrained=True, num_classes=2, 
                            input_size=(224, 224), epochs=5, lr=0.01)

    # test that network actually learns
    model = rule.learn(dataset='Waterbirds-train', cpu_workers=4)
    evals1 = model.predict(dataset='Waterbirds-test')
    assert evals1.accuracy >= 0.7
