"""Basic tests for learning rule functionality."""
import pytest
import sys

import basicdb
import numpy as np

from testbed.dataset import Dataset
from testbed import learning_rule
from testbed.mldb import MLDB
from testbed.model_families import debug_models


def rule_names(mldb):
    return set(x.name for x in mldb.get_learning_rules())


def test_mldb(tmp_path):
    """Basic test of mldb functionality."""
    db1 = basicdb.BasicDB(sql_string="sqlite:///:memory:",
                          stash_rootdir=tmp_path,
                          namespace="mockdb_1")
    db2 = basicdb.BasicDB(sql_string="sqlite:///:memory:",
                          stash_rootdir=tmp_path,
                          namespace="mockdb_2")

    mldb1 = MLDB(db1, data_cache_dir=tmp_path)
    mldb2 = MLDB(db2, data_cache_dir=tmp_path)

    mldb1.learning_rule(family='debug_1')
    assert rule_names(mldb1) == set()
    assert rule_names(mldb2) == set()
    
    mldb1.learning_rule(family='debug_1',
                        params={'param_a': 5, 'param_b': 'batch_norm'},
                        name='my_favorite_lr').store()
    assert rule_names(mldb1) == set(['my_favorite_lr'])
    assert rule_names(mldb2) == set()


def test_learning_rule_retrieval(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    rule1.store()
    assert rule1.db_obj is not None
    assert rule1.is_stored
    assert rule_names(mldb) == set(['lr1'])

    rule1_again = mldb.learning_rule(family='debug_1', name='lr1')
    assert rule1_again.db_obj is not None
    assert rule1.db_obj.uuid == rule1_again.db_obj.uuid
    assert rule1_again.is_stored

    with pytest.raises(ValueError):
        rule1_again = mldb.learning_rule(family='debug_1', name='NOT lr')
    
    rule2 = mldb.learning_rule(family='debug_1',
                               name='lr2',
                               params={'param_a': 2})
    assert rule2.db_obj is None
    assert not rule2.is_stored
    rule2.store()
    assert rule2.is_stored
    assert rule_names(mldb) == set(['lr1', 'lr2'])
    
    rule1_yet_again = mldb.learning_rule(family='debug_1',
                                         params={'param_b': 1.0})
    assert rule1_yet_again.db_obj is not None
    assert rule1.db_obj.uuid == rule1_yet_again.db_obj.uuid
    assert rule1_yet_again.is_stored
    
    rule2_again = mldb.learning_rule(family='debug_1',
                                     params={'param_a': 2, 'param_b': 1.0})
    assert rule2_again.is_stored
    assert rule2_again.db_obj.uuid == rule2.db_obj.uuid
    assert rule2.params.param_a == 2
    assert rule2.params.param_b == 1.0
    
    with pytest.raises(ValueError):
        rule3 = mldb.learning_rule(family='debug_2', name='lr1')

    rule3 = mldb.learning_rule(family='debug_2', name='lr3')
    assert not rule3.is_stored
    rule3.store()
    assert rule3.is_stored
    assert rule_names(mldb) == set(['lr1', 'lr2', 'lr3'])
    
    rule4 = mldb.learning_rule(family='debug_2', params={'param_b': 3.0})
    assert not rule4.is_stored
    rule4.store()
    assert rule4.is_stored
    assert rule_names(mldb) == set(['lr1', 'lr2', 'lr3', str(rule4.db_obj.uuid)])

    rule4_again = mldb.learning_rule(family='debug_2', params={'param_b': 3.0})
    assert rule4_again.is_stored
    assert rule4_again.db_obj.uuid == rule4.db_obj.uuid

    rule1_by_name = mldb.learning_rule(name='lr1')
    assert rule1_by_name.db_obj.uuid == rule1.db_obj.uuid


def test_learning_rule_learning_from_examples(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    rule1.store()
    assert rule1.impl is not None

    model1 = rule1.learn(examples=np.array([0, 1, 2]), labels=np.array([0, 0, 1]))
    assert model1.learning_rule == rule1
    assert model1.model_family == 'debug_1'

    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))
    
    model2 = rule1.learn(examples=np.array([0, 1, 2]), labels=np.array([0, 0, 1]))
    assert model1 != model2
    assert np.array_equal(model2.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))
    
    rule2 = mldb.learning_rule(family='debug_1', params={'output_value': 12})
    model3 = rule2.learn(examples=np.array([1]), labels=np.array([2]))
    assert np.array_equal(model3.predict(examples=np.zeros(5), return_eval=False),
                          rule2.params.output_value * np.ones(5))
    assert not np.array_equal(model3.predict(examples=np.zeros(5), return_eval=False),
                              rule1.params.output_value * np.ones(5))


def test_learning_rule_learning_from_dataset(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    rule1.store()
    assert rule1.impl is not None

    dset = Dataset(examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset)
    assert model1.learning_rule == rule1
    assert model1.model_family == 'debug_1'
    assert model1.name is None

    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))


def test_learning_rule_learning_from_dataset_name(mldb, rng):
    mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2])).store()

    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    rule1.store()
    assert rule1.impl is not None

    model1 = rule1.learn(dataset='dset1')
    assert model1.learning_rule == rule1
    assert model1.model_family == 'debug_1'

    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))


def test_learning_rule_learning_with_name(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1', name='lr1', output_value=5)
    rule1.store()
    assert rule1.impl is not None

    dset = Dataset(examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset, name='new_model')
    assert model1.learning_rule == rule1
    assert model1.model_family == 'debug_1'
    assert model1.name == 'new_model'

    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))


def test_learning_rule_learning_with_extra_args_dict(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1', name='lr1', params={'output_value': 5})
    rule1.store()
    assert rule1.impl is not None

    dset = Dataset(examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset, params={'additional_output': 2})
    assert model1.learning_rule == rule1
    assert model1.model_family == 'debug_1'
    assert model1.name is None

    assert not np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                              rule1.params.output_value * np.ones(5))
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          (rule1.params.output_value + 2) * np.ones(5))


def test_learning_rule_learning_with_extra_args_dataclass(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1', name='lr1', params={'output_value': 5})
    rule1.store()
    assert rule1.impl is not None

    dset = Dataset(examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset,
                         params=debug_models.TrainParameters(additional_output=2))
    assert model1.learning_rule == rule1
    assert model1.model_family == 'debug_1'
    assert model1.name is None

    assert not np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                              rule1.params.output_value * np.ones(5))
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          (rule1.params.output_value + 2) * np.ones(5))


def test_learning_rule_learning_with_extra_args_kwargs(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1', name='lr1', params={'output_value': 5})
    rule1.store()
    assert rule1.impl is not None

    dset = Dataset(examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset, additional_output=2)
    assert model1.learning_rule == rule1
    assert model1.model_family == 'debug_1'
    assert model1.name is None

    assert not np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                              rule1.params.output_value * np.ones(5))
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          (rule1.params.output_value + 2) * np.ones(5))


def test_learning_rule_learning_train_params_stored(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1', name='lr1', params={'output_value': 5})
    rule1.store()
    assert rule1.impl is not None

    dset = Dataset(examples=np.array([1, 1]), labels=np.array([2, 2]), mldb=mldb)
    model1 = rule1.learn(dataset=dset, additional_output=2)
    assert model1.learning_rule == rule1
    assert model1.model_family == 'debug_1'
    assert model1.name is None

    assert not np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                              rule1.params.output_value * np.ones(5))
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          (rule1.params.output_value + 2) * np.ones(5))
    
    assert len(mldb.get_models()) == 0
    model1.store()
    assert len(mldb.get_models()) == 1
    model1_again = mldb.get_models()[0]
    assert model1_again.uuid == model1.uuid
    assert model1_again != model1
    assert model1_again.params.additional_output == 2


def test_debug_learning_rule_explicitly_defined(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_3', name='lr1', params={'output_value': 5})
    rule1.store()
    assert rule1.impl is not None

    dset = Dataset(examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset, additional_output=2)
    assert model1.learning_rule == rule1
    assert model1.model_family == 'debug_3'
    assert model1.name is None

    assert not np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                              rule1.params.output_value * np.ones(5))
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          (rule1.params.output_value + 2) * np.ones(5))

def test_get_trained_models(mldb):
    rule = mldb.learning_rule(family='debug_1')
    dset = Dataset(examples=np.array([1, 1]), labels=np.array([2, 2]), mldb=mldb)
    model = rule.learn(dataset=dset)
    model.store()
    new_rule = mldb.learning_rule(family='debug_2')
    new_model = new_rule.learn(dataset=dset)
    new_model.store()

    def canon(models):
        return set([model.uuid for model in models])

    assert canon(rule.get_trained_models()) == canon([model])

    rule_again = mldb.learning_rule(family='debug_1')
    assert canon(rule.get_trained_models()) == canon([model])

    model_again = rule_again.learn(dataset=dset)
    model_again.store()
    assert canon(rule.get_trained_models()) == canon([model, model_again])


def test_avoid_duplicate_stores(mldb):
    rule = mldb.learning_rule(family='debug_1')
    duplicate = mldb.learning_rule(family='debug_1')
    rule.store()
    duplicate.store(tag="tag")
    assert len(mldb.get_learning_rules()) == 1
    assert duplicate.db_obj == rule.db_obj
