"""Basic tests for logging functionality."""
from functools import partial
import sys

import basicdb
import numpy as np

from testbed.mldb import MLDB
from testbed.dataset import Dataset
from testbed import learning_rule


def test_no_logging(mldb, rng):
    mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2])).store()

    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    rule1.store()
    assert rule1.impl is not None

    model1 = rule1.learn(dataset='dset1')
    assert model1.learning_rule == rule1
    assert model1.model_family == 'debug_1'
    assert model1.train_completed
    assert len(mldb.get_models()) == 0
    assert not model1.train_log.is_stored
    assert len(model1.train_log.entries) == 0
    assert model1.train_log.parent_model == model1
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))


def test_no_continuous_storing(mldb, rng):
    mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2])).store()

    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    rule1.store()
    assert rule1.impl is not None

    model1 = rule1.learn(dataset='dset1', store=True)
    assert model1.train_completed
    assert len(mldb.get_models()) == 1
    assert model1.train_log.is_stored
    assert len(model1.train_log.entries) == 0
    assert model1.train_log.parent_model == model1
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))


def test_two_log_entries_continuous_storing(mldb, rng):
    mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2])).store()
    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    
    model1 = rule1.learn(dataset='dset1', store=True, num_log_entries=2)
    assert model1.train_completed
    assert len(mldb.get_models()) == 1
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))
    assert model1.train_log.is_stored
    assert len(model1.train_log.entries) == 2
    assert model1.train_log.parent_model == model1
    
    def check_log_entry(train_log, ii):
        assert train_log.entries[str(ii)]['ii'] == ii
        assert train_log.load_data(str(ii)) == [ii, ii]
    
    check_log_entry(model1.train_log, 0)
    check_log_entry(model1.train_log, 1)
    assert model1.train_log.load_data(['0', '1']) == {'0': [0, 0], '1': [1, 1]}


def test_two_log_entries_store_afterwards(mldb, rng):
    mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2])).store()
    rule1 = mldb.learning_rule(family='debug_1', name='lr1')
    
    model1 = rule1.learn(dataset='dset1', num_log_entries=2)
    assert model1.train_completed
    assert len(mldb.get_models()) == 0
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))
    assert not model1.train_log.is_stored
    assert len(model1.train_log.entries) == 2
    assert model1.train_log.parent_model == model1
    
    def check_log_entry(train_log, ii):
        assert train_log.entries[str(ii)]['ii'] == ii
        assert train_log.load_data(str(ii)) == [ii, ii]
    
    check_log_entry(model1.train_log, 0)
    check_log_entry(model1.train_log, 1)
    assert model1.train_log.load_data(['0', '1']) == {'0': [0, 0], '1': [1, 1]}

    model1.store()
    assert len(mldb.get_models()) == 1
    assert model1.is_stored
    assert model1.train_log.is_stored
    assert len(model1.train_log.entries) == 2
    check_log_entry(model1.train_log, 0)
    check_log_entry(model1.train_log, 1)
    assert model1.train_log.load_data(['0', '1']) == {'0': [0, 0], '1': [1, 1]}
