"""Tests for mldb functionality."""
import pytest
import sys

import numpy as np

from testbed.dataset import Dataset
from testbed.model_families.debug_models import DebugModel


def canonicalize(objects):
    return set([x.db_obj.uuid for x in objects])


def test_tags(mldb):
    rule1 = mldb.learning_rule(family="debug_1", name="rule1")
    rule2 = mldb.learning_rule(family="debug_2", name="rule2")
    rule3 = mldb.learning_rule(family="debug_3", name="rule3")
    with pytest.raises(ValueError):
        mldb.add_tag(rule1, "test_exp")
    rule1.store()
    rule2.store(tag="exp1")
    rule3.store()
    mldb.add_tag(rule1, "exp1")
    mldb.add_tag(rule3, "exp2")

    assert canonicalize(mldb.get_learning_rules(tag="exp1")) == canonicalize([rule1, rule2])
    assert canonicalize(mldb.get_learning_rules(tag="exp2")) == canonicalize([rule3])
    assert canonicalize(mldb.get_learning_rules()) == canonicalize([rule1, rule2, rule3])

    assert len(mldb.get_learning_rules(tag="nonsense")) == 0

    dataset = mldb.dataset(examples=np.array([1, 1]), labels=np.array([0, 0]))

    model1 = rule1.learn(dataset=dataset, name="model1")
    model2 = rule2.learn(dataset=dataset, name="model2")
    model3 = rule3.learn(dataset=dataset, name="model3")

    model1.store()
    model2.store(tag="exp1")
    model3.store()
    mldb.add_tag(model1, "exp1")
    mldb.add_tag(model3, "exp2")

    assert canonicalize(mldb.get_models(tag="exp1")) == canonicalize([model1, model2])
    assert canonicalize(mldb.get_models(tag="exp2")) == canonicalize([model3])
    assert canonicalize(mldb.get_models()) == canonicalize([model1, model2, model3])
    assert len(mldb.get_models(tag="nonsense")) == 0

    assert canonicalize(mldb.get_models(tag="exp1", family="debug_1")) == canonicalize([model1])
    assert canonicalize(mldb.get_models(family="debug_3")) == canonicalize([model3])

    eval1 = model1.predict(dataset=dataset)
    eval2 = model2.predict(dataset=dataset)
    eval3 = model3.predict(dataset=dataset)
    eval1.store()
    eval2.store(tag="exp1")
    eval3.store()
    mldb.add_tag(eval1, "exp1")
    mldb.add_tag(eval3, "exp2")

    assert canonicalize(mldb.get_evaluations(tag="exp1")) == canonicalize([eval1, eval2])
    assert canonicalize(mldb.get_evaluations(tag="exp2")) == canonicalize([eval3])
    assert canonicalize(mldb.get_evaluations()) == canonicalize([eval1, eval2, eval3])
    assert len(mldb.get_evaluations(tag="nonsense")) == 0

    assert set(mldb.get_tags()) == set(["exp1", "exp2"])


def test_adding_tags_twice(mldb):
    rule = mldb.learning_rule(family="debug_1", name="rule1")
    rule.store(tag="tag")
    mldb.add_tag(rule, "tag") 


def test_model_import_without_train_set(mldb):
    rule = mldb.learning_rule(family="debug_1", name="rule1")
    rule.store()

    tmp_model = DebugModel(output_value=5,
                           extra_payload=[1, 2, 3])
    tmp_model_bytes = tmp_model.serialize()

    mldb.import_model(learning_rule=rule,
                      model_bytes=tmp_model_bytes,
                      name='model1')
    
    assert len(mldb.get_models()) == 1
    model = mldb.get_models()[0]
    assert model.name == 'model1'
    dset = Dataset(examples=np.array([1, 1, 1]), labels=np.array([2, 2, 2]))
    assert np.array_equal(model.predict(dataset=dset, return_eval=False),
                          5 * np.ones(3))
    assert model.impl.extra_payload == [1, 2, 3]
    assert model.imported
    assert not model.has_train_set


def test_model_import_with_train_set(mldb):
    rule = mldb.learning_rule(family="debug_1", name="rule1")
    rule.store()
    dset = Dataset(examples=np.array([1, 1, 1]), labels=np.array([2, 2, 2]), mldb=mldb)
    dset.store()

    tmp_model = DebugModel(output_value=2,
                           extra_payload=[1, 2, 3])
    tmp_model_bytes = tmp_model.serialize()

    mldb.import_model(learning_rule=rule,
                      model_bytes=tmp_model_bytes,
                      train_set=dset,
                      name='model1')
    
    assert len(mldb.get_models()) == 1
    model = mldb.get_models()[0]
    assert model.name == 'model1'
    assert np.array_equal(model.predict(dataset=dset, return_eval=False),
                          2 * np.ones(3))
    assert model.impl.extra_payload == [1, 2, 3]
    assert model.imported
    assert model.has_train_set
    assert model.train_set.uuid == dset.uuid