"""Basic tests for model functionality."""
from functools import partial
import pytest
import sys

import basicdb
import numpy as np

from testbed.dataset import Dataset
from testbed.evaluation import Evaluation
from testbed.mldb import MLDB
from testbed import model
from testbed.model_families import debug_models


def model_names(mldb):
    return set(x.name for x in mldb.get_models())


def test_mldb(tmp_path):
    """Basic test of mldb functionality."""
    db1 = basicdb.BasicDB(sql_string="sqlite:///:memory:",
                          stash_rootdir=tmp_path,
                          namespace="mockdb_1")
    db2 = basicdb.BasicDB(sql_string="sqlite:///:memory:",
                          stash_rootdir=tmp_path,
                          namespace="mockdb_2")

    mldb1 = MLDB(db1, data_cache_dir=tmp_path)
    mldb2 = MLDB(db2, data_cache_dir=tmp_path)

    rule1 = mldb1.learning_rule(family='debug_1')
    dset1 = mldb1.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset1, name='model1')
    assert model_names(mldb1) == set()
    assert model_names(mldb2) == set()

    model1.store() 
    assert model_names(mldb1) == set(['model1'])
    assert model_names(mldb2) == set()


def test_model_retrieval(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1')
    dset1 = mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset1, name='model1')
    assert model1.model_family == 'debug_1'
    assert model_names(mldb) == set()
    model1.store()
    assert model1.is_stored
    assert model1.has_train_set
    assert not model1.imported
    assert not model1.is_checkpoint
    assert model_names(mldb) == set(['model1'])

    model1_again = mldb.model(name='model1')
    assert model1_again.is_stored
    assert model1_again.db_obj.uuid == model1.db_obj.uuid
    assert model1_again.model_family == 'debug_1'
    
    
    model2 = rule1.learn(dataset=dset1)
    model2.store()
    assert model2.is_stored
    assert model2.uuid != model1.uuid
    assert model2.model_family == 'debug_1'
    assert model_names(mldb) == set(['model1', str(model2.db_obj.uuid)])


def test_extra_train_params(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1')
    dset1 = mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset1, name='model1', additional_output=5)
    assert model1.params.additional_output == 5
    model1.store()
    assert model1.params.additional_output == 5
    
    model1_again = mldb.model(name='model1')
    assert model1_again.is_stored
    assert model1_again.db_obj.uuid == model1.db_obj.uuid
    assert model1_again.params.additional_output == 5


def test_model_parents(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1')
    dset1 = mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset1, name='model1')
    assert model1.learning_rule == rule1
    assert model1.train_set == dset1
    model1.store()
    assert model1.is_stored
    assert dset1.is_stored
    assert rule1.is_stored
    
    model1_again = mldb.model(name='model1')
    assert model1_again.is_stored
    assert model1_again.uuid == model1.uuid
    assert model1_again.learning_rule.uuid == rule1.uuid
    assert model1_again.train_set.uuid == dset1.uuid


def test_predict_from_examples(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1')
    dset1 = mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset1, name='model1')
    assert np.array_equal(model1.predict(examples=np.zeros(5), return_eval=False),
                          rule1.params.output_value * np.ones(5))


def test_predict_from_dataset(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1')
    dset1 = mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset1, name='model1')
    dset2 = Dataset(examples=np.array([1, 1, 1]), labels=np.array([2, 2, 2]))
    assert np.array_equal(model1.predict(dataset=dset2, return_eval=False),
                          rule1.params.output_value * np.ones(3))


def test_predict_from_dataset_name(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1')
    dset1 = mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset1, name='model1')
    mldb.dataset(name='dset2', examples=np.array([1, 1, 1]), labels=np.array([2, 2, 2])).store()
    assert np.array_equal(model1.predict(dataset='dset2', return_eval=False),
                          rule1.params.output_value * np.ones(3))


def test_predict_with_kwargs(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1')
    dset1 = mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset1, name='model1')
    mldb.dataset(name='dset2', examples=np.array([1, 1, 1]), labels=np.array([2, 2, 2])).store()
    assert np.array_equal(model1.predict(dataset='dset2', return_eval=False, output_value=5),
                          5 * np.ones(3))


def test_predict_with_params(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1')
    dset1 = mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset1, name='model1')
    mldb.dataset(name='dset2', examples=np.array([1, 1, 1]), labels=np.array([2, 2, 2])).store()
    assert np.array_equal(model1.predict(dataset='dset2', return_eval=False, params={'output_value': 3}),
                          3 * np.ones(3))


def test_predict_with_model_family_params(mldb, rng):
    rule1 = mldb.learning_rule(family='debug_1')
    dset1 = mldb.dataset(name='dset1', examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule1.learn(dataset=dset1, name='model1')
    mldb.dataset(name='dset2', examples=np.array([1, 1, 1]), labels=np.array([2, 2, 2])).store()
    assert np.array_equal(model1.predict(dataset='dset2',
                                         return_eval=False,
                                         params=debug_models.EvalParameters(output_value=4)),
                          4 * np.ones(3))


def test_return_eval_default(mldb):
    rule = mldb.learning_rule(family='debug_1')
    dset = mldb.dataset(examples=np.array([1, 1]), labels=np.array([2, 2]))
    model = rule.learn(dataset=dset)
    assert isinstance(model.predict(dataset=dset), Evaluation)
    assert isinstance(model.predict(dataset=dset, return_eval=False), np.ndarray)
    assert np.allclose(model.predict(dataset=dset, return_eval=False),
                       rule.params.output_value)


def test_model_rename_stored(mldb):
    rule = mldb.learning_rule(family='debug_1')
    dset = mldb.dataset(examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule.learn(dataset=dset)
    model1.store()
    assert model1.name == str(model1.uuid)
    new_name = 'new_name'
    model1.rename(new_name)
    assert model1.name == new_name

    model_again = model.Model(name=new_name, mldb=mldb)
    assert model_again.name == new_name
    assert model_again.uuid == model1.uuid


def test_model_rename_not_stored(mldb):
    rule = mldb.learning_rule(family='debug_1')
    dset = mldb.dataset(examples=np.array([1, 1]), labels=np.array([2, 2]))
    model1 = rule.learn(dataset=dset)
    assert model1.name == None
    new_name = 'new_name'
    model1.rename(new_name)
    assert model1.name == new_name
    model1.store()
    assert model1.name == new_name

    model_again = model.Model(name=new_name, mldb=mldb)
    assert model_again.name == new_name
    assert model_again.uuid == model1.uuid


def test_get_evals(mldb):
    rule = mldb.learning_rule(family='debug_1')
    dset1 = mldb.dataset(examples=np.array([1, 1]), labels=np.array([2, 2]))
    dset2 = mldb.dataset(examples=np.array([1, 1]), labels=np.array([2, 2]))
    model = rule.learn(dataset=dset1)
    eval1 = model.predict(dataset=dset1)

    assert len(model.get_evaluations()) == 0

    eval1.store()
    assert len(model.get_evaluations()) == 1
    assert model.get_evaluations()[0].uuid == eval1.uuid
