"""Testing  neural learning rule functionality."""
import pytest
import sys

import basicdb
import numpy as np
import torch

from testbed.dataset import Dataset
from testbed import learning_rule
from testbed import init_db


@pytest.mark.gpu
def test_gpu(mldb, rng):
    seed = rng.randint(0, 99999)

    init_db.add_cifar10(mldb)
    cifar10_train = mldb.dataset(name="cifar10-train")
    cifar10_test = mldb.dataset(name="cifar10-test")
    
    rule = mldb.learning_rule(family='ResNet18', total_examples=50000*10)

    # test that network actually learns
    model = rule.learn(dataset=cifar10_train, name='model', seed=seed, device='cuda:0')
    evals1 = model.predict(dataset=cifar10_test, device='cuda:0')
    assert evals1.accuracy >= 0.7

    # test seeding during training eliminates randomness
    model2 = rule.learn(dataset=cifar10_train, name='model2', seed=seed, device='cuda:0')
    evals2 = model2.predict(dataset=cifar10_test, device='cuda:0')
    assert np.allclose(evals1.predictions, evals2.predictions)

    # test model serializing & deserializing
    model.store()
    model_again = mldb.model(name='model')
    evals3 = model_again.predict(dataset=cifar10_test, device='cuda:0')
    assert np.array_equal(evals1.predictions, evals3.predictions)


def test_cpu(mldb, rng):
    seed = rng.randint(0, 99999)

    init_db.add_cifar10(mldb)
    cifar10_train = mldb.dataset(name="cifar10-train").subsample(size=5000, seed=seed)
    cifar10_test = mldb.dataset(name="cifar10-test").subsample(size=1000, seed=seed)
    
    rule = mldb.learning_rule(family='LeNet', total_examples=5000*1)
    rule.store()
    
    # Check learning rule lookups work correctly
    assert mldb.learning_rule(family='LeNet', total_examples=5000*1).uuid == rule.uuid

    # test that network actually learns
    model = rule.learn(dataset=cifar10_train, name='model', seed=seed, num_workers=0)
    evals1 = model.predict(dataset=cifar10_test, num_workers=0)
    assert evals1.accuracy >= 0.13

    # test seeding during training eliminates randomness
    model2 = rule.learn(dataset=cifar10_train, name='model2', seed=seed, num_workers=0)
    evals2 = model2.predict(dataset=cifar10_test, num_workers=0)
    assert np.allclose(evals1.predictions, evals2.predictions)

    # test model serializing & deserializing
    model.store()
    model_again = mldb.model(name='model')
    evals3 = model_again.predict(dataset=cifar10_test, num_workers=0)
    assert np.array_equal(evals1.predictions, evals3.predictions)


def test_checkpointing(mldb, rng):
    seed = rng.randint(0, 99999)

    init_db.add_cifar10(mldb)
    cifar10_train = mldb.dataset(name="cifar10-train").subsample(size=10, seed=seed)
    
    rule = mldb.learning_rule(family='LeNet', total_examples=10*30)
    model = rule.learn(dataset=cifar10_train, name='model', seed=seed, num_workers=0, additional_checkpoint_epochs=[*range(30)])

    assert len(model.checkpoints) == 30


def test_subsample_logits(mldb, rng):
    seed = rng.randint(0, 99999)

    init_db.add_cifar10(mldb)
    cifar10_train = mldb.dataset(name="cifar10-train").subsample(size=10, seed=seed)

    rule = mldb.learning_rule(family='LeNet', total_examples=10*30)
    model = rule.learn(dataset=cifar10_train, name='model', seed=seed, num_workers=0)

    all_classes_eval = model.predict(dataset=cifar10_train, num_workers=0)
    all_classes_logits = all_classes_eval.extra_prediction_data

    cifar10_train.params.classes_to_exclude = [0, 9]
    subsample_classes_eval = model.predict(dataset=cifar10_train, num_workers=0)
    subsample_classes_logits = subsample_classes_eval.extra_prediction_data

    assert np.allclose(all_classes_logits[:, 1:-1], subsample_classes_logits)
