""""Transfer eval data from DB to a convenient pickle file."""
import pickle
import click
import os
import multiprocessing as mp
import numpy as np
import tqdm
from dataclasses import asdict
import datetime

import testbed


def extract_evals(print_progress, model_uuids, db_path):
    mldb = testbed.mldb.get_semilocal_mldb(db_path)
    models = mldb.get_models(uuids=model_uuids)
    if print_progress:
        models = tqdm.tqdm(models)
    evals = []
    for model in models:
        if model.is_checkpoint:
            model_id = str(model.parent_checkpoint.parent_model.uuid)
            epoch = model.parent_checkpoint.sequence_number
        else:
            model_id = str(model.uuid)
            epoch = -1
        model_evals = []
        for ev in model.get_evaluations():
            model_evals.append(dict(rule=model.learning_rule.name,
                                    rule_params=asdict(model.learning_rule.params),
                                    rule_params_class=model.learning_rule.params.__class__.__name__,
                                    model_family=model.model_family,
                                    model_name=model.name,
                                    epoch=epoch,
                                    test_set=ev.test_set.name,
                                    train_set=model.train_set.name if model.train_set is not None else 'na',
                                    checkpoint_id=str(model.uuid),
                                    model_id=model_id,
                                    eval_id=str(ev.uuid),
                                    stats=ev.computed_statistics
                                    ))
        evals.extend(model_evals)
    return evals


def extract_evals_(x):
    return extract_evals(*x)


@click.command()
@click.option('--out_path', type=str, default=f'../shared_mldb_root/mldb_{datetime.date.today()}.pickle')
@click.option('--db_path', type=str, default='../shared_mldb_root/exports/postgres_export.sqlite')
@click.option('--num_workers', type=int, default=os.cpu_count())
def pickle_evals(out_path, db_path, num_workers):
    if os.path.exists(out_path):
        raise ValueError(f'Output path {out_path} aleady exists!')

    mldb = testbed.mldb.get_semilocal_mldb(db_path)

    print('Loading all models in the DB')
    all_models = np.asarray([str(m.uuid) for m in mldb.get_models()])

    print(f'Extracting evals from {len(all_models)} models using {num_workers} '
          f'workers and saving the result to {out_path}')
    chunks = np.split(all_models, np.linspace(
        0, len(all_models), num_workers, endpoint=False).astype(int)[1:])
    inputs = [(i == 0, chunk, db_path) for i, chunk in enumerate(chunks)]

    with mp.Pool(num_workers) as pool:
        evals = [e for evs in pool.map(extract_evals_, inputs)
                 for e in evs]

    with open(out_path, 'wb') as f:
        pickle.dump(evals, f)


if __name__ == '__main__':
    pickle_evals()